/// <reference types="cypress" />
// ***********************************************************
// This example plugins/index.js can be used to load plugins
//
// You can change the location of this file or turn off loading
// the plugins file with the 'pluginsFile' configuration option.
//
// You can read more here:
// https://on.cypress.io/plugins-guide
// ***********************************************************

// This function is called when a project is opened or re-opened (e.g. due to
// the project's config changing)

/**
 * @type {Cypress.PluginConfig}
 */
const MongoClient = require('mongodb').MongoClient;
const Binary = require('mongodb').Binary;
const ObjectId = require('mongodb').ObjectId;

module.exports = (on, config) => {
	on('task', {
		initBinaryFileData({collection, data}) {
			console.log('initBinaryFileData');
			insertIntoDatabase(config, collection, parseBinaryFileData(data));
			return 0;
		},
		initData({collection, data}) {
			console.log('initData');
			insertIntoDatabase(config, collection, parseData(data));
			return 0;
		},
		dropCollection({collection}) {
			console.log('dropCollection');
			dropCollectionFromDatabase(config, collection);
			return 0;
		}
	});
};

function parseBinaryFileData(binaryFiles) {
	binaryFiles.forEach(binaryFile => parseBinaryFile(binaryFile));
	return binaryFiles;
}

function parseBinaryFile(binaryFile) {
	binaryFile._id = createObjectId(binaryFile);
	binaryFile.vorgangId = createObjectIdString(binaryFile.vorgangId);
	binaryFile.file.content = createBinary(binaryFile.file);
	binaryFile.createdAt = createDate(binaryFile.createdAt);
}

function createObjectId(field) {
	return ObjectId(field._id.$oid);
}

function parseData(data) {
	data.forEach(vorgang => parseVorgang(vorgang));
	return data;
}

function parseVorgang(vorgang) {
	vorgang.createdAt = createDate(vorgang.createdAt);
	vorgang._id = createObjectIdString(vorgang._id);

	vorgang.eingangs.forEach(eingang => parseEingang(eingang));

	if (vorgang.wiedervorlages) {
		vorgang.wiedervorlages.forEach(wiedervorlage => parseWiedervorlage(wiedervorlage));
	}
}

function createObjectIdString(field) {
	return `ObjectId("${field.$oid}")`;
}

function parseEingang(eingang) {
	eingang.header.createdAt = createDate(eingang.header.createdAt);

	if (eingang.attachments) {
		eingang.attachments.forEach(attachment => parseAttachment(attachment));
	}

	if (eingang.representations) {
		eingang.representations.forEach(representation => parseRepresentations(representation));
	}
}

function parseWiedervorlage(wiedervorlage) {
	wiedervorlage.frist = new Date(wiedervorlage.frist.$date);
	wiedervorlage.createdAt = createDate(wiedervorlage.createdAt);
}

function createDate(field) {
	return new Date(field.$date);
}

function parseAttachment(attachment) {
	attachment.files[ 0 ].content = createBinary(attachment.files[ 0 ]);
}

function parseRepresentations(representation) {
	representation.content = createBinary(representation);
}

function createBinary(field) {
	return Binary(field.content.$binary.base64, 'base64');
}

function insertIntoDatabase(config, collection, data) {
	MongoClient.connect(buildDatabaseUrl(config), (error, connection) => {
		if (!error) {
			var db = connection.db(config.env.database);

			db.collection(collection).drop(() => {
				db.createCollection(collection, (error) => handleCreateCollection(db, connection, collection, data, error));
			});
		}
	});
}

function handleCreateCollection(db, connection, collection, data, error) {
	if (error) {
		console.log(`Fehler beim Erstellen der Collection "${collection}": `, error);
	} else {
		console.log(`Collection ${collection} erfolgreich erstellt`);
		insertManyToDatabase(db, connection, collection, data);
	}
}

function insertManyToDatabase(db, connection, collection, data) {
	db.collection(collection).insertMany(data, (error) => handleInsertMany(connection, error));
}

function handleInsertMany(connection, error) {
	if (error) {
		console.log('Fehler beim Einlesen der Daten: ', error);
	} else {
		console.log('Die Daten wurden erfolgreich eingelesen.');
	}
	connection.close();
}

function dropCollectionFromDatabase(config, collection) {
	MongoClient.connect(buildDatabaseUrl(config), (error, connection) => {
		if (!error) {
			var db = connection.db(config.env.database);
			db.collection(collection).drop(() => {
				connection.close();
			});
		}
	});
}

function buildDatabaseUrl(config) {
	return `mongodb://${config.env.dbUrl}/`;
}