import { mock, useFromMock } from '@goofy-client/test-utils';
import { VorgangWithEingangLinkRel } from '@goofy-client/vorgang-shared';
import { Resource, ResourceFactory } from '@ngxp/rest';
import * as faker from 'faker';
import { cold, hot } from 'jest-marbles';
import { createCommand, createCommandListResource, createCommandResource } from 'libs/command-shared/test/command';
import { toResource } from 'libs/tech-shared/test/resource';
import { CommandLinkRel } from './command.linkrel';
import { Command, CommandListResource, CommandResource, CommandStatus } from './command.model';
import { CommandRepository } from './command.repository';

describe('CommandRepository', () => {

	let repository: CommandRepository;
	let resourceFactory = mock(ResourceFactory);
	let resourceWrapper = { post: jest.fn(), get: jest.fn(), patch: jest.fn() };

	const command: Command = createCommand();
	const commandResource: CommandResource = createCommandResource();
	const commandListResource: CommandListResource = createCommandListResource();
	const commandResourceWithPendingCommand: CommandResource = createCommandResource([VorgangWithEingangLinkRel.PENDING_COMMANDS]);

	beforeEach(() => {
		repository = new CommandRepository(useFromMock(resourceFactory));
		resourceFactory.from.mockReturnValue(resourceWrapper);
	})

	describe('create command', () => {

		const linkRel: string = faker.random.word();
		const resource: Resource = toResource({}, [linkRel]);

		beforeEach(() => {
			resourceWrapper.post.mockReturnValue(hot('a', { a: commandResource }));
		})

		it('should call resourceFactory with resource', () => {
			repository.createCommand(resource, linkRel, command);

			expect(resourceFactory.from).toHaveBeenCalledWith(resource);
		})

		it('should call resourceWrapper with linkel and command', () => {
			repository.createCommand(resource, linkRel, command);

			expect(resourceWrapper.post).toHaveBeenCalledWith(linkRel, command);
		})

		it('should return value', () => {
			const expected: CommandResource = commandResource;

			const result = repository.createCommand(resource, linkRel, command);

			expect(result).toBeObservable(cold('a', { a: expected }));
		})
	})

	describe('get command', () => {

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(hot('a', { a: commandResource }));
		})

		it('should call resourceFactory with resource', () => {
			repository.getCommand(commandResource);

			expect(resourceFactory.from).toHaveBeenCalledWith(commandResource);
		})

		it('should call resourceWrapper with selfLink', () => {
			repository.getCommand(commandResource);

			expect(resourceWrapper.get).toHaveBeenCalledWith(CommandLinkRel.SELF);
		})

		it('should return value', () => {
			const result = repository.getCommand(commandResource);

			expect(result).toBeObservable(cold('a', { a: commandResource }));
		})
	})

	describe('revoke command', () => {

		beforeEach(() => {
			resourceWrapper.patch.mockReturnValue(hot('a', { a: commandResource }));
		})

		it('should call resourceFactory with Resource', () => {
			repository.revokeCommand(commandResource);

			expect(resourceFactory.from).toHaveBeenCalledWith(commandResource);
		})

		it('should call resourceWrapper with deleteLink', () => {
			repository.revokeCommand(commandResource);

			expect(resourceWrapper.patch).toHaveBeenCalledWith(CommandLinkRel.REVOKE, { status: CommandStatus.REVOKED })
		})

		it('should return value', () => {
			const result = repository.revokeCommand(commandResource);

			expect(result).toBeObservable(cold('a', { a: commandResource }));
		})
	})

	describe('get pending commands', () => {
		const linkRel = VorgangWithEingangLinkRel.PENDING_COMMANDS;

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(hot('a', { a: commandListResource }));
		})

		it('should call resourceFactory with resource', () => {
			repository.getPendingCommands(commandResourceWithPendingCommand, linkRel);

			expect(resourceFactory.from).toHaveBeenCalledWith(commandResourceWithPendingCommand);
		})

		it('should call resourceWrapper with linkRel', () => {
			repository.getPendingCommands(commandResourceWithPendingCommand, linkRel);

			expect(resourceWrapper.get).toHaveBeenCalledWith(linkRel);
		})

		it('should return value', () => {
			const result = repository.getPendingCommands(commandResourceWithPendingCommand, linkRel);

			expect(result).toBeObservable(cold('a', { a: commandListResource }));
		})
	})

	describe('get effected resource', () => {
		const command: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);
		const listResource = {};

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(hot('a', { a: listResource }));
		})

		it('should call resourceFactory with resource', () => {
			repository.getEffectedResource(command);

			expect(resourceFactory.from).toHaveBeenCalledWith(command);
		})

		it('should call resourceWrapper with linkRel', () => {
			repository.getEffectedResource(command);

			expect(resourceWrapper.get).toHaveBeenCalledWith(CommandLinkRel.EFFECTED_RESOURCE);
		})

		it('should return value', () => {
			const result = repository.getEffectedResource(command);

			expect(result).toBeObservable(cold('a', { a: listResource }));
		})
	})
})
