import { ApiError, createEmptyStateResource, createStateResource, StateResource } from '@goofy-client/tech-shared';
import { mock, Mock, useFromMock } from '@goofy-client/test-utils';
import { Resource } from '@ngxp/rest';
import * as faker from 'faker';
import { cold, hot } from 'jest-marbles';
import { createCommand, createCommandResource } from 'libs/command-shared/test/command';
import { createApiError } from 'libs/tech-shared/test/error';
import { toResource } from 'libs/tech-shared/test/resource';
import { Observable } from 'rxjs';
import { CommandLinkRel } from './command.linkrel';
import { Command, CommandResource } from './command.model';
import { CommandRepository } from './command.repository';
import { CommandService, startInterval } from './command.service';

describe('CommandService', () => {
	let service;
	let repository: Mock<CommandRepository>;

	const command: Command = createCommand();

	const commandResource: CommandResource = createCommandResource();
	const commandStateResource: StateResource<CommandResource> = createStateResource(commandResource);

	const commandResourceWithUpdateLink: CommandResource = createCommandResource([CommandLinkRel.UPDATE]);

	beforeEach(() => {
		repository = mock(CommandRepository);
		service = new CommandService(useFromMock(repository));
	})

	describe('create command', () => {

		const commandResource: CommandResource = commandResourceWithUpdateLink;

		const resource: Resource = toResource({});
		const linkRel: string = faker.random.word();

		beforeEach(() => {
			repository.createCommand.mockReturnValue(cold('a', { a: commandResource }));
		})

		it('should call repository with resource, linkrel and command', () => {
			service.createCommand(resource, linkRel, command);

			expect(repository.createCommand).toHaveBeenCalledWith(resource, linkRel, command);
		})

		describe('recieve apiError', () => {

			const apiError: ApiError = createApiError();

			it('should return apiError as stateResource', () => {
				repository.createCommand.mockReturnValue(hot('a', { a: apiError }));

				const result = service.createCommand(resource, linkRel, command);

				expect(result).toBeObservable(cold('a', { a: createStateResource(apiError) }));
			})
		})
	})

	describe('revoke command', () => {

		beforeEach(() => {
			repository.revokeCommand.mockReturnValue(cold('a', { a: commandResource }));
		})

		it('should call repository with resource, linkrel and command', () => {
			service.revokeCommand(commandResource);

			expect(repository.revokeCommand).toHaveBeenCalledWith(commandResource);
		})

		describe('process command with status in progress(with update link)', () => {

			beforeEach(() => {
				repository.revokeCommand.mockReturnValue(cold('a', { a: commandResourceWithUpdateLink }));
				service.pollCommand = jest.fn();
				service.pollCommand.mockReturnValue(cold('a', { a: createStateResource(commandResourceWithUpdateLink, true) }));
			})

			it('should return value with loading true', () => {
				const result = service.revokeCommand(commandResource);

				expect(result).toBeObservable(hot('a', { a: createStateResource(commandResourceWithUpdateLink, true) }));
			})
		})

		describe('process command with status done(no update link)', () => {

			it('should return value with loading false', () => {
				const result = service.revokeCommand(commandResource);

				expect(result).toBeObservable(hot('a', { a: commandStateResource }));
			})
		})
	})

	describe('handle interval', () => {

		const interval = { handle: () => null };

		beforeEach(() => {
			service.clearInterval = jest.fn();
		})

		it('should call clearInterval', () => {
			service.handleInterval(createEmptyStateResource(), interval);

			expect(service.clearInterval).toHaveBeenCalledWith(interval.handle);
		})

		it('should NOT call clearInterval', () => {
			service.handleInterval(createEmptyStateResource(true), interval);

			expect(service.clearInterval).not.toHaveBeenCalled();
		})
	})

	describe('getAndUpdate', () => {

		beforeEach(() => {
			repository.getCommand.mockReturnValue(cold('a', { a: commandResourceWithUpdateLink }));
		})

		it('should call getCommand', () => {
			service.getAndUpdate(commandResource);

			expect(repository.getCommand).toHaveBeenCalled();
		})

		it('should return stateResource still loading', () => {
			const result = service.getAndUpdate(commandResource);

			expect(result).toBeObservable(hot('a', { a: createStateResource(commandResourceWithUpdateLink, true) }));
		})

		describe('command is loaded', () => {

			beforeEach(() => {
				repository.getCommand.mockReturnValue(cold('a', { a: commandResource }));
			})

			it('should loaded stateResource', () => {
				const result = service.getAndUpdate(commandResource);

				expect(result).toBeObservable(hot('a', { a: createStateResource(commandResource) }));
			})
		})
	})

	describe('get command', () => {

		beforeEach(() => {
			repository.getCommand.mockReturnValue(cold('a', { a: commandResource }));
		})

		it('should call repository with resource', () => {
			service.getCommand(commandResource);

			expect(repository.getCommand).toHaveBeenCalledWith(commandResource);
		})

		it('should return value', () => {
			const result = service.getCommand(commandResource);

			expect(result).toBeObservable(hot('a', { a: commandResource }))
		})
	})

	describe('get effected resource', () => {

		const listResource = {};
		const command: CommandResource = createCommandResource();

		beforeEach(() => {
			repository.getEffectedResource.mockReturnValue(cold('a', { a: listResource }));
		})

		it('should call repository', () => {
			service.getEffectedResource(command);

			expect(repository.getEffectedResource).toHaveBeenCalledWith(command);
		})

		it('should return value', () => {
			const result = service.getEffectedResource(command);

			expect(result).toBeObservable(hot('a', { a: listResource }))
		})
	})

	describe('get pending commands', () => {

		const listResource = {};

		beforeEach(() => {
			repository.getPendingCommands.mockReturnValue(cold('a', { a: listResource }));
		})

		it('should call repository', () => {
			service.getPendingCommands(command, CommandLinkRel.SELF);

			expect(repository.getPendingCommands).toHaveBeenLastCalledWith(command, CommandLinkRel.SELF);
		})
	})

	describe('start interval', () => {

		const intervall: number = 100;

		it('should not be null', () => {
			var result = startInterval(intervall);

			expect(result).not.toBeNull();
		})

		it('should have handle not null', () => {
			var result = startInterval(intervall);

			expect(result.handle).not.toBeNull();
		})

		it('should have tick as observable', () => {
			var result = startInterval(intervall);

			expect(result.tickObservable).toBeInstanceOf(Observable);
		})
	})
})