import { Injectable } from '@angular/core';
import { createErrorStateResource, createStateResource, StateResource } from '@goofy-client/tech-shared';
import { Resource } from '@ngxp/rest';
import { Observable, of, Subject } from 'rxjs';
import { catchError, flatMap, map, mergeMap, tap } from 'rxjs/operators';
import { CommandListResource, CommandResource, CreateCommand } from './command.model';
import { CommandRepository } from './command.repository';
import { isPending } from './command.util';

@Injectable({ providedIn: 'root' })
export class CommandService {

	intervalTimer: number = 500;

	constructor(private repository: CommandRepository) { }

	public createCommand(resource: Resource, linkRel: string, command: CreateCommand): Observable<StateResource<CommandResource>> {
		return this.repository.createCommand(resource, linkRel, command).pipe(
			mergeMap(createdCommand => this.startPolling(createdCommand)),
			catchError(apiError => of(createErrorStateResource(apiError))) //TODO only status code 422
		);
	}

	public revokeCommand(resource: CommandResource): Observable<StateResource<CommandResource>> {
		return this.repository.revokeCommand(resource)
			.pipe(mergeMap(commandResource => this.startPolling(commandResource)));
	}

	startPolling(commandResource: CommandResource): Observable<StateResource<CommandResource>> {
		return isPending(commandResource) ? this.pollCommand(commandResource) : of(createStateResource(commandResource));
	}

	pollCommand(commandResource: CommandResource): Observable<StateResource<CommandResource>> {
		const interval: IntervallHandleWithTickObservable = startInterval(this.intervalTimer);
		return interval.tickObservable.pipe(
			flatMap(() => this.getAndUpdate(commandResource)),
			tap(stateResource => this.handleInterval(stateResource, interval))
		)
	}

	handleInterval(stateResource: StateResource<CommandResource>, interval: IntervallHandleWithTickObservable): void {
		if (!stateResource.loading) this.clearInterval(interval.handle);
	}

	getAndUpdate(commandResource: CommandResource): Observable<StateResource<CommandResource>> {
		return this.getCommand(commandResource).pipe(map(res => createStateResource(res, isPending(res))));
	}

	getCommand(resource: CommandResource): Observable<CommandResource> {
		return this.repository.getCommand(resource);
	}

	clearInterval(handler): void {
		window.clearInterval(handler);
	}

	getPendingCommands(resource: Resource, linkRel: string): Observable<StateResource<CommandListResource>> {
		return this.repository.getPendingCommands(resource, linkRel).pipe(map(res => createStateResource(res)));
	}

	getEffectedResource(command: CommandResource): Observable<any> {
		return this.repository.getEffectedResource(command);
	}
}

export interface IntervallHandleWithTickObservable {
	handle: number,
	tickObservable: Observable<any>
}

export function startInterval(interval: number): IntervallHandleWithTickObservable {
	const subj: Subject<any> = new Subject();
	const handle = setInterval(() => subj.next('tick'), interval);

	return { handle, tickObservable: subj.asObservable() };
}
