import { Injectable } from '@angular/core';
import { Params } from '@angular/router';
import { CommandResource, CommandService, CreateCommand, isDone } from '@goofy-client/command-shared';
import { createEmptyStateResource, createStateResource, doIfLoadingRequired, NavigationService, StateResource } from '@goofy-client/tech-shared';
import { VorgangResource, VorgangService } from '@goofy-client/vorgang-shared';
import { Resource } from '@ngxp/rest';
import { isNil } from 'lodash';
import { BehaviorSubject, Observable, Subscription } from 'rxjs';
import { startWith, tap } from 'rxjs/operators';
import { KommentarLinkRel, KommentarListLinkRel } from './kommentar.linkrel';
import { CreateKommentarCommand, Kommentar, KommentarListResource, KommentarOrder, KommentarResource } from './kommentar.model';
import { KommentarRepository } from './kommentar.repository';

@Injectable({ providedIn: 'root' })
export class KommentarService {

	readonly kommentarList$: BehaviorSubject<StateResource<KommentarListResource>> = new BehaviorSubject(createEmptyStateResource());
	readonly formularVisibility$: BehaviorSubject<boolean> = new BehaviorSubject(false);

	private navigationSub: Subscription;

	constructor(
		private repository: KommentarRepository,
		private commandService: CommandService,
		private navigationService: NavigationService,
		private vorgangService: VorgangService
	) {
		this.listenToNavigation();
	}

	private listenToNavigation(): void {
		this.unsubscribe();
		this.navigationSub = this.navigationService.urlChanged().subscribe((params: Params) => this.onNavigation(params));
	}

	private unsubscribe(): void {
		if (!isNil(this.navigationSub)) this.navigationSub.unsubscribe();
	}

	onNavigation(params: Params): void {
		if (NavigationService.isVorgangListPage(params)) {
			this.kommentarList$.next({ ...this.kommentarList$.value, reload: true });
		}
	}

	public getKommentareByVorgang(vorgang: VorgangResource): Observable<StateResource<KommentarListResource>> {
		doIfLoadingRequired(this.kommentarList$.value, () => this.loadKommentare(vorgang));
		return this.kommentarList$.asObservable();
	}

	private loadKommentare(vorgang: VorgangResource): void {
		this.setListLoadingTrue();

		const sub: Subscription = this.repository.findKommentare(vorgang).subscribe((kommentarList: KommentarListResource) => {
			this.setKommentarList(kommentarList);
			sub.unsubscribe();
		});
	}

	setListLoadingTrue(): void {
		this.kommentarList$.next({ ...this.kommentarList$.value, loading: true });
	}

	setKommentarListReload() {
		this.kommentarList$.next({ ...this.kommentarList$.value, reload: true });
	}

	setKommentarList(kommentarList: KommentarListResource): void {
		this.kommentarList$.next(createStateResource(kommentarList));
	}

	public isFormularVisible(): Observable<boolean> {
		return this.formularVisibility$.asObservable();
	}

	public showFormular(): void {
		this.formularVisibility$.next(true);
	}

	public hideFormular(): void {
		this.formularVisibility$.next(false);
	}

	public createKommentar(kommentar: Kommentar): Observable<StateResource<CommandResource>> {
		return this.createKommentarCommand(this.kommentarList$.value.resource, KommentarListLinkRel.CREATE_KOMMENTAR, this.createCreateKommentarCommand(kommentar));
	}

	createCreateKommentarCommand(kommentar: Kommentar): CreateKommentarCommand {
		return { order: KommentarOrder.CREATE_KOMMENTAR, kommentar };
	}

	public editKommentar(kommentar: KommentarResource, toPatch: Kommentar): Observable<StateResource<CommandResource>> {
		return this.createKommentarCommand(kommentar, KommentarLinkRel.EDIT, this.createEditKommentarCommand(toPatch));
	}

	createEditKommentarCommand(kommentar: Kommentar): CreateKommentarCommand {
		return { order: KommentarOrder.EDIT_KOMMENTAR, kommentar }
	}

	createKommentarCommand(resource: Resource, linkRel: string, command: CreateCommand): Observable<StateResource<CommandResource>> {
		return this.commandService.createCommand(resource, linkRel, command).pipe(
			tap((createdCommand: StateResource<CommandResource>) => this.afterCreateOrEditKommentar(createdCommand)),
			startWith(createEmptyStateResource<CommandResource>(true)));
	}

	afterCreateOrEditKommentar(command: StateResource<CommandResource>): void {
		if (isDone(command.resource)) {
			this.hideFormular();
			this.setKommentarListReload();
			this.vorgangService.reloadCurrentVorgang();
		}
	}
}