import { HttpClient, HttpHeaders } from '@angular/common/http';
import { mock, useFromMock } from '@goofy-client/test-utils';
import { getUrl, ResourceFactory, ResourceUri } from '@ngxp/rest';
import { cold, hot } from 'jest-marbles';
import { createOzgFileListResource, createOzgFileResource } from 'libs/ozg-file-shared/test/ozg-file';
import { WiedervorlageLinkRel } from 'libs/wiedervorlage-shared/src/lib/wiedervorlage.linkrel';
import { WiedervorlageResource } from 'libs/wiedervorlage-shared/src/lib/wiedervorlage.model';
import { createWiedervorlageResource } from 'libs/wiedervorlage-shared/test/wiedervorlage';
import { OzgFileLinkRel } from './ozg-file.linkrel';
import { OzgFileListResource, OzgFileResource } from './ozg-file.model';
import { GetRequestOptions, OzgFileRepository } from './ozg-file.repository';

describe('OzgFileRepository', () => {
	const httpClient = mock(HttpClient);
	let repository;
	let resourceFactory = mock(ResourceFactory);
	let resourceWrapper = { get: jest.fn() };

	beforeEach(() => {
		repository = new OzgFileRepository(useFromMock(httpClient), useFromMock(resourceFactory));

		resourceFactory.from.mockReturnValue(resourceWrapper);
		resourceFactory.fromId.mockReturnValue(resourceWrapper);
	})

	describe('download', () => {

		const blob = {};
		const ozgFileResource: OzgFileResource = createOzgFileResource([OzgFileLinkRel.DOWNLOAD]);

		beforeEach(() => {
			httpClient.get.mockReturnValue(hot('a', { a: blob }));
		})

		it('should call httpClient', () => {
			const requestOptions = {};
			repository.buildRequestOptions = jest.fn();
			repository.buildRequestOptions.mockReturnValue(requestOptions);

			repository.download(ozgFileResource);

			expect(httpClient.get).toHaveBeenCalledWith(getUrl(ozgFileResource, OzgFileLinkRel.DOWNLOAD), requestOptions);
		})

		it('should return value', () => {
			const result = repository.download(ozgFileResource);

			expect(result).toBeObservable(cold('b', { b: blob }));
		})

		describe('buildRequestOptions', () => {

			it('should return httpHeaders', () => {
				const result: GetRequestOptions = repository.buildRequestOptions();

				expect(result.responseType).toEqual('blob');
			})

			it('should return responseType', () => {
				const result: GetRequestOptions = repository.buildRequestOptions();

				expect(result.headers).toEqual(getExpectedRequestOptions());
			})
		})

		function getExpectedRequestOptions(): HttpHeaders {
			let headers = new HttpHeaders();
			headers = headers.set('Accept', ['application/*', 'images/*']);
			return headers;
		}
	})

	describe('getFiles', () => {

		const ozgFileListResource: OzgFileListResource = createOzgFileListResource();
		const wiedervorlageResource: WiedervorlageResource = createWiedervorlageResource([WiedervorlageLinkRel.ATTACHMENTS]);

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(cold('a', { a: ozgFileListResource }));
		})

		it('should call resource factory', () => {
			repository.getFiles(wiedervorlageResource, WiedervorlageLinkRel.ATTACHMENTS);

			expect(resourceFactory.from).toHaveBeenCalledWith(wiedervorlageResource);
		})

		it('should call resource wrapper get with linkRel', () => {
			repository.getFiles(wiedervorlageResource, WiedervorlageLinkRel.ATTACHMENTS);

			expect(resourceWrapper.get).toHaveBeenCalledWith(WiedervorlageLinkRel.ATTACHMENTS);
		})
	})

	describe('get file', () => {

		const uri: ResourceUri = getUrl(createOzgFileResource());

		it('should call repository factory', () => {
			repository.getFile(uri);

			expect(resourceFactory.fromId).toHaveBeenCalledWith(uri);
		})

		it('should call repository wrapper', () => {
			repository.getFile(uri);

			expect(resourceWrapper.get).toHaveBeenCalled();
		})
	})
})