import { Injectable } from '@angular/core';
import { ActivatedRoute, NavigationEnd, Params, PRIMARY_OUTLET, Router } from '@angular/router';
import { isNil, isUndefined } from 'lodash-es';
import { BehaviorSubject } from 'rxjs';
import { Observable } from 'rxjs/internal/Observable';
import { filter, map, switchMap } from 'rxjs/operators';
import { isEmptyObject } from '../tech.util';

@Injectable({ providedIn: 'root' })
export class NavigationService {

	private readonly routeParameter$: BehaviorSubject<Params> = new BehaviorSubject({});

	constructor(private router: Router, private route: ActivatedRoute) {
		this.init();
	}

	private init(): void {
		this.prefillRouterData();

		this.router.events.pipe(
			filter((event) => event instanceof NavigationEnd),
			map(() => this.getLastFirstChild(this.route)),
			filter(route => route.outlet === PRIMARY_OUTLET),
			switchMap(route => route.params)).subscribe(params => {
				this.routeParameter$.next(params);
			});
	}

	private prefillRouterData(): void {
		this.getLastFirstChild(this.route.root).params.subscribe((params: Params) => this.routeParameter$.next(params));
	}

	private getLastFirstChild(route: ActivatedRoute): ActivatedRoute {
		while (route.firstChild) route = route.firstChild;
		return route;
	}

	public urlChanged(): Observable<Params> {
		return this.routeParameter$.asObservable();
	}

	public getParam(paramName: string): string {
		return this.routeParameter$.value[paramName];
	}

	public getDecodedParam(paramName: string): string {
		if (this.isParamPresent(paramName)) return undefined;
		return NavigationService.decodeParamUrl(this.routeParameter$.value[paramName]);
	}

	static decodeParamUrl(encodedUrl: string): string {
		let decodeUrl: string;

		try {
			decodeUrl = atob(decodeURIComponent(encodedUrl));
		} catch (e) {
			console.warn('cannot decode url param', e);
		}
		return decodeUrl;
	}

	public getDecodedUriParam(paramName: string): string {
		if (this.isParamPresent(paramName)) return undefined;
		return decodeURI(this.routeParameter$.value[paramName]);
	}

	private isParamPresent(paramName: string): boolean {
		return isEmptyObject(this.routeParameter$.value) || isUndefined(this.routeParameter$.value[paramName]);
	}

	static isVorgangListPage(params: Params): boolean {
		return isEmptyObject(params);
	}

	static isVorgangDetailPage(params: Params, paramName: string): boolean {
		return Object.keys(params).length === 1 && !isUndefined(params[paramName]);
	}

	static isSearch(params: Params): boolean {
		return Object.keys(params).length === 1 && !isNil(params['search']);
	}

	public navigateRelativeTo(relativePath: string, relativeTo: ActivatedRoute): void {
		this.router.navigate([relativePath], { relativeTo });
	}
}