import { FormControl, FormGroup } from '@angular/forms';
import { createIssue } from '../../../test/error';
import { Issue } from '../tech.model';
import { getControlForIssue, getMessageForIssue, setValidationError } from './tech.validation.util';

describe('tech.validation.utils', () => {
	describe('setValidationError', () => {

		const baseField1Control = new FormControl();
		const baseField2Control = new FormControl();
		const subGroupFieldControl = new FormControl();

		const form = new FormGroup({
			baseField1: baseField1Control,
			baseField2: baseField2Control,
			subGroup: new FormGroup({
				subGroupField1: subGroupFieldControl
			})
		});

		describe('get control for issue', () => {
			it('should return base field control', () => {
				const issue: Issue = { ...createIssue(), field: 'baseField1' };

				const control = getControlForIssue(form, issue);

				expect(control).toBe(baseField1Control);
			});

			it('should reeturn sub group field', () => {
				const issue: Issue = { ...createIssue(), field: 'subGroup.subGroupField1' };

				const control = getControlForIssue(form, issue);

				expect(control).toBe(subGroupFieldControl);
			})

			it('should ignore path prefix', () => {
				const issue: Issue = { ...createIssue(), field: 'command.wiedervorlage.baseField1' };

				const control = getControlForIssue(form, issue, 'command.wiedervorlage');

				expect(control).toBe(baseField1Control);
			})
		})

		describe('in base field', () => {
			const issue: Issue = { ...createIssue(), field: 'baseField1' };
			it('should set error in control', () => {
				setValidationError(form, issue);

				expect(baseField1Control.errors).not.toBeNull();
			});

			it('should set message code in control', () => {
				setValidationError(form, issue);

				expect(baseField1Control.hasError(issue.messageCode)).toBe(true);
			});

			it('should set control touched', () => {
				setValidationError(form, issue);

				expect(baseField1Control.touched).toBe(true);
			})

			it('should not set error in other control', () => {
				setValidationError(form, issue);

				expect(baseField2Control.errors).toBeNull();
			});
		});

		describe('in subGroup Field', () => {
			const issue: Issue = { ...createIssue(), field: 'subGroup.subGroupField1' };

			it('should set error in control', () => {
				setValidationError(form, issue);

				expect(subGroupFieldControl.errors).not.toBeNull();
			});
		});
	});

	describe('get message for issue', () => {
		const fieldLabel = 'Field Label';
		it('should return message', () => {
			const msg = getMessageForIssue(fieldLabel, { ...createIssue(), messageCode: 'validation_field_size' });

			expect(msg).toContain('muss aus mindestens');
		});

		it('should set field label', () => {
			const msg = getMessageForIssue(fieldLabel, { ...createIssue(), messageCode: 'validation_field_size' });

			expect(msg).toContain(fieldLabel);
		});

		it('should replace min param', () => {
			const msg = getMessageForIssue(fieldLabel, { ...createIssue(), messageCode: 'validation_field_size', parameters: [{ name: 'min', value: '3' }] });

			expect(msg).toContain('3');
		});
	})
})
