import { Injectable } from '@angular/core';
import { Params } from '@angular/router';
import { ApiRootService } from '@goofy-client/api-root-shared';
import { createEmptyStateResource, createStateResource, doIfLoadingRequired, NavigationService, StateResource } from '@goofy-client/tech-shared';
import { VorgangService } from '@goofy-client/vorgang-shared';
import { getLink } from '@ngxp/rest';
import { Resource, ResourceUri } from '@ngxp/rest/lib/resource.model';
import { isNil } from 'lodash-es';
import { BehaviorSubject, Observable, Subscription } from 'rxjs';
import { filter, first, flatMap } from 'rxjs/operators';
import { UserProfileListResource, UserProfileResource } from './user-profile.model';
import { UserProfileRepository } from './user-profile.repository';

@Injectable({ providedIn: 'root' })
export class UserProfileService {

	private userProfiles = <any>{};
	private userProfileSearchList: BehaviorSubject<StateResource<UserProfileListResource>> = new BehaviorSubject(createEmptyStateResource());
	private userProfileSearchVisibility: BehaviorSubject<boolean> = new BehaviorSubject(false);

	private navigationSubscription: Subscription;

	private searchSubscription: Subscription;

	constructor(private repository: UserProfileRepository, private apiRootService: ApiRootService, private navigationService: NavigationService) {
		this.listenOnNavigation();
	}

	private listenOnNavigation(): void {
		if (!isNil(this.navigationSubscription)) this.navigationSubscription.unsubscribe();
		this.navigationSubscription = this.navigationService.urlChanged().subscribe(params => this.onNavigation(params));
	}

	onNavigation(params: Params): void {
		if (NavigationService.isVorgangListPage(params)) {
			this.userProfiles = <any>{};
			this.hideUserProfileSearch();
		} else if (NavigationService.isVorgangDetailPage(params, VorgangService.VORGANG_WITH_EINGANG_URL)) {
			this.userProfiles = <any>{};
		}
	}

	public isUserProfileSearchVisible(): Observable<boolean> {
		return this.userProfileSearchVisibility.asObservable();
	}

	public showUserProfileSearch(): void {
		this.userProfileSearchVisibility.next(true);
	}

	public hideUserProfileSearch(): void {
		this.userProfileSearchVisibility.next(false);
		this.userProfileSearchList.next(createEmptyStateResource());
	}

	public getAssignedUserProfile(vorgang: Resource, linkRel: string): Observable<StateResource<UserProfileResource>> {
		const uri: ResourceUri = this.getUserProfileUri(vorgang, linkRel);

		this.createIfNotExist(uri);

		doIfLoadingRequired(this.userProfiles[uri].value, () => {
			this.userProfiles[uri].next({ ...this.userProfiles[uri].value, loading: true });
			this.loadUserProfile(vorgang, linkRel)
		});
		return this.userProfiles[uri].asObservable();
	}

	createIfNotExist(uri: ResourceUri): void {
		if (isNil(this.userProfiles[uri])) this.userProfiles[uri] = new BehaviorSubject(createEmptyStateResource<UserProfileResource>());
	}

	loadUserProfile(vorgang: Resource, linkRel: string): void {
		this.repository.getUserProfile(vorgang, linkRel).pipe(first()).subscribe((userProfile: UserProfileResource) => this.updateUserProfile(vorgang, linkRel, userProfile));
	}

	updateUserProfile(vorgang: Resource, linkRel: string, userProfile: UserProfileResource): void {
		this.userProfiles[this.getUserProfileUri(vorgang, linkRel)].next(createStateResource(userProfile));
	}

	private getUserProfileUri(vorgang: Resource, linkRel: string): ResourceUri {
		return getLink(vorgang, linkRel).href;
	}

	public getSearchedUserProfiles(): Observable<StateResource<UserProfileListResource>> {
		return this.userProfileSearchList.asObservable();
	}

	public search(searchBy: string): Observable<StateResource<UserProfileListResource>> {
		if (!isNil(this.searchSubscription)) this.searchSubscription.unsubscribe();
		if (isNil(searchBy)) {
			this.clearSearchList();
		} else {
			this.searchSubscription = this.apiRootService.getApiRoot().pipe(
				filter(resource => resource.loaded),
				flatMap(apiRootResource => this.repository.search(apiRootResource.resource, searchBy)
				)).subscribe(userProfileList => {
					if (userProfileList !== null) {
						this.userProfileSearchList.next(createStateResource(userProfileList))
					}
				});
		}
		return this.userProfileSearchList.asObservable();
	}

	public clearSearchList() {
		this.userProfileSearchList.next(createEmptyStateResource());
	}
}