import { CommandResource, CommandService } from '@goofy-client/command-shared';
import { createEmptyStateResource, createStateResource, NavigationService, StateResource } from '@goofy-client/tech-shared';
import { Mock, mock, useFromMock } from '@goofy-client/test-utils';
import { SnackBarService } from '@goofy-client/ui';
import { hot } from 'jest-marbles';
import { CommandLinkRel } from 'libs/command-shared/src/lib/command.linkrel';
import { createCommandResource } from 'libs/command-shared/test/command';
import { createVorgangWithEingangResource } from 'libs/vorgang-shared/test/vorgang';
import { of } from 'rxjs';
import { VorgangCommandService } from './vorgang-command.service';
import { VorgangListService } from './vorgang-list.service';
import { VorgangWithEingangLinkRel } from './vorgang.linkrel';
import { VorgangMessages } from './vorgang.messages';
import { VorgangOrder, VorgangWithEingangResource } from './vorgang.model';
import { VorgangService } from './vorgang.service';
import { createAbschliessenCommand, createAnnehmenCommand, createBearbeitenCommand, createBescheidenCommand, createVerwerfenCommand, createWiedereroeffnenCommand, createZurueckholenCommand, createZurueckstellenCommand } from './vorgang.util';

describe('VorgangCommandService', () => {
	let service: VorgangCommandService;
	let navigationService: Mock<NavigationService>;
	let commandService: Mock<CommandService>;
	let snackBarService: Mock<SnackBarService>;
	let vorgangListService: Mock<VorgangListService>;
	let vorgangService: Mock<VorgangService>;

	beforeEach(() => {
		navigationService = { ...mock(NavigationService) };
		commandService = mock(CommandService);
		snackBarService = mock(SnackBarService);
		vorgangListService = mock(VorgangListService);
		vorgangService = mock(VorgangService);
		navigationService.urlChanged = jest.fn();
		navigationService.urlChanged.mockReturnValue(of({}));

		service = new VorgangCommandService(useFromMock(navigationService), useFromMock(commandService), useFromMock(snackBarService), useFromMock(vorgangListService), useFromMock(vorgangService));
	})

	describe('onNavigation', () => {

		describe('to vorgang list page', () => {

			const paramsForVorgangList = {};

			describe('should reset stateResource for order', () => {

				it('abschliessen', () => testOrderOnNavigation(VorgangOrder.ABSCHLIESSEN));

				it('annehmen', () => testOrderOnNavigation(VorgangOrder.ANNEHMEN));

				it('bearbeiten', () => testOrderOnNavigation(VorgangOrder.BEARBEITEN));

				it('bescheiden', () => testOrderOnNavigation(VorgangOrder.BESCHEIDEN));

				it('verwerfen', () => testOrderOnNavigation(VorgangOrder.VERWERFEN));

				it('wiedereroeffnen', () => testOrderOnNavigation(VorgangOrder.WIEDEREROEFFNEN));

				it('zurueckholen', () => testOrderOnNavigation(VorgangOrder.ZURUECKHOLEN));

				it('zurueckstellen', () => testOrderOnNavigation(VorgangOrder.ZURUECKSTELLEN));

				function testOrderOnNavigation(order: VorgangOrder) {
					service.commandStateResourceByOrder[order].next(createStateResource(createCommandResource()));

					service.onNavigation(paramsForVorgangList);

					expect(service.commandStateResourceByOrder[order].value).toEqual(createEmptyStateResource());
				}
			})
		});
	})

	describe('order actions', () => {

		const vorgangResource: VorgangWithEingangResource = createVorgangWithEingangResource();
		const commandResource: CommandResource = createCommandResource();
		const commandWithApiError: StateResource<CommandResource> = createStateResource(commandResource);

		beforeEach(() => {
			commandService.createCommand.mockReturnValue(hot('-a', { a: commandWithApiError }));
			service.showSnackbar = jest.fn();
			vorgangService.getVorgang.mockReturnValue(vorgangResource);

			service.doVorgangOrderAction = jest.fn();
		})

		it('should call commandService on vorgang annehmen', () => {
			service.annehmen();

			expect(service.doVorgangOrderAction).toHaveBeenCalledWith(VorgangWithEingangLinkRel.ANNEHMEN, createAnnehmenCommand, VorgangMessages.ANGENOMMEN, VorgangOrder.ANNEHMEN);
		})

		it('should call commandService on vorgang verwerfen', () => {
			service.verwerfen();

			expect(service.doVorgangOrderAction).toHaveBeenCalledWith(VorgangWithEingangLinkRel.VERWERFEN, createVerwerfenCommand, VorgangMessages.VERWORFEN, VorgangOrder.VERWERFEN);
		})

		it('should call commandService on vorgang zurueckholen', () => {
			service.zurueckholen();

			expect(service.doVorgangOrderAction).toHaveBeenCalledWith(VorgangWithEingangLinkRel.ZURUECKHOLEN, createZurueckholenCommand, VorgangMessages.ZURUECKGEHOLT, VorgangOrder.ZURUECKHOLEN);
		})

		it('should call commandService on vorgang bearbeiten', () => {
			service.bearbeiten();

			expect(service.doVorgangOrderAction).toHaveBeenCalledWith(VorgangWithEingangLinkRel.BEARBEITEN, createBearbeitenCommand, VorgangMessages.BEARBEITET, VorgangOrder.BEARBEITEN);
		})

		it('should call commandService on vorgang bescheiden', () => {
			service.bescheiden();

			expect(service.doVorgangOrderAction).toHaveBeenCalledWith(VorgangWithEingangLinkRel.BESCHEIDEN, createBescheidenCommand, VorgangMessages.BESCHIEDEN, VorgangOrder.BESCHEIDEN);
		})

		it('should call commandService on vorgang zurueckstellen', () => {
			service.zurueckstellen();

			expect(service.doVorgangOrderAction).toHaveBeenCalledWith(VorgangWithEingangLinkRel.ZURUECKSTELLEN, createZurueckstellenCommand, VorgangMessages.ZURUECKGESTELLT, VorgangOrder.ZURUECKSTELLEN);
		})

		it('should call commandService on vorgang abschliessen', () => {
			service.abschliessen();

			expect(service.doVorgangOrderAction).toHaveBeenCalledWith(VorgangWithEingangLinkRel.ABSCHLIESSEN, createAbschliessenCommand, VorgangMessages.ABGESCHLOSSEN, VorgangOrder.ABSCHLIESSEN);
		})

		it('should call commandService on vorgang wiedereroeffnen', () => {
			service.wiedereroeffnen();

			expect(service.doVorgangOrderAction).toHaveBeenCalledWith(VorgangWithEingangLinkRel.WIEDEREROEFFNEN, createWiedereroeffnenCommand, VorgangMessages.WIEDEREROEFFNET, VorgangOrder.WIEDEREROEFFNEN);
		})
	})

	describe('snack bar', () => {

		const snackBarMessage: string = 'Lorem ipsum pardeu ikromale'
		const commandResourceWithResourceLink: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);
		const commandResourceWithUpdateLink: CommandResource = createCommandResource([CommandLinkRel.UPDATE]);

		describe('when command is done', () => {

			it('should call service if command is done', () => {
				service.showSnackbar(createStateResource(commandResourceWithResourceLink), snackBarMessage);

				expect(snackBarService.showSnackbar).toHaveBeenCalled();
			})

			it('should not call service if command is pending', () => {
				service.showSnackbar(createStateResource(commandResourceWithUpdateLink), snackBarMessage);

				expect(snackBarService.showSnackbar).not.toHaveBeenCalled();
			})
		})
	})

	describe('proceedWithRevokeCommand', () => {

		it('should proceed if command is done', () => {
			service.proceedWithRevokeCommand(createStateResource(createCommandResource([CommandLinkRel.EFFECTED_RESOURCE])));

			expect(vorgangService.reloadVorgang).toHaveBeenCalled();
		})

		it('should reload vorgang list if command is done', () => {
			service.proceedWithRevokeCommand(createStateResource(createCommandResource([CommandLinkRel.EFFECTED_RESOURCE])));

			expect(vorgangListService.reloadVorgangList).toHaveBeenCalled();
		})

		it('should not proceed if command is pending', () => {
			service.proceedWithRevokeCommand(createStateResource(createCommandResource()));

			expect(vorgangService.reloadVorgang).not.toHaveBeenCalled();
			expect(vorgangListService.reloadVorgangList).not.toHaveBeenCalled();
		});
	})
})