import { Injectable } from '@angular/core';
import { Params } from '@angular/router';
import { CommandResource, CommandService, CreateCommand, isDone } from '@goofy-client/command-shared';
import { createEmptyStateResource, NavigationService, StateResource } from '@goofy-client/tech-shared';
import { SnackBarService } from '@goofy-client/ui';
import { isNil } from 'lodash-es';
import { BehaviorSubject, Observable, Subscription } from 'rxjs';
import { VorgangListService } from './vorgang-list.service';
import { VorgangWithEingangLinkRel } from './vorgang.linkrel';
import { VorgangMessages } from './vorgang.messages';
import { VorgangOrder } from './vorgang.model';
import { VorgangService } from './vorgang.service';
import { createAbschliessenCommand, createAnnehmenCommand, createBearbeitenCommand, createBescheidenCommand, createVerwerfenCommand, createWiedereroeffnenCommand, createZurueckholenCommand, createZurueckstellenCommand } from './vorgang.util';

@Injectable({ providedIn: 'root' })
export class VorgangCommandService {

	readonly commandStateResourceByOrder: CommandStateResourceByOrderMap = {
		[VorgangOrder.ABSCHLIESSEN]: new BehaviorSubject(createEmptyStateResource()),
		[VorgangOrder.ANNEHMEN]: new BehaviorSubject(createEmptyStateResource()),
		[VorgangOrder.BEARBEITEN]: new BehaviorSubject(createEmptyStateResource()),
		[VorgangOrder.BESCHEIDEN]: new BehaviorSubject(createEmptyStateResource()),
		[VorgangOrder.VERWERFEN]: new BehaviorSubject(createEmptyStateResource()),
		[VorgangOrder.WIEDEREROEFFNEN]: new BehaviorSubject(createEmptyStateResource()),
		[VorgangOrder.ZURUECKHOLEN]: new BehaviorSubject(createEmptyStateResource()),
		[VorgangOrder.ZURUECKSTELLEN]: new BehaviorSubject(createEmptyStateResource())
	};

	private readonly revokeCommand$: BehaviorSubject<StateResource<CommandResource>> = new BehaviorSubject(createEmptyStateResource());

	private subscription: Subscription;

	constructor(
		private navigationService: NavigationService,
		private commandService: CommandService,
		private snackBarService: SnackBarService,
		private vorgangListService: VorgangListService,
		private vorgangService: VorgangService
	) {
		this.listenForLeavingVorgang();
	}

	public getCommandStateResources(order: VorgangOrder): Observable<StateResource<CommandResource>> {
		return this.commandStateResourceByOrder[order].asObservable();
	}

	public getRevokeCommand(): Observable<StateResource<CommandResource>> {
		return this.revokeCommand$;
	}

	private listenForLeavingVorgang(): void {
		this.unsubscribe();
		this.subscription = this.navigationService.urlChanged().subscribe(params => this.onNavigation(params));
	}

	onNavigation(params: Params): void {
		if (NavigationService.isVorgangListPage(params)) {
			this.clearCommandStateResourceByOrder();
		}
	}

	private clearCommandStateResourceByOrder(): void {
		this.commandStateResourceByOrder[VorgangOrder.ABSCHLIESSEN].next(createEmptyStateResource());
		this.commandStateResourceByOrder[VorgangOrder.ANNEHMEN].next(createEmptyStateResource());
		this.commandStateResourceByOrder[VorgangOrder.BEARBEITEN].next(createEmptyStateResource());
		this.commandStateResourceByOrder[VorgangOrder.BESCHEIDEN].next(createEmptyStateResource());
		this.commandStateResourceByOrder[VorgangOrder.VERWERFEN].next(createEmptyStateResource());
		this.commandStateResourceByOrder[VorgangOrder.WIEDEREROEFFNEN].next(createEmptyStateResource());
		this.commandStateResourceByOrder[VorgangOrder.ZURUECKHOLEN].next(createEmptyStateResource());
		this.commandStateResourceByOrder[VorgangOrder.ZURUECKSTELLEN].next(createEmptyStateResource());
	}

	public abschliessen(): void {
		this.doVorgangOrderAction(VorgangWithEingangLinkRel.ABSCHLIESSEN, createAbschliessenCommand, VorgangMessages.ABGESCHLOSSEN, VorgangOrder.ABSCHLIESSEN);
	}

	public annehmen(): void {
		this.doVorgangOrderAction(VorgangWithEingangLinkRel.ANNEHMEN, createAnnehmenCommand, VorgangMessages.ANGENOMMEN, VorgangOrder.ANNEHMEN);
	}

	public bearbeiten(): void {
		this.doVorgangOrderAction(VorgangWithEingangLinkRel.BEARBEITEN, createBearbeitenCommand, VorgangMessages.BEARBEITET, VorgangOrder.BEARBEITEN);
	}

	public bescheiden(): void {
		this.doVorgangOrderAction(VorgangWithEingangLinkRel.BESCHEIDEN, createBescheidenCommand, VorgangMessages.BESCHIEDEN, VorgangOrder.BESCHEIDEN);
	}

	public verwerfen(): void {
		this.doVorgangOrderAction(VorgangWithEingangLinkRel.VERWERFEN, createVerwerfenCommand, VorgangMessages.VERWORFEN, VorgangOrder.VERWERFEN);
	}

	public wiedereroeffnen(): void {
		this.doVorgangOrderAction(VorgangWithEingangLinkRel.WIEDEREROEFFNEN, createWiedereroeffnenCommand, VorgangMessages.WIEDEREROEFFNET, VorgangOrder.WIEDEREROEFFNEN);
	}

	public zurueckholen(): void {
		this.doVorgangOrderAction(VorgangWithEingangLinkRel.ZURUECKHOLEN, createZurueckholenCommand, VorgangMessages.ZURUECKGEHOLT, VorgangOrder.ZURUECKHOLEN);
	}

	public zurueckstellen(): void {
		this.doVorgangOrderAction(VorgangWithEingangLinkRel.ZURUECKSTELLEN, createZurueckstellenCommand, VorgangMessages.ZURUECKGESTELLT, VorgangOrder.ZURUECKSTELLEN);
	}

	doVorgangOrderAction(linkRel: string, commandProvider: () => CreateCommand, snackBarMessage: string, order: VorgangOrder): void {
		this.setOrderOnLoading(order);

		const subscription: Subscription = this.commandService.createCommand(this.vorgangService.getVorgang(), linkRel, commandProvider())
			.subscribe((commandStateResource: StateResource<CommandResource>) => {
				if (isDone(commandStateResource.resource)) {
					this.vorgangService.reloadVorgang(commandStateResource.resource);
					this.showSnackbar(commandStateResource, snackBarMessage);
					this.commandStateResourceByOrder[order].next(commandStateResource);
					subscription.unsubscribe();
				}
			});
	}

	setOrderOnLoading(order: VorgangOrder) {
		this.commandStateResourceByOrder[order].next({ ...this.commandStateResourceByOrder[order].value, loading: true });
	}

	showSnackbar(commandResource: StateResource<CommandResource>, snackBarMessage: string): void {
		if (isDone(commandResource.resource)) {
			this.snackBarService.showSnackbar(commandResource.resource, snackBarMessage, () => this.handleRevokeCommand(commandResource.resource));
		}
	}

	private handleRevokeCommand(commandResource: CommandResource): void {
		this.revokeCommand$.next(createEmptyStateResource(true));

		const subscription: Subscription = this.commandService.revokeCommand(commandResource).subscribe(command => {
			this.proceedWithRevokeCommand(command);
			subscription.unsubscribe();
		});
	}

	proceedWithRevokeCommand(commandStateResource: StateResource<CommandResource>) {
		if (isDone(commandStateResource.resource)) {
			this.revokeCommand$.next(commandStateResource);
			this.vorgangService.reloadVorgang(commandStateResource.resource);
			this.vorgangListService.reloadVorgangList();
		}
	}

	ngOnDestroy(): void {
		this.unsubscribe();
	}

	private unsubscribe(): void {
		if (!isNil(this.subscription)) this.subscription.unsubscribe();
	}
}

interface CommandStateResourceByOrderMap {
	[VorgangOrder.ABSCHLIESSEN]: BehaviorSubject<StateResource<CommandResource>>,
	[VorgangOrder.ANNEHMEN]: BehaviorSubject<StateResource<CommandResource>>,
	[VorgangOrder.BEARBEITEN]: BehaviorSubject<StateResource<CommandResource>>,
	[VorgangOrder.BESCHEIDEN]: BehaviorSubject<StateResource<CommandResource>>,
	[VorgangOrder.VERWERFEN]: BehaviorSubject<StateResource<CommandResource>>,
	[VorgangOrder.WIEDEREROEFFNEN]: BehaviorSubject<StateResource<CommandResource>>,
	[VorgangOrder.ZURUECKHOLEN]: BehaviorSubject<StateResource<CommandResource>>,
	[VorgangOrder.ZURUECKSTELLEN]: BehaviorSubject<StateResource<CommandResource>>
}
