import { ApiRootResource } from '@goofy-client/api-root-shared';
import { createEmptyStateResource, createStateResource, NavigationService } from '@goofy-client/tech-shared';
import { Mock, mock, useFromMock } from '@goofy-client/test-utils';
import { getEmbeddedResource } from '@ngxp/rest';
import { cold } from 'jest-marbles';
import { ApiRootService } from 'libs/api-root-shared/src/lib/api-root.service';
import { createApiRootResource } from 'libs/api-root-shared/test/api-root';
import { createVorgangListResource, createVorgangResource } from 'libs/vorgang-shared/test/vorgang';
import { of } from 'rxjs';
import { VorgangListService } from './vorgang-list.service';
import { VorgangListLinkRel } from './vorgang.linkrel';
import { VorgangListResource, VorgangResource } from './vorgang.model';
import { VorgangRepository } from './vorgang.repository';

describe('VorgangListService', () => {
	let service;
	let navigationService: Mock<NavigationService>;
	let repository: Mock<VorgangRepository>;
	let apiRootService: Mock<ApiRootService>;

	const vorgangListResource: VorgangListResource = createVorgangListResource();
	const vorgaengeOfVorgangList: VorgangResource[] = getEmbeddedResource(vorgangListResource, VorgangListLinkRel.VORGANG_HEADER_LIST);

	beforeEach(() => {
		repository = mock(VorgangRepository);
		apiRootService = mock(ApiRootService);
		navigationService = { ...mock(NavigationService) };
		navigationService.urlChanged = jest.fn();
		navigationService.urlChanged.mockReturnValue(of({}));

		service = new VorgangListService(useFromMock(apiRootService), useFromMock(repository), useFromMock(navigationService));
	})

	describe('getVorgangList', () => {
		const apiRootResource: ApiRootResource = createApiRootResource();

		beforeEach(() => {
			apiRootService.getApiRoot.mockReturnValue(of(createStateResource(apiRootResource)));
		})

		it('should call repository if not loaded', () => {
			service.vorgangList$.next(of(null));

			service.getVorgangList();

			expect(repository.loadVorgangList).toHaveBeenCalledWith(apiRootResource);
		})

		it('should not call repository if already loading', () => {
			service.vorgangList$.next(createEmptyStateResource(true));

			service.getVorgangList();

			expect(repository.loadVorgangList).not.toHaveBeenCalled();
		})

		describe('set loading', () => {

			beforeEach(() => {
				repository.loadVorgangList.mockReturnValue(of(null));
			})

			it('should set loading flag', () => {
				service.getVorgangList();

				expect(service.vorgangList$.value.loading).toBe(true);
			});
		})

		describe('check vorgangList after load ', () => {

			beforeEach(() => {
				repository.loadVorgangList.mockReturnValue(of(vorgangListResource));
			})

			it('should set loaded', () => {
				service.getVorgangList();

				expect(service.vorgangList$.value.loaded).toBe(true);
			});

			it('should set resource', () => {
				service.getVorgangList();

				expect(service.vorgangList$.value.resource).toBe(vorgangListResource);
			});

			it('should unset loading', () => {
				service.getVorgangList();

				expect(service.vorgangList$.value.loading).toBe(false);
			});
		})

		describe('check vorgaenge after load', () => {

			beforeEach(() => {
				repository.loadVorgangList.mockReturnValue(of(vorgangListResource));
			})

			it('should set vorgaenge to loaded vorgaenge', () => {
				service.getVorgangList();

				expect(service.vorgaenge$.value).toHaveLength(vorgaengeOfVorgangList.length);
				expect(service.vorgaenge$.value).toBe(vorgaengeOfVorgangList);
			})
		})

		describe('check hasNextPage after load', () => {

			it('should be true if link exists', () => {
				repository.loadVorgangList.mockReturnValue(of(createVorgangListResource([VorgangListLinkRel.NEXT])));

				service.getVorgangList();

				expect(service.hasNextPage()).toBeObservable(cold('a', { a: true }));
			})

			it('should be false if link not exists', () => {
				repository.loadVorgangList.mockReturnValue(of(createVorgangListResource()));

				service.getVorgangList();

				expect(service.hasNextPage()).toBeObservable(cold('a', { a: false }));
			})
		})

		describe('load vorgangList with empty ApiRoot resource', () => {

			beforeEach(() => {
				apiRootService.getApiRoot.mockReturnValue(of(createEmptyStateResource()));
			})

			it('should not call repository loadVorgangList', () => {
				service.getVorgangList();

				expect(repository.loadVorgangList).not.toHaveBeenCalled();
			})
		})
	})

	describe('load next page', () => {
		const vorgangListResource: VorgangListResource = createVorgangListResource();

		beforeEach(() => {
			repository.getNextVorgangListPage.mockReturnValue(of(vorgangListResource));
		})

		it('should not call repository while vorgangList is loading', () => {
			service.vorgangList$.next(createEmptyStateResource(true));

			service.loadNextPage();

			expect(repository.getNextVorgangListPage).not.toHaveBeenCalled();
		})

		it('should call repository', () => {
			service.vorgangList$.next(createStateResource(vorgangListResource));

			service.loadNextPage();

			expect(repository.getNextVorgangListPage).toHaveBeenCalledWith(vorgangListResource);
		})

		describe('set loading', () => {

			beforeEach(() => {
				repository.getNextVorgangListPage.mockReturnValue(of(null));
			})

			it('should set loading flag', () => {
				service.loadNextPage();

				expect(service.vorgangList$.value.loading).toBe(true);
			});
		})

		describe('check vorgangList after load', () => {

			beforeEach(() => {
				repository.getNextVorgangListPage.mockReturnValue(of(vorgangListResource));
			})

			it('should set loaded', () => {
				service.loadNextPage();

				expect(service.vorgangList$.value.loaded).toBe(true);
			});

			it('should set resource', () => {
				service.loadNextPage();

				expect(service.vorgangList$.value.resource).toBe(vorgangListResource);
			});

			it('should unset loading', () => {
				service.loadNextPage();

				expect(service.vorgangList$.value.loading).toBe(false);
			});
		})

		describe('check hasNextPage after load', () => {

			it('should be true if link exists', () => {
				repository.getNextVorgangListPage.mockReturnValue(of(createVorgangListResource([VorgangListLinkRel.NEXT])));

				service.loadNextPage();

				expect(service.hasNextPage$.value).toBe(true);
			})

			it('should be false if link not exists', () => {
				repository.getNextVorgangListPage.mockReturnValue(of(createVorgangListResource()));

				service.loadNextPage();

				expect(service.hasNextPage$.value).toBe(false);
			})
		})
	});

	describe('onNavigation', () => {

		describe('to vorgang list page', () => {

			beforeEach(() => {
				service.setSearchString = jest.fn();
			})

			it('should clear searchString', () => {
				service.onNavigation({});

				expect(service.setSearchString).toHaveBeenCalledWith(null);
			})
		});

		describe('search', () => {

			beforeEach(() => {
				service.clearVorgangList = jest.fn();
				service.isVorgangListLoaded = jest.fn();
				service.isVorgangListLoaded.mockReturnValue(true);
				service.setSearchString = jest.fn();
				service.getVorgangList = jest.fn();
			})

			it('should clear vorgang list', () => {
				service.onNavigation({ search: 'X' });

				expect(service.clearVorgangList).toHaveBeenCalled();
			})

			it('should load vorgang list', () => {
				service.onNavigation({ search: 'X' });

				expect(service.getVorgangList).toHaveBeenCalled();
			})

			it('should set search sting', () => {
				const searchString: string = 'X';
				navigationService.getDecodedUriParam.mockReturnValue(searchString);

				service.onNavigation({ search: searchString });

				expect(service.setSearchString).toHaveBeenCalledWith(searchString);
			});
		});

		describe('to vorgang detail page', () => {

			beforeEach(() => {
				service.setSearchString = jest.fn();
			})

			it('should clear searchString', () => {
				service.onNavigation({ 'vorgangWithEingangUrl': 'X' });

				expect(service.setSearchString).toHaveBeenCalledWith(null);
			})
		});
	})

	describe('addVorgangListToVorgaenge', () => {

		const vorgangResource: VorgangResource = createVorgangResource();
		const additionalVorgangResource: VorgangResource = createVorgangResource();

		beforeEach(() => {
			service.vorgaenge$.next([vorgangResource]);
		})

		it('should add vorgaenge to existing vorgaenge', () => {
			const expectedVorgaenge: VorgangResource[] = [vorgangResource, additionalVorgangResource];

			service.addVorgangListToVorgaenge([additionalVorgangResource]);

			expect(service.vorgaenge$.value).toHaveLength(2);
			expect(service.vorgaenge$.value).toEqual(expectedVorgaenge);
		})
	})

	describe('loadVorgangList', () => {

		const apiRootResource: ApiRootResource = createApiRootResource();

		beforeEach(() => {
			service.getSearchByCriteria = jest.fn();
		})

		it('should call repository searchVorgaengeBy on existing search by criteria', () => {
			const searchValue: string = 'searchParam';
			repository.searchVorgaengeBy.mockReturnValue(of(vorgangListResource));

			service.setSearchString(searchValue);
			service.loadVorgangList(apiRootResource);

			expect(repository.searchVorgaengeBy).toHaveBeenCalledWith(apiRootResource, searchValue);
		})

		it('should call repository loadVorgangList on null search criteria', () => {
			service.getSearchByCriteria.mockReturnValue(null);
			repository.loadVorgangList.mockReturnValue(of(vorgangListResource));

			service.loadVorgangList(apiRootResource);

			expect(repository.loadVorgangList).toHaveBeenCalledWith(apiRootResource);
		})
	})
})
