import { Injectable } from '@angular/core';
import { Params } from '@angular/router';
import { ApiRootResource } from '@goofy-client/api-root-shared';
import { createEmptyStateResource, createStateResource, doIfLoadingRequired, NavigationService, StateResource } from '@goofy-client/tech-shared';
import { hasLink } from '@ngxp/rest';
import { ApiRootService } from 'libs/api-root-shared/src/lib/api-root.service';
import { isNil, isNull } from 'lodash-es';
import { BehaviorSubject, Observable, Subscription } from 'rxjs';
import { filter, first, flatMap } from 'rxjs/operators';
import { VorgangListLinkRel } from './vorgang.linkrel';
import { VorgangListResource, VorgangResource } from './vorgang.model';
import { VorgangRepository } from './vorgang.repository';
import { VorgangService } from './vorgang.service';
import { getVorgaengeFromList } from './vorgang.util';

@Injectable({ providedIn: 'root' })
export class VorgangListService {

	public static readonly SEARCH: string = 'search';

	private readonly vorgangList$: BehaviorSubject<StateResource<VorgangListResource>> = new BehaviorSubject(createEmptyStateResource());
	private readonly vorgaenge$: BehaviorSubject<VorgangResource[]> = new BehaviorSubject([]);
	private readonly hasNextPage$: BehaviorSubject<boolean> = new BehaviorSubject(false);

	private readonly searchString$: BehaviorSubject<string> = new BehaviorSubject(null);

	private subscription: Subscription;

	constructor(
		private apiRootService: ApiRootService,
		private vorgangRepository: VorgangRepository,
		private navigationService: NavigationService
	) {
		this.listenOnNavigation();
	}

	private listenOnNavigation(): void {
		this.unsubscribe();
		this.subscription = this.navigationService.urlChanged().subscribe(params => this.onNavigation(params));
	}

	onNavigation(params: Params): void {
		if (NavigationService.isVorgangListPage(params)) {
			this.setSearchString(null);
			this.reloadIfNeccessary();
		}
		if (NavigationService.isSearch(params)) {
			this.setSearchString(this.navigationService.getDecodedUriParam(VorgangListService.SEARCH));
			this.reloadIfNeccessary();
		}
		if (NavigationService.isVorgangDetailPage(params, VorgangService.VORGANG_WITH_EINGANG_URL)) {
			this.setSearchString(null);
		}
	}

	setSearchString(value: string): void {
		this.searchString$.next(value);
	}

	reloadIfNeccessary() {
		if (this.isVorgangListLoaded()) {
			this.reloadVorgangList();
		}
	}

	isVorgangListLoaded(): boolean {
		return this.vorgangList$.value.loaded;
	}

	public getVorgaenge(): Observable<VorgangResource[]> {
		return this.vorgaenge$;
	}

	public hasNextPage(): Observable<boolean> {
		return this.hasNextPage$;
	}

	clearVorgangList(): void {
		this.vorgangList$.next({ ...this.vorgangList$.value, reload: true });
	}

	public reloadVorgangList(): void {
		this.clearVorgangList();
		this.getVorgangList();
	}

	public getVorgangList(): Observable<StateResource<VorgangListResource>> {
		doIfLoadingRequired(this.vorgangList$.value, () => this.apiRootService.getApiRoot().pipe(
			filter(resource => resource.loaded),
			flatMap(apiRootResource => this.loadVorgangList(apiRootResource.resource)
			)).subscribe(vorgangList => {
				if (vorgangList !== null) {
					this.updateVorgangList(vorgangList);
					this.updateVorgaenge(getVorgaengeFromList(vorgangList));
				}
			}));

		return this.vorgangList$;
	}

	loadVorgangList(apiRootResource: ApiRootResource): Observable<VorgangListResource> {
		this.setVorgangListLoading();

		if (!isNull(this.searchString$.value)) {
			return this.vorgangRepository.searchVorgaengeBy(apiRootResource, this.searchString$.value);
		}
		return this.vorgangRepository.loadVorgangList(apiRootResource);
	}

	public loadNextPage(): void {
		if (!this.isVorgangListLoading()) {
			this.loadNextListPage();
		}
	}

	private isVorgangListLoading(): boolean {
		return this.vorgangList$.value.loading;
	}

	private loadNextListPage(): void {
		this.setVorgangListLoading();

		this.vorgangRepository.getNextVorgangListPage(this.vorgangList$.value.resource).pipe(first())
			.subscribe(vorgangList => {
				if (vorgangList !== null) {
					this.updateVorgangList(vorgangList);
					this.addVorgangListToVorgaenge(getVorgaengeFromList(vorgangList));
				}
			});
	}

	private setVorgangListLoading(): void {
		this.vorgangList$.next({ ...this.vorgangList$.value, loading: true })
	}

	private updateVorgangList(vorgangList: VorgangListResource): void {
		this.vorgangList$.next(createStateResource(vorgangList));
		this.updateHasNextPage();
	}

	private updateHasNextPage(): void {
		this.hasNextPage$.next(hasLink(this.vorgangList$.value.resource, VorgangListLinkRel.NEXT));
	}

	addVorgangListToVorgaenge(vorgaenge: VorgangResource[]): void {
		const updatedVorgaenge: VorgangResource[] = [...this.vorgaenge$.value].concat(vorgaenge);

		this.updateVorgaenge(updatedVorgaenge);
	}

	private updateVorgaenge(vorgaenge: VorgangResource[]): void {
		this.vorgaenge$.next(vorgaenge);
	}

	search(searchString: string): void {
		this.searchString$.next(searchString);
		this.reloadVorgangList();
	}

	getSearchString(): Observable<string> {
		return this.searchString$.asObservable();
	}

	ngOnDestroy(): void {
		this.unsubscribe();
	}

	private unsubscribe(): void {
		if (!isNil(this.subscription)) this.subscription.unsubscribe();
	}
}