import { ApiRootLinkRel, ApiRootResource } from '@goofy-client/api-root-shared';
import { mock, useFromMock } from '@goofy-client/test-utils';
import { getUrl, ResourceFactory } from '@ngxp/rest';
import { cold, hot } from 'jest-marbles';
import { createApiRootResource } from 'libs/api-root-shared/test/api-root';
import { createVorgangListResource, createVorgangResource, createVorgangWithEingangResource } from 'libs/vorgang-shared/test/vorgang';
import { VorgangHeaderLinkRel, VorgangListLinkRel, VorgangWithEingangLinkRel } from './vorgang.linkrel';
import { VorgangListResource, VorgangResource, VorgangWithEingangResource } from './vorgang.model';
import { VorgangRepository } from './vorgang.repository';

describe('VorgangRepository', () => {
	let repository: VorgangRepository;
	let resourceFactory = mock(ResourceFactory);
	let resourceWrapper = { get: jest.fn() };

	beforeEach(() => {
		repository = new VorgangRepository(useFromMock(resourceFactory));

		resourceFactory.fromId.mockReturnValue(resourceWrapper);
		resourceFactory.from.mockReturnValue(resourceWrapper);
	})

	describe('loadVorgangList', () => {
		const vorgangListResource: VorgangListResource = createVorgangListResource();
		const apiRootResource: ApiRootResource = createApiRootResource();

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(hot('a', { a: vorgangListResource }));
		})

		it('should call resourceFactory with resource', () => {
			repository.loadVorgangList(apiRootResource);

			expect(resourceFactory.from).toHaveBeenCalledWith(apiRootResource);
		})

		it('should call resourceWrapper with link', () => {
			repository.loadVorgangList(apiRootResource);

			expect(resourceWrapper.get).toHaveBeenCalledWith(ApiRootLinkRel.VORGAENGE)
		})

		it('should return result', () => {
			let result = repository.loadVorgangList(apiRootResource);

			expect(result).not.toBeNull();
			expect(result).toBeObservable(cold('a', { a: vorgangListResource }));
		})
	})

	describe('getNextVorgangListPage', () => {
		const vorgangListResource: VorgangListResource = createVorgangListResource([VorgangListLinkRel.NEXT]);

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(hot('a', { a: vorgangListResource }));
		})

		it('should call resourceFactory with resource', () => {
			repository.getNextVorgangListPage(vorgangListResource);

			expect(resourceFactory.from).toHaveBeenCalledWith(vorgangListResource);
		})

		it('should call resourceWrapper with link', () => {
			repository.getNextVorgangListPage(vorgangListResource);

			expect(resourceWrapper.get).toHaveBeenCalledWith(VorgangListLinkRel.NEXT)
		})

		it('should return result', () => {
			let result = repository.getNextVorgangListPage(vorgangListResource);

			expect(result).not.toBeNull();
			expect(result).toBeObservable(cold('a', { a: vorgangListResource }));
		})
	})

	describe('getVorgangWithEingang', () => {
		const vorgangResource: VorgangResource = createVorgangResource([VorgangHeaderLinkRel.VORGANG_WITH_EINGANG]);
		const vorgangWithEingangUrl = getUrl(vorgangResource, VorgangHeaderLinkRel.VORGANG_WITH_EINGANG);
		const vorgangWithEingangResource: VorgangWithEingangResource = createVorgangWithEingangResource();

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(hot('a', { a: vorgangWithEingangResource }));
		})

		it('should call resourceFactory with resource', () => {
			repository.getVorgang(vorgangWithEingangUrl);

			expect(resourceFactory.fromId).toHaveBeenCalledWith(vorgangWithEingangUrl);
		})

		it('should call resourceWrapper with link', () => {
			repository.getVorgang(vorgangWithEingangUrl);

			expect(resourceWrapper.get).toHaveBeenCalled()
		})

		it('should return result', () => {
			let result = repository.getVorgang(vorgangWithEingangUrl);

			expect(result).not.toBeNull();
			expect(result).toBeObservable(cold('a', { a: vorgangWithEingangResource }));
		})
	})

	describe('getAttachments', () => {

		const vorgangWithEingangResource: VorgangWithEingangResource = createVorgangWithEingangResource([VorgangWithEingangLinkRel.ATTACHMENTS]);

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(hot('a', { a: vorgangWithEingangResource }));
		})

		it('should call resourceFactory with resource', () => {
			repository.getAttachments(vorgangWithEingangResource);

			expect(resourceFactory.from).toHaveBeenCalledWith(vorgangWithEingangResource);
		})

		it('should call resourceWrapper with link', () => {
			repository.getAttachments(vorgangWithEingangResource);

			expect(resourceWrapper.get).toHaveBeenCalledWith(VorgangWithEingangLinkRel.ATTACHMENTS)
		})

		it('should return result', () => {
			let result = repository.getAttachments(vorgangWithEingangResource);

			expect(result).not.toBeNull();
			expect(result).toBeObservable(cold('a', { a: vorgangWithEingangResource }));
		})
	})

	describe('searchVorgaengeBy', () => {

		const vorgangListResource: VorgangListResource = createVorgangListResource();
		const apiRootResource: ApiRootResource = createApiRootResource();

		beforeEach(() => {
			resourceWrapper.get.mockReturnValue(hot('a', { a: vorgangListResource }));
			repository.buildSearchByUrl = jest.fn();
		})

		it('should call resourceFactory', () => {
			repository.searchVorgaengeBy(apiRootResource, null);

			expect(resourceFactory.fromId).toHaveBeenCalled();
		})

		it('should call resourceWrapper ', () => {
			repository.searchVorgaengeBy(apiRootResource, null);

			expect(resourceWrapper.get).toHaveBeenCalled();
		})

		it('should call buildSearchByUrl', () => {
			repository.searchVorgaengeBy(apiRootResource, null);

			expect(repository.buildSearchByUrl).toHaveBeenCalled()
		})
	})

	it('should build search by url', () => {
		const _links = {
			self: { href: '/api/1' }, search: { href: '/vorgangs?page=0{&' + VorgangRepository.SEARCH_PARAM + '}' }
		};
		const apiRootResource: ApiRootResource = { ...createApiRootResource([ApiRootLinkRel.SEARCH]), _links };

		const resultUri = repository.buildSearchByUrl(apiRootResource, 'searchParam');

		expect(resultUri).toEqual('/vorgangs?page=0&' + VorgangRepository.SEARCH_PARAM + '=searchParam');
	})
})