import { CommandListResource, CommandResource, CommandService } from '@goofy-client/command-shared';
import { createEmptyStateResource, createStateResource, NavigationService, StateResource } from '@goofy-client/tech-shared';
import { Mock, mock, useFromMock } from '@goofy-client/test-utils';
import { UserProfileResource, UserProfileService } from '@goofy-client/user-profile-shared';
import { getUrl, ResourceUri } from '@ngxp/rest';
import { cold, hot } from 'jest-marbles';
import { CommandLinkRel } from 'libs/command-shared/src/lib/command.linkrel';
import { createCommandListResource, createCommandResource } from 'libs/command-shared/test/command';
import { createOzgFileListResource } from 'libs/ozg-file-shared/test/ozg-file';
import { createVorgangWithEingangResource } from 'libs/vorgang-shared/test/vorgang';
import { of } from 'rxjs';
import { createUserProfileResource } from '../../../user-profile-shared/test/user-profile';
import { VorgangWithEingangLinkRel } from './vorgang.linkrel';
import { CreateAssignUserCommand, VorgangOrder, VorgangWithEingangResource } from './vorgang.model';
import { VorgangRepository } from './vorgang.repository';
import { VorgangService } from './vorgang.service';

describe('VorgangService', () => {
	let service;
	let navigationService: Mock<NavigationService>;
	let repository: Mock<VorgangRepository>;
	let commandService: Mock<CommandService>;
	let userProfileService: Mock<UserProfileService>

	beforeEach(() => {
		repository = mock(VorgangRepository);
		navigationService = { ...mock(NavigationService) };
		commandService = mock(CommandService);
		navigationService.urlChanged = jest.fn();
		navigationService.urlChanged.mockReturnValue(of({}));
		userProfileService = mock(UserProfileService);

		service = new VorgangService(useFromMock(repository), useFromMock(navigationService), useFromMock(commandService), useFromMock(userProfileService));
	})

	describe('get vorgang with eingang', () => {
		const vorgangWithEingang: VorgangWithEingangResource = createVorgangWithEingangResource();
		const url: ResourceUri = getUrl(vorgangWithEingang);

		beforeEach(() => {
			repository.getVorgang.mockReturnValue(of(vorgangWithEingang));
			service.setStateResourceOnLoading = jest.fn();
		})

		it('should set loading', () => {
			service.loadVorgangWithEingang(url);

			expect(service.setStateResourceOnLoading).toHaveBeenCalled();
		});

		it('should put vorgang in subject', () => {
			service.loadVorgangWithEingang(url);

			expect(service.vorgangWithEingang$.value).toEqual(createStateResource(vorgangWithEingang));
		});

	})

	describe('loadVorgangWithEingang', () => {
		const vorgangWithEingang = createVorgangWithEingangResource();
		const url = getUrl(vorgangWithEingang);

		beforeEach(() => {
			repository.getVorgang.mockReturnValue(cold('a', { a: vorgangWithEingang }));
		})

		it('should call repository', () => {
			service.loadVorgangWithEingang(url);

			expect(repository.getVorgang).toHaveBeenCalledWith(url);
		});
	})

	describe('updateVorgang', () => {

		beforeEach(() => {
			service.loadPendingCommandsByVorgang = jest.fn();
			service.loadAttachments = jest.fn();
			service.loadForwardings = jest.fn();
		})

		it('should load pendingCommands', () => {
			service.updateVorgang(createVorgangWithEingangResource());

			expect(service.loadPendingCommandsByVorgang).toHaveBeenCalled();
		})

		it('should load attachments', () => {
			service.updateVorgang(createVorgangWithEingangResource());

			expect(service.loadAttachments).toHaveBeenCalled();
		})
	})

	describe('loadPendingCommandsByVorgang', () => {
		const vorgangWithEingang = createVorgangWithEingangResource([VorgangWithEingangLinkRel.PENDING_COMMANDS]);

		beforeEach(() => {
			commandService.getPendingCommands = jest.fn();
			commandService.getPendingCommands.mockReturnValue(of(createCommandResource()));

			service.setStateResourceOnLoading = jest.fn();
			service.clearPendingCommands = jest.fn();
			service.clearPendingForwardCommand = jest.fn();
		})

		it('should do nothing without link', () => {
			service.loadPendingCommandsByVorgang(createVorgangWithEingangResource());

			expect(service.setStateResourceOnLoading).not.toHaveBeenCalled();
		})

		it('should call commandService', () => {
			service.loadPendingCommandsByVorgang(vorgangWithEingang);

			expect(commandService.getPendingCommands).toHaveBeenCalledWith(vorgangWithEingang, VorgangWithEingangLinkRel.PENDING_COMMANDS);
		});

		describe('setPendingForwardCommand', () => {

			beforeEach(() => {
				service.existsPendingForwardCommand = jest.fn();
				service.setPendingForwardSingleCommand = jest.fn();
			})

			it('should set pendingForwardCommand if its empty', () => {
				service.existsPendingForwardCommand.mockReturnValue(false);
				const command: CommandResource = { ...createCommandResource(), order: VorgangOrder.FORWARD };
				const commandList: CommandListResource = createCommandListResource([command]);

				service.setPendingForwardCommand(commandList);

				expect(service.setPendingForwardSingleCommand).toHaveBeenCalledWith(createStateResource(command));
			})

			it('should NOT set pendingForwardCommand if its exists', () => {
				service.existsPendingForwardCommand.mockReturnValue(true);

				service.setPendingForwardCommand(createCommandListResource());

				expect(service.setPendingForwardSingleCommand).not.toHaveBeenCalled();
			})
		})
	})

	describe('onNavigation', () => {

		describe('to vorgang list page', () => {

			beforeEach(() => {
				service.clearVorgang = jest.fn();
				service.clearPendingForwardCommand = jest.fn();
			})

			it('should clear clearPendingForwardCommand', () => {
				service.onNavigation({});

				expect(service.clearPendingForwardCommand).toHaveBeenCalled();
			})

			it('should clear vorgang', () => {
				service.onNavigation({});

				expect(service.clearVorgang).toHaveBeenCalled();
			});
		});
	})

	describe('loadAttachments', () => {

		beforeEach(() => {
			repository.getAttachments.mockReturnValue(of(createOzgFileListResource()));
			service.doRepositoryCall = jest.fn();
		})

		it('should call repository', () => {
			const vorgang: VorgangWithEingangResource = createVorgangWithEingangResource([VorgangWithEingangLinkRel.ATTACHMENTS]);

			service.loadAttachments(vorgang);

			expect(service.doRepositoryCall).toHaveBeenCalledWith(service.getAttachments(), repository.getAttachments(vorgang));
		})

		it('should NOT call repository', () => {
			const vorgang: VorgangWithEingangResource = createVorgangWithEingangResource();

			service.loadAttachments(vorgang);

			expect(service.doRepositoryCall).not.toHaveBeenCalled();
		})
	})

	describe('reload vorgang', () => {

		const commandResourceWithLink: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);

		beforeEach(() => {
			service.loadVorgangWithEingang = jest.fn();
		})

		it('should reload with given uri', () => {
			service.reloadVorgang(commandResourceWithLink);

			expect(service.loadVorgangWithEingang).toHaveBeenCalledWith(getUrl(commandResourceWithLink, CommandLinkRel.EFFECTED_RESOURCE));
		})
	})

	describe('reloadVorgangOnDone', () => {

		const snackBarMessage: string = 'Lorem ipsum pardeu ikromale'
		const commandResourceWithResourceLink: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);
		const commandResourceWithUpdateLink: CommandResource = createCommandResource([CommandLinkRel.UPDATE]);

		beforeEach(() => {
			service.loadVorgangWithEingang = jest.fn();
		})

		it('should call load vorgang on command is done', () => {
			service.reloadVorgangOnDone(createStateResource(commandResourceWithResourceLink), snackBarMessage);

			expect(service.loadVorgangWithEingang).toHaveBeenCalled();
		})

		it('should not load vorgang on command is pending', () => {
			service.reloadVorgangOnDone(createStateResource(commandResourceWithUpdateLink), snackBarMessage);

			expect(service.loadVorgangWithEingang).not.toHaveBeenCalled();
		})
	})

	describe('reloadCurrentVorgang', () => {

		beforeEach(() => {
			service.loadVorgangWithEingang = jest.fn();
		})

		it('should call navigation service', () => {
			service.reloadCurrentVorgang();

			expect(navigationService.getDecodedParam).toHaveBeenCalledWith(VorgangService.VORGANG_WITH_EINGANG_URL);
		})

		it('should call loadVorgangWithEingang', () => {
			service.reloadCurrentVorgang();

			expect(service.loadVorgangWithEingang).toHaveBeenCalled();
		})
	})

	describe('assign user', () => {

		const userProfile: UserProfileResource = createUserProfileResource();
		const userProfileUri: ResourceUri = getUrl(userProfile);

		const command: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);
		const commandStateResource: StateResource<CommandResource> = createStateResource(command);

		describe('init and next value', () => {

			beforeEach(() => {
				commandService.createCommand.mockReturnValue(hot('-a', { a: commandStateResource }));
				service.reloadVorgangOnDone = jest.fn();
			})

			it('should be return', () => {
				const result = service.assignUser(userProfileUri);

				expect(result).toBeObservable(cold('ab', { a: createEmptyStateResource(true), b: commandStateResource }));
			})
		})

		describe('further process', () => {

			beforeEach(() => {
				commandService.createCommand.mockReturnValue(of(commandStateResource));
			})

			it('should call command service', () => {
				const order: CreateAssignUserCommand = { order: VorgangOrder.ASSIGN_USER, body: { assignedTo: userProfileUri } };

				service.assignUser(userProfileUri);

				expect(commandService.createCommand).toHaveBeenCalledWith(service.vorgangWithEingang$.value.resource, VorgangWithEingangLinkRel.ASSIGN, order);
			})

			it('should reload vorgang on command is done', () => {
				service.reloadVorgang = jest.fn();

				service.assignUser(userProfileUri).subscribe();

				expect(service.reloadVorgang).toHaveBeenCalled();
			})
		})
	})
})
