import { Injectable } from '@angular/core';
import { Params } from '@angular/router';
import { CommandListResource, CommandResource, CommandService, isDone } from '@goofy-client/command-shared';
import { OzgFileListResource } from '@goofy-client/ozg-file-shared';
import { createEmptyStateResource, createStateResource, doIfLoadingRequired, NavigationService, StateResource } from '@goofy-client/tech-shared';
import { UserProfileService } from '@goofy-client/user-profile-shared';
import { getUrl, hasLink, Resource, ResourceUri } from '@ngxp/rest';
import { CommandLinkRel } from 'libs/command-shared/src/lib/command.linkrel';
import { isNil, isUndefined } from 'lodash-es';
import { BehaviorSubject, Observable, Subscription } from 'rxjs';
import { first, startWith, tap } from 'rxjs/operators';
import { VorgangWithEingangLinkRel } from './vorgang.linkrel';
import { CreateAssignUserCommand, VorgangOrder, VorgangResource, VorgangWithEingangResource } from './vorgang.model';
import { VorgangRepository } from './vorgang.repository';
import { getPendingForwardingCommand } from './vorgang.util';

@Injectable({ providedIn: 'root' })
export class VorgangService {

	readonly vorgangWithEingang$: BehaviorSubject<StateResource<VorgangWithEingangResource>> = new BehaviorSubject(createEmptyStateResource());

	private readonly pendingForwardCommand$: BehaviorSubject<StateResource<CommandResource>> = new BehaviorSubject(createEmptyStateResource());
	private readonly attachments$: BehaviorSubject<StateResource<OzgFileListResource>> = new BehaviorSubject(createEmptyStateResource());

	private subscription: Subscription;

	public static readonly VORGANG_WITH_EINGANG_URL: string = 'vorgangWithEingangUrl';

	constructor(
		private vorgangRepository: VorgangRepository,
		private navigationService: NavigationService,
		private commandService: CommandService,
		private userProfileService: UserProfileService
	) {
		this.listenForLeavingVorgang();
	}

	public getVorgangWithEingang(): Observable<StateResource<VorgangWithEingangResource>> {
		doIfLoadingRequired(this.vorgangWithEingang$.value, () => {
			const vorgangUri: ResourceUri = this.getVorgangWithEingangUrl();
			if (!isUndefined(vorgangUri)) this.loadVorgangWithEingang(vorgangUri);
		});
		return this.vorgangWithEingang$.asObservable();
	}

	private getVorgangWithEingangUrl(): ResourceUri {
		return this.navigationService.getDecodedParam(VorgangService.VORGANG_WITH_EINGANG_URL);
	}

	reloadCurrentVorgang(): void {
		this.loadVorgangWithEingang(this.navigationService.getDecodedParam(VorgangService.VORGANG_WITH_EINGANG_URL));
	}

	public loadVorgangWithEingang(vorgangUri: ResourceUri): void {
		this.setStateResourceOnLoading(this.vorgangWithEingang$);
		this.vorgangRepository.getVorgang(vorgangUri).subscribe(vorgang => this.updateVorgang(vorgang))
	}

	updateVorgang(vorgang: VorgangWithEingangResource): void {
		this.vorgangWithEingang$.next(createStateResource(vorgang));
		this.loadPendingCommandsByVorgang(vorgang);
		this.loadAttachments(vorgang);
	}

	loadPendingCommandsByVorgang(vorgang: VorgangWithEingangResource): void {
		if (this.hasPendingCommands(vorgang)) {
			const pendingCommandsSubscription: Subscription = this.commandService.getPendingCommands(vorgang, VorgangWithEingangLinkRel.PENDING_COMMANDS).subscribe(result => {
				this.setPendingForwardCommand(result.resource);
				pendingCommandsSubscription.unsubscribe();
			})
		}
	}

	setPendingForwardCommand(pendingCommandList: CommandListResource): void {
		if (!this.existsPendingForwardCommand()) {
			this.setPendingForwardSingleCommand(createStateResource(getPendingForwardingCommand(pendingCommandList)))
		}
	}

	existsPendingForwardCommand(): boolean {
		return this.pendingForwardCommand$.value.loaded;
	}

	public setPendingForwardSingleCommand(pendingCommand: StateResource<CommandResource>): void {
		this.pendingForwardCommand$.next(pendingCommand);
	}

	public setPendingForwardSingleCommandLoading(): void {
		this.pendingForwardCommand$.next(createEmptyStateResource(true));
	}

	public getPendingForwardCommand(): Observable<StateResource<CommandResource>> {
		return this.pendingForwardCommand$.asObservable();
	}

	private hasPendingCommands(vorgang: VorgangWithEingangResource): boolean {
		return hasLink(vorgang, VorgangWithEingangLinkRel.PENDING_COMMANDS);
	}

	setStateResourceOnLoading(stateResource$: BehaviorSubject<StateResource<any>>): void {
		stateResource$.next({ ...stateResource$.value, loading: true });
	}

	loadAttachments(vorgang: VorgangWithEingangResource): void {
		if (this.hasAttachments(vorgang)) {
			this.doRepositoryCall(this.attachments$, this.vorgangRepository.getAttachments(vorgang));
		}
	}

	doRepositoryCall(stateSubject: BehaviorSubject<StateResource<Resource>>, repositoryCall: Observable<any>): void {
		this.setStateResourceOnLoading(stateSubject);
		repositoryCall.pipe(first()).subscribe(result => {
			stateSubject.next(createStateResource(result));
		})
	}

	private hasAttachments(vorgang: VorgangWithEingangResource): boolean {
		return hasLink(vorgang, VorgangWithEingangLinkRel.ATTACHMENTS);
	}

	getAttachments(): Observable<StateResource<OzgFileListResource>> {
		return this.attachments$;
	}

	private listenForLeavingVorgang(): void {
		this.unsubscribe();
		this.subscription = this.navigationService.urlChanged().subscribe(params => this.onNavigation(params));
	}

	onNavigation(params: Params): void {
		if (NavigationService.isVorgangListPage(params)) {
			this.clearVorgang();
			this.clearPendingForwardCommand();
		}
	}

	private clearVorgang(): void {
		this.vorgangWithEingang$.next(createEmptyStateResource());
	}

	clearPendingForwardCommand(): void {
		this.pendingForwardCommand$.next(createEmptyStateResource());
	}

	getVorgang(): VorgangResource {
		return this.vorgangWithEingang$.value.resource;
	}


	reloadVorgang(commandResource: CommandResource): void {
		this.loadVorgangWithEingang(getUrl(commandResource, CommandLinkRel.EFFECTED_RESOURCE));
	}

	public assignUser(userUri: ResourceUri): Observable<StateResource<CommandResource>> {
		return this.commandService.createCommand(this.vorgangWithEingang$.value.resource, VorgangWithEingangLinkRel.ASSIGN, this.createAssignUserOrder(userUri)).pipe(
			tap(commandStateResource => this.reloadVorgangOnDone(commandStateResource)),
			startWith(createEmptyStateResource<CommandResource>(true))
		)
	}

	reloadVorgangOnDone(commandResource: StateResource<CommandResource>): void {
		if (isDone(commandResource.resource)) {
			this.reloadVorgang(commandResource.resource);
		}
	}

	private createAssignUserOrder(assignedTo: ResourceUri): CreateAssignUserCommand {
		return { order: VorgangOrder.ASSIGN_USER, body: { assignedTo } };
	}

	ngOnDestroy(): void {
		this.unsubscribe();
	}

	private unsubscribe(): void {
		if (!isNil(this.subscription)) this.subscription.unsubscribe();
	}
}
