import { BinaryFileService } from '@goofy-client/binary-file-shared';
import { CommandResource, CommandService } from '@goofy-client/command-shared';
import { OzgFileListResource, OzgFileService } from '@goofy-client/ozg-file-shared';
import { createEmptyStateResource, createStateResource, NavigationService, StateResource } from '@goofy-client/tech-shared';
import { Mock, mock, useFromMock } from '@goofy-client/test-utils';
import { SnackBarService } from '@goofy-client/ui';
import { VorgangWithEingangResource } from '@goofy-client/vorgang-shared';
import { getUrl } from '@ngxp/rest';
import { cold, hot } from 'jest-marbles';
import { CommandLinkRel } from 'libs/command-shared/src/lib/command.linkrel';
import { createCommandResource } from 'libs/command-shared/test/command';
import { createOzgFileListResource } from 'libs/ozg-file-shared/test/ozg-file';
import { createVorgangWithEingangResource } from 'libs/vorgang-shared/test/vorgang';
import { createWiedervorlage, createWiedervorlageListResource, createWiedervorlageResource } from 'libs/wiedervorlage-shared/test/wiedervorlage';
import { of } from 'rxjs';
import { WiedervorlageLinkRel, WiedervorlageListLinkRel } from './wiedervorlage.linkrel';
import { WiedervorlageMessages } from './wiedervorlage.message';
import { Wiedervorlage, WiedervorlageListResource, WiedervorlageOrder, WiedervorlageResource } from './wiedervorlage.model';
import { WiedervorlageRepository } from './wiedervorlage.repository';
import { WiedervorlageService } from './wiedervorlage.service';

describe('WiedervorlageService', () => {
	let service;
	let repository: Mock<WiedervorlageRepository>;
	let navigationService: Mock<NavigationService>;
	let commandService: Mock<CommandService>;
	let snackbarService: Mock<SnackBarService>;
	let ozgFileService: Mock<OzgFileService>;
	let binaryFileService: Mock<BinaryFileService>;

	const vorgangResource: VorgangWithEingangResource = createVorgangWithEingangResource();
	const wiedervorlageList: WiedervorlageListResource = createWiedervorlageListResource();
	const wiedervorlageResource: WiedervorlageResource = createWiedervorlageResource();
	const wiedervorlage: Wiedervorlage = createWiedervorlage();
	const commandResource: CommandResource = createCommandResource();
	const commandResourceWithEffectedResourceLink: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);

	beforeEach(() => {
		repository = mock(WiedervorlageRepository);
		navigationService = mock(NavigationService);
		commandService = mock(CommandService);
		snackbarService = mock(SnackBarService);
		ozgFileService = mock(OzgFileService);
		binaryFileService = mock(BinaryFileService);

		navigationService.urlChanged.mockReturnValue(of(null));

		service = new WiedervorlageService(
			useFromMock(repository),
			useFromMock(navigationService),
			useFromMock(commandService),
			useFromMock(snackbarService),
			useFromMock(ozgFileService),
			useFromMock(binaryFileService),
		);
	})

	describe('getWiedervorlagenByVorgang', () => {

		beforeEach(() => {
			repository.findByVorgang.mockReturnValue(of(wiedervorlageList));
			service.setWiedergabeListLoading = jest.fn();
			service.setWiedervorlagenList = jest.fn();
		})

		it('should set loading to true', () => {
			service.getWiedervorlagenByVorgang(vorgangResource);

			expect(service.setWiedergabeListLoading).toHaveBeenCalled();
		})

		it('should call repository', () => {
			service.getWiedervorlagenByVorgang(vorgangResource).subscribe();

			expect(repository.findByVorgang).toHaveBeenCalledWith(vorgangResource);
		})

		it('should set loading to false', () => {
			service.getWiedervorlagenByVorgang(vorgangResource);

			expect(service.setWiedervorlagenList).toHaveBeenCalled();
		})
	})

	describe('getWiedervorlage', () => {

		beforeEach(() => {
			navigationService.getDecodedParam.mockReturnValue(getUrl(wiedervorlageResource));
			repository.getWiedervorlage.mockReturnValue(of(wiedervorlageResource));
		})

		it('should set loading to true', () => {
			service.setWiedervorlageLoading = jest.fn();

			service.getWiedervorlage();

			expect(service.setWiedervorlageLoading).toHaveBeenCalled();
		})

		it('should call navigationService', () => {
			service.getWiedervorlage();

			expect(navigationService.getDecodedParam).toHaveBeenCalledWith(WiedervorlageService.encodedWiedervorlageUriParam);
		})

		it('should call repository', () => {
			service.getWiedervorlage();

			expect(repository.getWiedervorlage).toHaveBeenCalledWith(getUrl(wiedervorlageResource));
		})

		it('should set returnvalue', () => {
			service.getWiedervorlage();

			expect(service.wiedervorlage$.value).toEqual(createStateResource(wiedervorlageResource));
		})
	})

	describe('saveWiedervorlage', () => {

		beforeEach(() => {
			commandService.createCommand.mockReturnValue(of(createStateResource(commandResourceWithEffectedResourceLink)));
		})

		it('should return initial value and mapped value', () => {
			commandService.createCommand.mockReturnValue(hot('-a', { a: createStateResource(commandResourceWithEffectedResourceLink) }));

			const result = service.saveWiedervorlage(wiedervorlageResource, wiedervorlage);

			expect(result).toBeObservable(cold('ab', { a: createEmptyStateResource(true), b: createStateResource(commandResourceWithEffectedResourceLink) }));
		})

		it('should call command service', () => {
			service.saveWiedervorlage(wiedervorlageResource, wiedervorlage).subscribe();

			expect(commandService.createCommand).toHaveBeenCalledWith(wiedervorlageResource, WiedervorlageLinkRel.EDIT, { order: WiedervorlageOrder.EDIT_WIEDERVORLAGE, wiedervorlage });
		})

		it('should set submit-in-progress on loading', () => {
			service.setSubmitInProgressLoading = jest.fn();

			service.saveWiedervorlage(wiedervorlage);

			expect(service.setSubmitInProgressLoading).toHaveBeenCalled();
		})

		it('should call proceedAfterRecieveCommand', () => {
			service.proceedAfterReceiveCommand = jest.fn();

			service.saveWiedervorlage(wiedervorlageResource, wiedervorlage).subscribe();

			const expectedMessage = WiedervorlageMessages.SAVED.replace('{betreff}', wiedervorlage.betreff);
			expect(service.proceedAfterReceiveCommand).toHaveBeenCalledWith(createStateResource(commandResourceWithEffectedResourceLink), expectedMessage);
		})
	})

	describe('createWiedervorlage', () => {

		beforeEach(() => {
			commandService.createCommand.mockReturnValue(of(createStateResource(commandResource)));
		})

		it('should return intitial value and mapped value', () => {
			commandService.createCommand.mockReturnValue(hot('-a', { a: createStateResource(commandResource) }));

			const result = service.createWiedervorlage(wiedervorlage);

			expect(result).toBeObservable(cold('ab', { a: createEmptyStateResource(true), b: createStateResource(commandResource) }));
		})

		it('should set submit-in-progress on loading', () => {
			service.setSubmitInProgressLoading = jest.fn();

			service.createWiedervorlage(wiedervorlage);

			expect(service.setSubmitInProgressLoading).toHaveBeenCalled();
		})

		it('should call command service', () => {
			service.createWiedervorlage(wiedervorlage).subscribe();

			const command = { order: WiedervorlageOrder.CREATE_WIEDERVORLAGE, wiedervorlage };
			expect(commandService.createCommand).toHaveBeenCalledWith(service.wiedervorlageList$.value.resource, WiedervorlageListLinkRel.CREATE_WIEDERVORLAGE, command);
		})

		it('should call proceedAfterReceiveCommand', () => {
			service.proceedAfterReceiveCommand = jest.fn();

			service.createWiedervorlage(wiedervorlage).subscribe();

			const expectedMessage = WiedervorlageMessages.CREATED.replace('{betreff}', wiedervorlage.betreff);
			expect(service.proceedAfterReceiveCommand).toHaveBeenCalledWith(createStateResource(commandResource), expectedMessage);
		})
	})

	describe('proceedAfterRecieveCommand', () => {

		it('should set reload on wiedervorlageList', () => {
			service.proceedAfterReceiveCommand(createStateResource(commandResourceWithEffectedResourceLink), null);

			expect(service.wiedervorlageList$.value.reload).toBe(true);
		})

		it('should do nothing on apiError', () => {
			service.proceedAfterReceiveCommand(createEmptyStateResource(), null);

			expect(service.wiedervorlageList$.value.reload).toBe(false);
		})

		it('should call showSnackbar', () => {
			snackbarService.showSnackbar = jest.fn();
			const snackbarMessage: string = 'XX wurde XX';

			service.proceedAfterReceiveCommand(createStateResource(commandResourceWithEffectedResourceLink), snackbarMessage);

			expect(snackbarService.showSnackbar).toHaveBeenCalledWith(commandResourceWithEffectedResourceLink, snackbarMessage);
		});
	})

	describe('order actions', () => {
		const wiedervorlageResource: WiedervorlageResource = createWiedervorlageResource();

		beforeEach(() => {
			commandService.createCommand.mockReturnValue(of(createStateResource(commandResourceWithEffectedResourceLink)));
			service.reloadWiedervorlageList = jest.fn();
			service.loadAndSetWiedervorlageByUrl = jest.fn();

			service.wiedervorlage$.next(createStateResource(wiedervorlageResource));
		})

		describe('wiedervorlage erledigen', () => {

			it('should call commandService', () => {
				service.erledigen();

				expect(commandService.createCommand).toHaveBeenCalledWith(wiedervorlageResource, WiedervorlageLinkRel.ERLEDIGEN, { order: WiedervorlageOrder.ERLEDIGEN });
			})

			it('should reload wiedervorlage', () => {
				service.erledigen();

				expect(service.loadAndSetWiedervorlageByUrl).toHaveBeenCalled();
			})

			it('should show snackbar', () => {
				service.erledigen();

				expect(snackbarService.showSnackbar).toHaveBeenCalledWith(commandResourceWithEffectedResourceLink, `Die Wiedervorlage ${wiedervorlageResource.betreff} wurde erledigt`);
			})
		})

		describe('wiedervorlage wiedereroeffnen', () => {

			it('should call commandService', () => {
				service.wiedereroeffnen();

				expect(commandService.createCommand).toHaveBeenCalledWith(wiedervorlageResource, WiedervorlageLinkRel.WIEDEREROEFFNEN, { order: WiedervorlageOrder.WIEDEREROEFFNEN });
			});

			it('should reload wiedervorlage', () => {
				service.wiedereroeffnen();

				expect(service.loadAndSetWiedervorlageByUrl).toHaveBeenCalled();
			});

			it('should show snackbar', () => {
				service.wiedereroeffnen();

				expect(snackbarService.showSnackbar).toHaveBeenCalledWith(commandResourceWithEffectedResourceLink, `Die Wiedervorlage ${wiedervorlageResource.betreff} wurde wiedereröffnet`);
			})
		})
	})

	describe('showSnackbar', () => {

		const message: string = 'SnackbarMessage';

		it('should call service if command is done', () => {
			const commandResource: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);

			service.showSnackBar(commandResource, message);

			expect(snackbarService.showSnackbar).toHaveBeenCalledWith(commandResource, message);
		})

		it('should NOT call service if command is pending', () => {
			const commandResource: CommandResource = createCommandResource();

			service.showSnackBar(commandResource, message);

			expect(snackbarService.showSnackbar).not.toHaveBeenCalled();
		})
	})

	describe('on navigation', () => {

		beforeEach(() => {
			service.navigateToVorgangDetailPage = jest.fn();

			service.setWiedervorlageListReload = jest.fn();
			service.setWiedervorlageReload = jest.fn();
			service.clearAttachmentList = jest.fn();
			service.clearWiedervorlagenList = jest.fn();
		})

		describe('to vorgang detail', () => {

			beforeEach(() => {
				service.navigateToVorgangDetailPage.mockReturnValue(true);
			})

			it('should set wiedervorlage list on reload', () => {
				service.onNavigation({});

				expect(service.setWiedervorlageListReload).toHaveBeenCalled();
			})

			it('should set wiedervorlage on reload', () => {
				service.onNavigation({});

				expect(service.setWiedervorlageReload).toHaveBeenCalled();
			})

			it('should clear attachments', () => {
				service.onNavigation({});

				expect(service.clearAttachmentList).toHaveBeenCalled();
			})
		})

		describe('to vorgang list', () => {

			it('should clear list', () => {
				service.onNavigation({});

				expect(service.clearWiedervorlagenList).toHaveBeenCalled();
			})

			it('should clear attachments', () => {
				service.onNavigation({});

				expect(service.clearAttachmentList).toHaveBeenCalled();
			})
		})
	})

	describe('load attachments', () => {

		const ozgFileListResource: OzgFileListResource = createOzgFileListResource();
		const ozgFileStateResource: StateResource<OzgFileListResource> = createStateResource(ozgFileListResource);

		beforeEach(() => {
			ozgFileService.getFileList.mockReturnValue(of(ozgFileStateResource));
			service.setAttachmentLoading = jest.fn();
		})

		it('should set attachment state resource on loading', () => {
			service.loadAttachments(wiedervorlageResource);

			expect(service.setAttachmentLoading).toHaveBeenCalledWith(wiedervorlageResource);
		})

		it('should call file service', () => {
			service.loadAttachments(wiedervorlageResource);

			expect(ozgFileService.getFileList).toHaveBeenCalledWith(wiedervorlageResource, WiedervorlageLinkRel.ATTACHMENTS);
		})

		it('should set loaded resource into state resource', () => {
			service.loadAttachments(wiedervorlageResource);

			service.getAttachmentList(wiedervorlageResource);
			const result: OzgFileListResource = service.attachmentListByWiedervorlage[getUrl(wiedervorlageResource)].value;

			expect(result).toEqual(ozgFileStateResource);
		})
	})
})