import { Injectable, OnDestroy } from '@angular/core';
import { Params } from '@angular/router';
import { BinaryFileService } from '@goofy-client/binary-file-shared';
import { Command, CommandResource, CommandService, CreateCommand, isDone } from '@goofy-client/command-shared';
import { OzgFileListResource, OzgFileResource, OzgFileService } from '@goofy-client/ozg-file-shared';
import { createEmptyStateResource, createStateResource, doIfLoadingRequired, hasError, NavigationService, replacePlaceholder, StateResource } from '@goofy-client/tech-shared';
import { SnackBarService } from '@goofy-client/ui';
import { VorgangResource, VorgangService } from '@goofy-client/vorgang-shared';
import { getUrl, ResourceUri } from '@ngxp/rest';
import { isEqual, isNil, isNull, isUndefined } from 'lodash-es';
import { BehaviorSubject, Observable, Subscription } from 'rxjs';
import { filter, mergeMap, startWith, tap } from 'rxjs/operators';
import { WiedervorlageLinkRel, WiedervorlageListLinkRel } from './wiedervorlage.linkrel';
import { WiedervorlageMessages } from './wiedervorlage.message';
import { CreateWiedervorlageCommand, Wiedervorlage, WiedervorlageListResource, WiedervorlageOrder, WiedervorlageResource } from './wiedervorlage.model';
import { WiedervorlageRepository } from './wiedervorlage.repository';
import { WiedervorlageRoutes } from './wiedervorlage.route';
import { createErledigenCommand, createWiedereroeffnenCommand } from './wiedervorlage.util';

@Injectable({ providedIn: 'root' })
export class WiedervorlageService implements OnDestroy {

	static encodedWiedervorlageUriParam: string = 'wiedervorlageUrl';

	readonly wiedervorlageList$: BehaviorSubject<StateResource<WiedervorlageListResource>> = new BehaviorSubject<StateResource<WiedervorlageListResource>>(createEmptyStateResource());
	readonly wiedervorlage$: BehaviorSubject<StateResource<WiedervorlageResource>> = new BehaviorSubject<StateResource<WiedervorlageResource>>(createEmptyStateResource());

	readonly submitInProgress$: BehaviorSubject<StateResource<CommandResource>> = new BehaviorSubject<StateResource<CommandResource>>(createEmptyStateResource());

	readonly erledigenCommand$: BehaviorSubject<StateResource<CommandResource>> = new BehaviorSubject<StateResource<CommandResource>>(createEmptyStateResource());
	readonly wiedereroeffnenCommand$: BehaviorSubject<StateResource<CommandResource>> = new BehaviorSubject<StateResource<CommandResource>>(createEmptyStateResource());

	attachmentListByWiedervorlage = <any>{};

	private subscription: Subscription;
	private attachmentSubscription: Subscription;

	constructor(
		private repository: WiedervorlageRepository,
		private navigationService: NavigationService,
		private commandService: CommandService,
		private snackbarService: SnackBarService,
		private ozgFileService: OzgFileService,
		private binaryFileService: BinaryFileService
	) {
		this.listenToNavigation();
	}

	getSubmitInProgress(): Observable<StateResource<CommandResource>> {
		return this.submitInProgress$.asObservable();
	}

	public getWiedervorlagenByVorgang(vorgangResource: VorgangResource): Observable<StateResource<WiedervorlageListResource>> {
		doIfLoadingRequired(this.wiedervorlageList$.value, () => {
			this.setWiedergabeListLoading();
			this.loadWiedervorlagenByVorgang(vorgangResource);
		});
		return this.wiedervorlageList$.asObservable();
	}

	private loadWiedervorlagenByVorgang(vorgangResource: VorgangResource): void {
		const subscription: Subscription = this.repository.findByVorgang(vorgangResource).subscribe(wiedervorlagenList => {
			if (!isNull(wiedervorlagenList)) {
				this.setWiedervorlagenList(wiedervorlagenList);
				subscription.unsubscribe();
			}
		});
	}

	loadStrictWiedervorlagenNotDone(vorgangResource: VorgangResource): Observable<StateResource<WiedervorlageListResource>> {
		this.wiedervorlageList$.next(createEmptyStateResource());
		return this.getWiedervorlagenByVorgang(vorgangResource);
	}

	setWiedergabeListLoading(): void {
		this.wiedervorlageList$.next({ ...this.wiedervorlageList$.value, loading: true });
	}

	setWiedervorlagenList(wiedervorlagenList: WiedervorlageListResource): void {
		this.wiedervorlageList$.next(createStateResource(wiedervorlagenList));
	}

	public getWiedervorlage(): Observable<StateResource<WiedervorlageResource>> {
		doIfLoadingRequired(this.wiedervorlage$.value, () => {
			this.isNewWiedervorlage()
				? this.wiedervorlage$.next(createEmptyStateResource())
				: this.updateWiedervorlage();
		});
		return this.wiedervorlage$;
	}

	private updateWiedervorlage(): void {
		this.setWiedervorlageLoading();

		this.loadAndSetWiedervorlageByUrl(this.getWiedervorlageUri());
	}

	private loadAndSetWiedervorlageByUrl(url: ResourceUri): void {
		this.setWiedervorlageLoading();

		const subscription: Subscription = this.repository.getWiedervorlage(url).subscribe(wiedervorlage => {
			this.wiedervorlage$.next(createStateResource(wiedervorlage));
			subscription.unsubscribe();
		});
	}

	private isNewWiedervorlage(): boolean {
		return isEqual(this.navigationService.getParam(WiedervorlageService.encodedWiedervorlageUriParam), WiedervorlageRoutes.NEW);
	}

	private listenToNavigation(): void {
		this.unsubscribe();
		this.subscription = this.navigationService.urlChanged().subscribe(params => this.onNavigation(params));
	}

	onNavigation(params: Params): void {
		if (this.navigateToVorgangDetailPage(params)) {
			this.setWiedervorlageListReload();
			this.setWiedervorlageReload();
			this.submitInProgress$.next(createEmptyStateResource());
			this.clearAttachmentList();
		}
		if (NavigationService.isVorgangListPage(params)) {
			this.clearWiedervorlagenList();
			this.submitInProgress$.next(createEmptyStateResource());
			this.clearAttachmentList();
		}
	}

	clearAttachmentList(): void {
		this.attachmentListByWiedervorlage = <any>{};
	}

	clearWiedervorlagenList(): void {
		this.wiedervorlageList$.next(createEmptyStateResource());
	}

	private setWiedervorlageReload(): void {
		this.wiedervorlage$.next({ ...this.wiedervorlage$.value, reload: true });
	}

	private navigateToVorgangDetailPage(params: Params): boolean {
		return isUndefined(params[WiedervorlageService.encodedWiedervorlageUriParam] && !isUndefined(params[VorgangService.VORGANG_WITH_EINGANG_URL]));
	}

	private getWiedervorlageUri(): ResourceUri {
		return this.navigationService.getDecodedParam(WiedervorlageService.encodedWiedervorlageUriParam);
	}

	setWiedervorlageLoading(): void {
		this.wiedervorlage$.next({ ...this.wiedervorlage$.value, loading: true });
	}

	public createWiedervorlage(wiedervorlage: Wiedervorlage): Observable<StateResource<CommandResource>> {
		this.setSubmitInProgressLoading();

		return this.createCreateWiedervorlageCommand(this.wiedervorlageList$.value.resource, wiedervorlage).pipe(
			filter(commandStateResource => !commandStateResource.loading),
			tap(commandStateResource => this.proceedAfterReceiveCommand(commandStateResource, this.buildMessage(WiedervorlageMessages.CREATED, wiedervorlage))),
			startWith(createEmptyStateResource<CommandResource>(true))
		);
	}

	createCreateWiedervorlageCommand(wiedervorlageListResource: WiedervorlageListResource, wiedervorlage: Wiedervorlage): Observable<StateResource<CommandResource>> {
		return this.commandService.createCommand(wiedervorlageListResource, WiedervorlageListLinkRel.CREATE_WIEDERVORLAGE, this.buildCreateWiedervorlageCommand(wiedervorlage));
	}

	private buildCreateWiedervorlageCommand(wiedervorlage: Wiedervorlage): Command {
		return <any>{ order: WiedervorlageOrder.CREATE_WIEDERVORLAGE, wiedervorlage };
	}

	public saveWiedervorlage(wiedervorlageResource: WiedervorlageResource, wiedervorlage: Wiedervorlage): Observable<StateResource<CommandResource>> {
		this.setSubmitInProgressLoading();

		return this.createSaveWiedervorlageCommand(wiedervorlageResource, wiedervorlage).pipe(
			filter(commandStateResource => !commandStateResource.loading),
			tap(commandStateResource => this.proceedAfterReceiveCommand(commandStateResource, this.buildMessage(WiedervorlageMessages.SAVED, wiedervorlage))),
			startWith(createEmptyStateResource<CommandResource>(true))
		);
	}

	private setSubmitInProgressLoading(): void {
		this.submitInProgress$.next({ ...this.submitInProgress$.value, loading: true });
	}

	createSaveWiedervorlageCommand(wiedervorlageResource: WiedervorlageResource, wiedervorlage: Wiedervorlage): Observable<StateResource<CommandResource>> {
		return this.commandService.createCommand(wiedervorlageResource, WiedervorlageLinkRel.EDIT, this.buildEditWiedervorlageCommand(wiedervorlage));
	}


	private buildEditWiedervorlageCommand(wiedervorlage: Wiedervorlage): CreateWiedervorlageCommand {
		return { order: WiedervorlageOrder.EDIT_WIEDERVORLAGE, wiedervorlage };
	}

	proceedAfterReceiveCommand(commandStateResource: StateResource<CommandResource>, message: string): void {
		this.submitInProgress$.next({ ...this.submitInProgress$.value, resource: commandStateResource.resource });
		if (isDone(commandStateResource.resource)) {
			this.submitInProgress$.next(createStateResource(commandStateResource.resource));
			this.snackbarService.showSnackbar(commandStateResource.resource, message)
			this.setWiedervorlageListReload();
		} else if (hasError(commandStateResource)) {
			this.submitInProgress$.next(createStateResource(commandStateResource.resource));
		}
	}

	setWiedervorlageListReload(): void {
		this.wiedervorlageList$.next({ ...this.wiedervorlageList$.value, reload: true });
	}

	public erledigen(): Observable<StateResource<CommandResource>> {
		return this.doAction(this.erledigenCommand$, WiedervorlageLinkRel.ERLEDIGEN, createErledigenCommand, WiedervorlageMessages.ERLEDIGT);
	}

	public wiedereroeffnen(): Observable<StateResource<CommandResource>> {
		return this.doAction(this.wiedereroeffnenCommand$, WiedervorlageLinkRel.WIEDEREROEFFNEN, createWiedereroeffnenCommand, WiedervorlageMessages.WIEDEREROEFFNET);
	}

	private doAction(commandStateSubj: BehaviorSubject<StateResource<CommandResource>>, linkRel: string, commandProvider: () => CreateCommand, snackBarMessage: string): Observable<StateResource<any>> {
		commandStateSubj.next(createEmptyStateResource(true));

		const commandSubscription: Subscription = this.commandService.createCommand(this.wiedervorlage$.value.resource, linkRel, commandProvider())
			.subscribe(commandStateResource => {
				if (isDone(commandStateResource.resource)) {
					this.handleSource(commandStateResource.resource);
					this.showSnackBar(commandStateResource.resource, this.buildMessage(snackBarMessage, this.wiedervorlage$.value.resource));
					commandStateSubj.next(commandStateResource);
					commandSubscription.unsubscribe();
				}
			})

		return commandStateSubj.asObservable();
	}

	private handleSource(commandResource: CommandResource): void {
		if (isDone(commandResource)) {
			this.loadAndSetWiedervorlageByUrl(this.getWiedervorlageUri());
		}
	}

	private buildMessage(message: string, wiedervorlage: Wiedervorlage) {
		return replacePlaceholder(message, 'betreff', wiedervorlage.betreff);
	}

	showSnackBar(commandResource: CommandResource, message: string) {
		if (isDone(commandResource)) {
			this.snackbarService.showSnackbar(commandResource, message);
		}
	}

	public getAttachmentList(wiedervorlage: WiedervorlageResource): Observable<StateResource<OzgFileListResource>> {
		const uri: ResourceUri = getUrl(wiedervorlage);

		if (isNil(this.attachmentListByWiedervorlage[uri])) {
			this.attachmentListByWiedervorlage[uri] = new BehaviorSubject(createEmptyStateResource());
		} else {
			this.attachmentListByWiedervorlage[uri].next(createEmptyStateResource());
		}

		doIfLoadingRequired(this.attachmentListByWiedervorlage[uri].value, () => this.loadAttachments(wiedervorlage));

		return this.attachmentListByWiedervorlage[uri].asObservable();
	}

	loadAttachments(wiedervorlage: WiedervorlageResource): void {
		this.setAttachmentLoading(wiedervorlage);
		if (!isNil(this.attachmentSubscription)) this.attachmentSubscription.unsubscribe();
		this.attachmentSubscription = this.ozgFileService.getFileList(wiedervorlage, WiedervorlageLinkRel.ATTACHMENTS).subscribe(fileList => {
			if (fileList.loaded) {
				this.attachmentListByWiedervorlage[getUrl(wiedervorlage)].next(fileList);
			}
		});
	}

	setAttachmentLoading(wiedervorlage: WiedervorlageResource): void {
		this.attachmentListByWiedervorlage[getUrl(wiedervorlage)].next({ ...this.attachmentListByWiedervorlage[getUrl(wiedervorlage)].value, loading: true });
	}

	public uploadFile(file: File): Observable<StateResource<OzgFileResource>> {
		return this.binaryFileService.uploadFile(this.wiedervorlageList$.value.resource, WiedervorlageListLinkRel.UPLOAD_FILE, file).pipe(
			mergeMap(a => this.ozgFileService.getFile(a)),
			startWith(createEmptyStateResource<OzgFileResource>(true))
		);
	}

	ngOnDestroy(): void {
		this.unsubscribe();
	}

	private unsubscribe(): void {
		if (!isNil(this.subscription)) this.subscription.unsubscribe();
	}
}