import { Injectable, OnDestroy } from '@angular/core';
import { FormBuilder, FormControl, FormGroup } from '@angular/forms';
import { OzgFileListLinkRel, OzgFileListResource, OzgFileResource } from '@goofy-client/ozg-file-shared';
import { AbstractFormService, createFutureDate, createStateResource, format, StateResource } from '@goofy-client/tech-shared';
import { Wiedervorlage, WiedervorlageLinkRel, WiedervorlageResource, WiedervorlageService } from '@goofy-client/wiedervorlage-shared';
import { getEmbeddedResource, getUrl, hasLink } from '@ngxp/rest';
import { isNil, isNull } from 'lodash-es';
import { BehaviorSubject, Observable, Subscription } from 'rxjs';

@Injectable()
export class WiedervorlageFormService extends AbstractFormService implements OnDestroy {

	static readonly FIELD_BETREFF = 'betreff';
	static readonly FIELD_BESCHREIBUNG = 'beschreibung';
	static readonly FIELD_FRIST = 'frist';

	static readonly FIELD_PATH_PREFIX = 'command.wiedervorlage';

	private attachments: BehaviorSubject<StateResource<OzgFileResource[]>>;
	private attachmentSubscription: Subscription;

	constructor(
		formBuilder: FormBuilder,
		private wiedervorlageService: WiedervorlageService
	) {
		super(formBuilder);

		this.attachments = new BehaviorSubject(createStateResource([]));
	}

	doAfterPatch(wiedervorlage: WiedervorlageResource): void {
		if (!isNil(this.attachmentSubscription)) this.attachmentSubscription.unsubscribe();

		if (this.hasAttachments(wiedervorlage)) {
			this.attachments.next({ ...this.attachments.value, reload: true });
			this.subscribeToAttachments(wiedervorlage);
		} else {
			this.attachments.next(createStateResource([]));
		}
	}

	private hasAttachments(wiedervorlage: WiedervorlageResource): boolean {
		return !isNull(wiedervorlage) && hasLink(wiedervorlage, WiedervorlageLinkRel.ATTACHMENTS);
	}

	private subscribeToAttachments(wiedervorlage: WiedervorlageResource): void {
		this.attachmentSubscription = this.wiedervorlageService.getAttachmentList(wiedervorlage).subscribe(attachmentsList => this.handleAttachmentList(attachmentsList));
	}

	handleAttachmentList(attachmentsList: StateResource<OzgFileListResource>): void {
		if (attachmentsList.resource) {
			this.attachments.next(createStateResource(getEmbeddedResource<OzgFileResource[]>(attachmentsList.resource, OzgFileListLinkRel.FILE_LIST)));
		}
	}

	protected initForm(): FormGroup {
		return this.formBuilder.group({
			[WiedervorlageFormService.FIELD_BETREFF]: new FormControl(null),
			[WiedervorlageFormService.FIELD_FRIST]: new FormControl(createFutureDate(14)),
			[WiedervorlageFormService.FIELD_BESCHREIBUNG]: new FormControl()
		})
	}

	protected getPathPrefix(): string {
		return WiedervorlageFormService.FIELD_PATH_PREFIX;
	}

	protected doSubmit(): Observable<StateResource<any>> {
		if (this.isPatch()) {
			return this.wiedervorlageService.saveWiedervorlage(this.getSourceValue(), this.getValue());
		} else {
			return this.wiedervorlageService.createWiedervorlage(this.getValue());
		}
	}

	getValue(): Wiedervorlage {
		const formValue = super.getFormValue();
		return { ...formValue, frist: format(formValue.frist), attachments: this.attachments.value.resource.map(ozgFile => getUrl(ozgFile)) };
	}

	public getAttachments(): Observable<StateResource<OzgFileResource[]>> {
		return this.attachments.asObservable();
	}

	public addAttachment(attachment: OzgFileResource): void {
		let attachments = this.attachments.value;
		attachments.resource.push(attachment);
		this.attachments.next(createStateResource(attachments.resource));
	}

	public deleteAttachment(attachment: OzgFileResource): void {
		let attachments = this.attachments.value;
		attachments.resource.splice(attachments.resource.indexOf(attachment), 1);
		this.attachments.next(createStateResource(attachments.resource));
	}

	ngOnDestroy(): void {
		this.unsubscribe();
	}

	unsubscribe(): void {
		if (!isNil(this.attachmentSubscription)) this.attachmentSubscription.unsubscribe();
	}
}