package de.itvsh.goofy.binary;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.io.IOException;
import java.util.List;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.ByteArrayResource;
import org.springframework.core.io.Resource;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestPart;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;

import com.google.protobuf.ByteString;

import de.itvsh.goofy.common.errorhandling.TechnicalException;
import de.itvsh.goofy.common.file.OzgFile;
import de.itvsh.goofy.common.file.OzgFileData;

@RestController
@RequestMapping(BinaryFileController.PATH)
public class BinaryFileController {

	static final String PATH = "/api/binaryFiles"; // NOSONAR

	@Autowired
	private BinaryFileService service;
	@Autowired
	private BinaryFileModelAssembler assembler;
	@Autowired
	private FileIdMapper fileIdMapper;

	@GetMapping("/{fileId}")
	public EntityModel<OzgFile> getFile(@PathVariable FileId fileId) {
		return assembler.toModel(service.getFile(fileId));
	}

	@PostMapping("/{vorgangId}/{fieldName}/file")
	public ResponseEntity<Void> uploadFile(@PathVariable String vorgangId, @PathVariable String fieldName, @RequestPart MultipartFile file) {
		var fileId = service.uploadFile(vorgangId, fieldName, buildOzgFileData(file));

		return ResponseEntity.created(linkTo(BinaryFileController.class).slash(fileId).toUri()).build();
	}

	private OzgFileData buildOzgFileData(MultipartFile file) {
		return OzgFileData.builder()
				.name(file.getOriginalFilename())
				.contentType(file.getContentType())
				.content(ByteString.copyFrom(getData(file)))
				.size(file.getSize())
				.build();
	}

	public CollectionModel<EntityModel<OzgFile>> getFiles(List<FileId> fileIds) {
		var files = service.getFiles(mapToFileId(fileIds)).collect(Collectors.toList());

		return assembler.toCollectionModel(files);
	}

	private List<FileId> mapToFileId(List<FileId> fileIds) {
		return fileIds.stream()
				.map(fileIdMapper::toString)
				.map(idString -> FileId.from(idString.substring(idString.lastIndexOf("/") + 1, idString.length())))
				.collect(Collectors.toList());
	}

	private byte[] getData(MultipartFile file) {
		try {
			return file.getBytes();
		} catch (IOException e) {
			throw new TechnicalException("Error reading File data.", e.getCause());
		}
	}

	@GetMapping(value = "/{fileId}", //
			produces = { MediaType.APPLICATION_OCTET_STREAM_VALUE, "images/*" }, //
			consumes = MediaType.APPLICATION_OCTET_STREAM_VALUE)
	public ResponseEntity<Resource> getFileData(@PathVariable FileId fileId) {
		var ozgFileData = service.getFileData(fileId);

		return buildResponseEntity(ozgFileData);
	}

	private ResponseEntity<Resource> buildResponseEntity(OzgFileData ozgFileData) {
		return ResponseEntity.ok()
				.header(HttpHeaders.CONTENT_DISPOSITION, String.format("attachment; filename=%s", ozgFileData.getName()))
				.contentLength(ozgFileData.getSize())
				.contentType(MediaType.valueOf(ozgFileData.getContentType()))
				.body(new ByteArrayResource(ozgFileData.getContent().toByteArray()));
	}
}