package de.itvsh.goofy.common;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.reflect.ConstructorUtils;

import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.databind.BeanProperty;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.JavaType;
import com.fasterxml.jackson.databind.JsonDeserializer;
import com.fasterxml.jackson.databind.deser.ContextualDeserializer;
import com.fasterxml.jackson.databind.deser.std.StdDeserializer;

import de.itvsh.goofy.common.datatypes.StringBasedValue;
import de.itvsh.goofy.common.errorhandling.TechnicalException;

public class LinkedResourceDesiralizer extends StdDeserializer<Object> implements ContextualDeserializer {

	private static final long serialVersionUID = 1L;

	private BeanProperty beanProperty;
	private final JavaType targetType;
	private LinkedResource annotation;

	protected LinkedResourceDesiralizer() {
		super(Object.class);
		targetType = null;
	}

	protected LinkedResourceDesiralizer(BeanProperty beanProperty) {
		super(Object.class);
		this.beanProperty = beanProperty;
		this.targetType = beanProperty.getType();
		this.annotation = beanProperty.getAnnotation(LinkedResource.class);
	}

	@Override
	public Object deserialize(JsonParser jsonParser, DeserializationContext ctxt) throws IOException {
		if (jsonParser.isExpectedStartArrayToken()) {
			Collection<Object> idList = targetType.getRawClass().isAssignableFrom(Set.class) ? new HashSet<>() : new ArrayList<>();

			while (!jsonParser.nextToken().isStructEnd()) {
				idList.add(extractId(jsonParser.getText()));
			}
			return idList;
		} else {
			return extractId(jsonParser.getText());
		}
	}

	Object extractId(String url) {
		Class<?> type;
		if (targetType.isCollectionLikeType()) {
			type = targetType.getContentType().getRawClass();
		} else {
			type = targetType.getRawClass();
		}

		if (String.class.isAssignableFrom(type)) {
			return extractStringId(url);
		}
		if (Long.class.isAssignableFrom(type) || Long.TYPE.isAssignableFrom(type)) {
			return extractLongId(url);
		}
		if (StringBasedValue.class.isAssignableFrom(type)) {
			return extractStringBasedValue(url);
		}
		return buildByBuilder(url);
	}

	private Object buildByBuilder(String url) {
		ObjectBuilder<?> builder;
		try {
			builder = ConstructorUtils.invokeConstructor(annotation.builder()).constructContextAware(beanProperty);
		} catch (NoSuchMethodException | IllegalAccessException | InvocationTargetException | InstantiationException e) {
			throw new TechnicalException("Error instanciating Builder.", e);
		}
		return builder.build(extractStringId(url));
	}

	public static Long extractLongId(String uri) {
		var trimedUri = StringUtils.trimToNull(uri);
		if (Objects.isNull(trimedUri)) {
			return null;
		}
		return Long.parseLong(URLDecoder.decode(trimedUri.substring(trimedUri.lastIndexOf('/') + 1), StandardCharsets.UTF_8));
	}

	private StringBasedValue extractStringBasedValue(String url) {
		String value = extractStringId(url);
		Method fromMethod;
		try {
			fromMethod = targetType.getRawClass().getMethod("from", String.class);
		} catch (NoSuchMethodException e) {
			throw new IllegalStateException(
					String.format("Cannot generate Id from type '%s'. Missing 'from' Method.", targetType.getRawClass().getSimpleName()));
		}
		try {
			return (StringBasedValue) fromMethod.invoke(null, value);
		} catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
			throw new IllegalStateException(
					String.format("Cannot generate Id from type '%s'. Error calling 'from' Method.", targetType.getRawClass().getSimpleName()),
					e);
		}
	}

	public static String extractStringId(String url) {
		return URLDecoder.decode(url.substring(url.lastIndexOf('/') + 1), StandardCharsets.UTF_8);
	}

	@Override
	public JsonDeserializer<?> createContextual(DeserializationContext ctxt, BeanProperty property) {
		return new LinkedResourceDesiralizer(property);
	}

}
