package de.itvsh.goofy.common;

import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.BooleanSupplier;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.function.UnaryOperator;
import java.util.stream.Collectors;

import org.apache.commons.lang3.reflect.FieldUtils;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
public class ModelBuilder<T> {

	private static final Map<Class<?>, List<Field>> ANNOTATED_FIELDS = new ConcurrentHashMap<>();

	private final T entity;
	private final EntityModel<T> model;

	private final List<Link> links = new LinkedList<>();
	private final List<Function<EntityModel<T>, EntityModel<T>>> mapper = new LinkedList<>();

	private ModelBuilder(T entity) {
		this.entity = entity;
		this.model = null;
	}

	private ModelBuilder(EntityModel<T> model) {
		this.entity = null;
		this.model = model;
	}

	public static <T> ModelBuilder<T> fromEntity(T entity) {
		return new ModelBuilder<>(entity);
	}

	public static <T> ModelBuilder<T> fromModel(EntityModel<T> model) {
		return new ModelBuilder<>(model);
	}

	public ModelBuilder<T> addLink(Link link) {
		links.add(link);
		return this;
	}

	public ModelBuilder<T> addLinks(Link... links) {
		this.links.addAll(Arrays.asList(links));
		return this;
	}

	public ModelBuilder<T> addLinks(Collection<Link> links) {
		this.links.addAll(links);
		return this;
	}

	public ConditionalLinkAdder ifMatch(Predicate<T> predicate) {
		return new ConditionalLinkAdder(predicate.test(Objects.isNull(entity) ? model.getContent() : entity));
	}

	public ConditionalLinkAdder ifMatch(BooleanSupplier guard) {
		return new ConditionalLinkAdder(guard.getAsBoolean());
	}

	public ModelBuilder<T> map(UnaryOperator<EntityModel<T>> mapper) {
		this.mapper.add(mapper);
		return this;
	}

	public EntityModel<T> buildModel() {
		var filteredLinks = links.stream().filter(Objects::nonNull).collect(Collectors.toSet());

		EntityModel<T> buildedModel = Objects.isNull(model) ? EntityModel.of(entity) : model;
		buildedModel = buildedModel.add(filteredLinks);
		addLinkToLinkedResourcesIfMissing(buildedModel);

		return applyMapper(buildedModel);
	}

	private EntityModel<T> applyMapper(EntityModel<T> resource) {
		Iterator<Function<EntityModel<T>, EntityModel<T>>> i = mapper.iterator();
		EntityModel<T> result = resource;
		while (i.hasNext()) {
			result = i.next().apply(result);
		}
		return result;
	}

	private T getEntity() {
		return Optional.ofNullable(entity == null ? model.getContent() : entity)
				.orElseThrow(() -> new IllegalStateException("Entity must not null for ModelBuilding"));
	}

	private void addLinkToLinkedResourcesIfMissing(EntityModel<T> resource) {
		var fields = ANNOTATED_FIELDS.get(getEntity().getClass());
		if (fields == null) {
			fields = FieldUtils.getFieldsListWithAnnotation(getEntity().getClass(), LinkedResource.class);
			ANNOTATED_FIELDS.put(getEntity().getClass(), fields);
		}

		fields.forEach(field -> {
			String fieldName = sanitizeName(field.getName());
			if (field.getType().isArray() || Collection.class.isAssignableFrom(field.getType()) || resource.hasLink(fieldName)) {
				return;
			}
			Class<?> controllerClass = field.getAnnotation(LinkedResource.class).controllerClass();
			getEntityFieldValue(field).ifPresent(val -> resource.add(WebMvcLinkBuilder.linkTo(controllerClass).slash(val).withRel(fieldName)));
		});

	}

	private String sanitizeName(String fieldName) {
		if (fieldName.endsWith("Id")) {
			return fieldName.substring(0, fieldName.indexOf("Id"));
		}
		return fieldName;
	}

	private Optional<Object> getEntityFieldValue(Field field) {
		try {
			field.setAccessible(true);
			Optional<Object> value = Optional.ofNullable(field.get(getEntity()));
			field.setAccessible(false);
			return value;
		} catch (IllegalArgumentException | IllegalAccessException e) {
			LOG.warn("Cannot access field value of LinkedResource field.", e);
		}
		return Optional.empty();
	}

	@RequiredArgsConstructor
	public class ConditionalLinkAdder {

		public final boolean conditionFulfilled;

		public ModelBuilder<T> addLink(Supplier<Link> linkSupplier) {
			if (conditionFulfilled) {
				addLink(linkSupplier.get());
			}
			return ModelBuilder.this;
		}

		public ModelBuilder<T> addLink(Function<T, Link> linkBuilder) {
			if (conditionFulfilled) {
				addLink(linkBuilder.apply(getEntity()));
			}
			return ModelBuilder.this;
		}

		public ModelBuilder<T> addLink(Link link) {
			if (conditionFulfilled) {
				links.add(link);
			}
			return ModelBuilder.this;
		}

		public ModelBuilder<T> addLinks(Link... links) {
			if (conditionFulfilled) {
				ModelBuilder.this.links.addAll(Arrays.asList(links));
			}
			return ModelBuilder.this;
		}

		@SafeVarargs
		public final ModelBuilder<T> addLinks(Supplier<Link>... linkSuppliers) {
			if (conditionFulfilled) {
				for (int i = 0; i < linkSuppliers.length; i++) {
					ModelBuilder.this.links.add(linkSuppliers[i].get());
				}
			}
			return ModelBuilder.this;
		}

		public final ModelBuilder<T> addLinks(Supplier<Collection<Link>> linksSupplier) {
			if (conditionFulfilled) {
				linksSupplier.get().forEach(ModelBuilder.this.links::add);
			}

			return ModelBuilder.this;
		}

	}
}