package de.itvsh.goofy.common.command;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PatchMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import de.itvsh.goofy.common.LinkedResourceDesiralizer;
import lombok.Getter;

@RestController
@RequestMapping(CommandController.COMMANDS_PATH)
public class CommandController {

	static final String COMMANDS_PATH = "/api/commands"; // NOSONAR

	static final String PARAM_PENDING = "pending";
	static final String PARAM_VORGANG_ID = "vorgangId";

	@Autowired
	private CommandService service;
	@Autowired
	private CommandModelAssembler modelAssembler;

	@GetMapping("{commandId}")
	public EntityModel<Command> getById(@PathVariable String commandId) {
		return modelAssembler.toModel(service.getById(commandId));
	}

	@PatchMapping("{commandId}")
	public ResponseEntity<EntityModel<Command>> revoke(@PathVariable String commandId, @RequestBody StatusPatch patch) {
		if (patch.getStatus() == CommandStatus.REVOKED) {
			return ResponseEntity.ok(modelAssembler.toModel(service.revoke(commandId)));
		}

		return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
	}

	@Getter
	static class StatusPatch {
		private CommandStatus status;
	}

	@GetMapping(params = { PARAM_PENDING, PARAM_VORGANG_ID })
	public CollectionModel<EntityModel<Command>> getPendingCommands(@RequestParam boolean pending, @RequestParam String vorgangId) {
		return modelAssembler.toCollectionModel(service.getPendingCommands(vorgangId));
	}

	@RestController
	@RequestMapping(CommandByRelationController.COMMAND_BY_RELATION_PATH)
	public static class CommandByRelationController {

		private static final Set<String> PARSE_URI_BODY_FIELDS = Set.of("assignedTo");

		static final String COMMAND_BY_RELATION_PATH = "/api/vorgangs/{vorgangId}/relations/{relationId}"; // NOSONAR
		static final String COMMAND_BY_RELATION_AND_VERSION_PATH = COMMAND_BY_RELATION_PATH + "/{relationVersion}/commands";
		static final String PARAM_ORDER = "order";

		@Autowired
		private CommandService service;

		@PostMapping("{relationVersion}/commands")
		public ResponseEntity<Void> createCommandByRelation(@PathVariable String vorgangId, @PathVariable String relationId,
				@PathVariable long relationVersion,
				@RequestBody CreateCommand command) {
			var created = service.createCommand(vorgangId, relationId, relationVersion,
					command.toBuilder().body(bodyProcessor(Optional.ofNullable(command.getBody()))).build());

			return ResponseEntity.created(linkTo(CommandController.class).slash(created.getId()).toUri()).build();
		}

		private Map<String, String> bodyProcessor(Optional<Map<String, String>> bodyMap) {
			return bodyMap.map(map -> map.entrySet().stream().map(entry -> Pair.of(entry.getKey(), processValue(entry.getKey(), entry.getValue())))
					.collect(Collectors.toMap(Pair::getKey, Pair::getValue))).orElse(null);
		}

		private String processValue(String key, String value) {
			if (PARSE_URI_BODY_FIELDS.contains(key)) {
				return LinkedResourceDesiralizer.extractStringId(value);
			}
			return value;
		}
	}

	public boolean existsPendingCommands(String relationId) {
		return service.existsPendingCommands(relationId);
	}
}