package de.itvsh.goofy.common.command;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Optional;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.LinkRelation;
import org.springframework.hateoas.server.RepresentationModelAssembler;
import org.springframework.stereotype.Component;

import de.itvsh.goofy.common.errorhandling.FunctionalException;
import de.itvsh.goofy.kommentar.KommentarController;
import de.itvsh.goofy.vorgang.VorgangController;
import de.itvsh.goofy.vorgang.forwarding.ForwardingController;
import de.itvsh.goofy.wiedervorlage.WiedervorlageController;

@Component
class CommandModelAssembler implements RepresentationModelAssembler<Command, EntityModel<Command>> {

	static final LinkRelation REL_EFFECTED_RESOURCE = LinkRelation.of("effected_resource");
	static final LinkRelation REL_REVOKE = LinkRelation.of("revoke");
	static final LinkRelation REL_UPDATE = LinkRelation.of("update");

	private static final Predicate<Command> IS_DONE = command -> command.getStatus() == CommandStatus.FINISHED
			|| command.getStatus() == CommandStatus.REVOKED || command.getStatus() == CommandStatus.ERROR;

	private static final Predicate<Command> IS_PENDING = command -> command.getStatus() == CommandStatus.PENDING
			|| command.getStatus() == CommandStatus.REVOKE_PENDING;

	@Override
	public EntityModel<Command> toModel(Command entity) {
		Link selfLink = linkTo(CommandController.class).slash(entity.getId()).withSelfRel();
		Link revokeLink = linkTo(methodOn(CommandController.class).revoke(entity.getId(), null)).withRel(REL_REVOKE);
		Link updateLink = linkTo(methodOn(CommandController.class).getById(entity.getId())).withRel(REL_UPDATE);

		var resultModel = EntityModel.of(entity, selfLink);
		resultModel = addIf(IS_DONE, resultModel, effectedResourceLinkByOrder(entity));
		resultModel = addIf(IS_PENDING, resultModel, updateLink);
		resultModel = addIf(command -> command.getOrder().isRevokeable(), resultModel, revokeLink);
		return resultModel;
	}

	private Link effectedResourceLinkByOrder(Command entity) {
		switch (entity.getOrder().getType()) {
		case VORGANG:
			return linkTo(methodOn(VorgangController.class).getVorgangWithEingang(entity.getRelationId())).withRel(REL_EFFECTED_RESOURCE);
		case WIEDERVORLAGE:
			return linkTo(methodOn(WiedervorlageController.class).getById(entity.getRelationId())).withRel(REL_EFFECTED_RESOURCE);
		case KOMMENTAR:
			return linkTo(KommentarController.class).slash(entity.getRelationId()).withRel(REL_EFFECTED_RESOURCE);
		case FORWARDING:
			return linkTo(methodOn(ForwardingController.class).findByVorgangId(entity.getVorgangId())).withRel(REL_EFFECTED_RESOURCE);
		default:
			throw new FunctionalException(() -> "command_order_unkown");
		}
	}

	private EntityModel<Command> addIf(Predicate<Command> condition, EntityModel<Command> model, Link link) {
		return Optional.ofNullable(model.getContent()).filter(condition).map(command -> model.add(link)).orElse(model);
	}

	public CollectionModel<EntityModel<Command>> toCollectionModel(Stream<Command> entities) {
		return CollectionModel.of(entities.map(this::toModel).collect(Collectors.toList()), linkTo(CommandController.class).withSelfRel());
	}
}