package de.itvsh.goofy.common.command;

import java.util.Optional;
import java.util.stream.Stream;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.itvsh.goofy.common.ContextService;
import de.itvsh.ozg.pluto.grpc.command.CommandServiceGrpc.CommandServiceBlockingStub;
import de.itvsh.ozg.pluto.grpc.command.GrpcCreateCommandRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcExistsPendingCommandsRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcExistsPendingCommandsResponse;
import de.itvsh.ozg.pluto.grpc.command.GrpcFindCommandsRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcGetCommandRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcGetPendingCommandsRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcGetPendingCommandsResponse;
import de.itvsh.ozg.pluto.grpc.command.GrpcOrder;
import de.itvsh.ozg.pluto.grpc.command.GrpcRevokeCommandRequest;
import net.devh.boot.grpc.client.inject.GrpcClient;

@Service
class CommandRemoteService {

	@Autowired
	private ContextService contextService;
	@Autowired
	private CommandMapper mapper;
	@Autowired
	private CommandBodyMapper bodyMapper;

	@GrpcClient("pluto")
	private CommandServiceBlockingStub commandServiceStub;

	public Command createCommand(CreateCommand command, long version) {
		GrpcCreateCommandRequest request = createCreateCommandRequest(command, version);
		var response = commandServiceStub.createCommand(request);

		return mapper.toCommand(response.getCommand());
	}

	GrpcCreateCommandRequest createCreateCommandRequest(CreateCommand command, long version) {
		var requestBuilder = GrpcCreateCommandRequest.newBuilder()
				.setCallContext(contextService.createCallContext())
				.setVorgangId(command.getVorgangId())
				.setRelationId(command.getRelationId())
				.setRelationVersion(version)
				.addAllBody(bodyMapper.mapToBodyFields(command.getBody()))
				.setOrder(GrpcOrder.valueOf(command.getOrder().name()));

		Optional.ofNullable(command.getRedirectRequest()).map(mapper::toGrpcRedirectRequest).ifPresent(requestBuilder::setRedirectRequest);

		return requestBuilder.build();
	}

	public Command revokeCommand(String commandId) {
		var response = commandServiceStub.revokeCommand(createRevokeRequest(commandId));

		return mapper.toCommand(response.getCommand());
	}

	GrpcRevokeCommandRequest createRevokeRequest(String commandId) {
		return GrpcRevokeCommandRequest.newBuilder()
				.setContext(contextService.createCallContext())
				.setId(commandId)
				.build();
	}

	public Command getCommand(String commandId) {
		var response = commandServiceStub.getCommand(createGetCommandRequest(commandId));

		return mapper.toCommand(response);
	}

	GrpcGetCommandRequest createGetCommandRequest(String commandId) {
		return GrpcGetCommandRequest.newBuilder()
				.setContext(contextService.createCallContext())
				.setId(commandId)
				.build();
	}

	public boolean existsPendingCommands(String relationId) {
		GrpcExistsPendingCommandsResponse response = commandServiceStub.existsPendingCommands(buildGrpcExistsPendingCommandsRequest(relationId));

		return response.getExistsPendingCommands();
	}

	private GrpcExistsPendingCommandsRequest buildGrpcExistsPendingCommandsRequest(String relationId) {
		return GrpcExistsPendingCommandsRequest.newBuilder()
				.setContext(contextService.createCallContext())
				.setVorgangId(relationId)
				.build();
	}

	public Stream<Command> getPendingCommands(String vorgangId) {
		GrpcGetPendingCommandsResponse response = commandServiceStub.getPendingCommands(buildGrpcGetPendingCommandsRequest(vorgangId));

		return response.getCommandList().stream().map(mapper::toCommand);
	}

	private GrpcGetPendingCommandsRequest buildGrpcGetPendingCommandsRequest(String vorgangId) {
		return GrpcGetPendingCommandsRequest.newBuilder()
				.setContext(contextService.createCallContext())
				.setVorgangId(vorgangId)
				.build();
	}

	public Stream<Command> findCommands(String vorgangId, Optional<CommandStatus> status, Optional<CommandOrder> order) {
		var request = buildFindCommandsRequest(vorgangId, status, order);

		var response = commandServiceStub.findCommands(request);

		return response.getCommandList().stream().map(mapper::toCommand);
	}

	private GrpcFindCommandsRequest buildFindCommandsRequest(String vorgangId, Optional<CommandStatus> status, Optional<CommandOrder> order) {
		var builder = GrpcFindCommandsRequest.newBuilder()
				.setContext(contextService.createCallContext())
				.setVorgangId(vorgangId);

		status.map(CommandStatus::name).ifPresent(builder::addStatus);
		order.map(commandOrder -> GrpcOrder.valueOf(commandOrder.name())).ifPresent(builder::setOrder);
		return builder.build();
	}
}