package de.itvsh.goofy.common.errorhandling;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Path;
import javax.validation.metadata.ConstraintDescriptor;

import org.springframework.core.annotation.Order;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.ResponseStatus;

import lombok.extern.log4j.Log4j2;

@ControllerAdvice
@Log4j2
@Order(99)
public class ExceptionController {

	private static final Set<String> IGNORABLE_CONSTRAINT_VIOLATION_ATTRIBUTES = new HashSet<>(Arrays.asList("groups", "payload", "message"));

	@ExceptionHandler(ResourceNotFoundException.class)
	@ResponseStatus(HttpStatus.NOT_FOUND)
	@ResponseBody
	public ApiError handleResourceNotFoundException(ResourceNotFoundException e) {
		LOG.debug("EntityModel not found: {}", e.getMessage());
		return ApiError.builder().issue(Issue.builder().message(e.getMessage()).messageCode("resource.not_found").build()).build();
	}

	@ExceptionHandler(RuntimeException.class)
	@ResponseStatus(HttpStatus.INTERNAL_SERVER_ERROR)
	@ResponseBody
	public ApiError handleRuntimeException(RuntimeException e) {
		LOG.error("RuntimeException on Request", e);
		return ApiError.builder().issue(Issue.builder().messageCode("generale.server_error").message(e.getMessage()).build()).build();
	}

	@ExceptionHandler(ConstraintViolationException.class)
	@ResponseStatus(HttpStatus.UNPROCESSABLE_ENTITY)
	@ResponseBody
	public ApiError handleConstaintViolationException(ConstraintViolationException e) {
		LOG.info("Validation Exception: {}", e.getMessage());
		LOG.debug("Validation Exception", e);

		return ApiError.builder()//
				.issues(e.getConstraintViolations().stream().map(this::buildIssue).collect(Collectors.toList()))//
				.build();
	}

	private Issue buildIssue(ConstraintViolation<?> violation) {
		return Issue.builder()//
				.field(buildFieldPath(violation.getPropertyPath()))//
				.messageCode(violation.getMessageTemplate().replace("{", "").replace("}", ""))//
				.message(violation.getMessage())//
				.parameters(buildParameters(violation).collect(Collectors.toList()))
				.build();
	}

	private Stream<IssueParam> buildParameters(ConstraintViolation<?> violation) {
		return Optional.ofNullable(violation.getConstraintDescriptor()).map(ConstraintDescriptor::getAttributes)
				.map(descr -> descr.entrySet().stream()
						.filter(entry -> !IGNORABLE_CONSTRAINT_VIOLATION_ATTRIBUTES.contains(entry.getKey()))
						.map(entry -> IssueParam.builder().name(entry.getKey()).value(entry.getValue().toString()).build()))
				.orElse(Stream.empty());
	}

	private String buildFieldPath(Path propertyPath) {
		return propertyPath.toString().substring(propertyPath.toString().indexOf('.') + 1);
	}
}