package de.itvsh.goofy.common.logging;

import java.util.Collection;
import java.util.StringJoiner;

import org.apache.commons.lang3.ArrayUtils;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.itvsh.goofy.common.datatypes.StringBasedValue;
import lombok.extern.log4j.Log4j2;

@Log4j2
public class AspectLoggingUtils {

	private static final String LOGGING_TEMPLATE = "%s: %s";

	private static final String LOGGING_CALL_TEMPLATE = "%s(%s)";
	private static final String LOGGING_RESULT_TEMPLATE = "%s => %s";

	private AspectLoggingUtils() {
	}

	static void log(final JoinPoint joinPoint) {
		try {
			MethodSignature sig = (MethodSignature) joinPoint.getSignature();
			var log = getLogger(joinPoint);
			if (log.isDebugEnabled()) {
				log.debug(String.format(LOGGING_CALL_TEMPLATE, sig.getName(), formatDebuggingParameter(joinPoint)));
			} else {
				log.info(String.format(LOGGING_CALL_TEMPLATE, sig.getName(), formatInfoParameters(joinPoint)));
			}
		} catch (RuntimeException e) {
			LOG.error("Error on generating log output for " + joinPoint, e);
		}
	}

	static void logException(JoinPoint joinPoint, Exception exception) {
		try {
			var log = getLogger(joinPoint);
			if (log.isWarnEnabled()) {
				log.warn(String.format(LOGGING_TEMPLATE, exception.getClass().getSimpleName(), exception.getMessage()));
			}
			if (log.isDebugEnabled()) {
				log.debug(exception.getMessage(), exception);
			}
		} catch (RuntimeException e) {
			LOG.error("Error on generating log on exception.", e);
			LOG.error("Original exception log", exception);
		}
	}

	static void logReturnValue(JoinPoint joinPoint, Object returnValue) {
		try {

			MethodSignature sig = (MethodSignature) joinPoint.getSignature();
			var log = getLogger(joinPoint);
			if (log.isInfoEnabled()) {
				log.info(String.format(LOGGING_RESULT_TEMPLATE, sig.getName(),
						sig.getReturnType() == Void.TYPE ? "<Void>" : formatInfoParameter(returnValue)));
			}
		} catch (RuntimeException e) {
			LOG.error("Error on generating log for returnValue.", e);
		}
	}

	private static Logger getLogger(final JoinPoint joinPoint) {
		return LoggerFactory.getLogger(joinPoint.getSignature().getDeclaringType());
	}

	private static String formatInfoParameters(final JoinPoint joinPoint) {
		String[] paramNames = getSignature(joinPoint).getParameterNames();
		Object[] paramValues = joinPoint.getArgs();

		var stringJoiner = new StringJoiner(", ");

		for (int n = 0; n < paramValues.length; n++) {
			if (paramNames != null) {
				stringJoiner.add(String.format(LOGGING_TEMPLATE, paramNames[n], formatInfoParameter(paramValues[n])));
			} else {
				stringJoiner.add(formatInfoParameter(paramValues[n]));
			}
		}

		return stringJoiner.toString();
	}

	private static String formatInfoParameter(Object value) {
		if (value != null && printValueOnInfoLevel(value.getClass())) {
			return formatParameter(value);
		} else if (value != null) {
			var sb = new StringBuilder();
			sb.append("<").append(value.getClass().getSimpleName());
			if (value.getClass().isArray()) {
				sb.deleteCharAt(sb.length() - 1);
				sb.append(ArrayUtils.getLength(value)).append("]");
			}
			sb.append(">");
			return sb.toString();
		} else {
			return "-null-";
		}
	}

	private static boolean printValueOnInfoLevel(Class<?> paramClass) {
		return paramClass.isPrimitive() || //
				StringBasedValue.class.isAssignableFrom(paramClass) || //
				String.class.isAssignableFrom(paramClass) || //
				Number.class.isAssignableFrom(paramClass);
	}

	private static String formatDebuggingParameter(final JoinPoint joinPoint) {
		String[] paramNames = getSignature(joinPoint).getParameterNames();
		Object[] paramValues = joinPoint.getArgs();

		var sb = new StringBuilder();
		for (int n = 0; n < paramValues.length; n++) {
			if (n > 0) {
				sb.append(", ");
			}
			if (paramNames != null) {
				sb.append(paramNames[n]).append(": ");
			}
			sb.append(formatParameter(paramValues[n]));
		}
		return sb.toString();
	}

	private static String formatParameter(final Object param) {
		if (param == null) {
			return "NULL";
		}
		if (Collection.class.isAssignableFrom(param.getClass())) {
			return "Collection with size " + ((Collection<?>) param).size();
		} else {
			return param.toString();
		}
	}

	private static MethodSignature getSignature(final JoinPoint joinPoint) {
		return (MethodSignature) joinPoint.getSignature();
	}
}