package de.itvsh.goofy.common.user;

import java.util.Collection;
import java.util.Optional;
import java.util.function.Predicate;

import org.apache.commons.lang3.StringUtils;
import org.springframework.security.authentication.AuthenticationTrustResolver;
import org.springframework.security.authentication.AuthenticationTrustResolverImpl;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;

@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class CurrentUserHelper {
	static final String ROLE_PREFIX = "ROLE_";

	public static final Predicate<String> HAS_ROLE = CurrentUserHelper::hasRole;

	private static final AuthenticationTrustResolver TRUST_RESOLVER = new AuthenticationTrustResolverImpl();
	private static final Predicate<Authentication> TRUSTED = auth -> !TRUST_RESOLVER.isAnonymous(auth);

	public static Authentication getAuthentication() {
		return findAuthentication().orElseThrow(() -> new IllegalStateException("No authenticated User found"));
	}

	public static Optional<Authentication> findAuthentication() {
		return Optional.ofNullable(SecurityContextHolder.getContext().getAuthentication()).filter(TRUSTED);
	}

	public static boolean hasRole(String role) {
		var auth = getAuthentication();

		if ((auth == null) || (auth.getPrincipal() == null)) {
			return false;
		}
		Collection<? extends GrantedAuthority> authorities = auth.getAuthorities();
		return containsRole(authorities, role);

	}

	public static boolean containsRole(Collection<? extends GrantedAuthority> authorities, String role) {
		String roleToCheck;

		if (role != null && !role.startsWith(ROLE_PREFIX)) {
			roleToCheck = ROLE_PREFIX + role;
		} else {
			roleToCheck = role;
		}

		if (authorities == null) {
			return false;
		}

		return containsRole(roleToCheck, authorities);
	}

	public static String prepareRoleForCheck(String role) {
		if ((role != null) && (!role.startsWith(ROLE_PREFIX))) {
			return ROLE_PREFIX + role;
		} else {
			return role;
		}
	}

	public static boolean containsRole(String role, Collection<? extends GrantedAuthority> authorities) {
		return authorities.stream().anyMatch(a -> isAuthorityEquals(role, a.getAuthority()));
	}

	private static boolean isAuthorityEquals(String role, String authority) {
		String roleToCheck = prepareRoleForCheck(role);
		return StringUtils.equalsIgnoreCase(role, authority) || StringUtils.equalsIgnoreCase(roleToCheck, authority);
	}

	public static UserId getCurrentUserId() {
		return UserId.from(getAuthentication().getName());
	}

}
