package de.itvsh.goofy.vorgang;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.server.RepresentationModelAssembler;
import org.springframework.stereotype.Component;

import de.itvsh.goofy.common.ModelBuilder;
import de.itvsh.goofy.common.command.CommandController;
import de.itvsh.goofy.common.command.CommandController.CommandByRelationController;
import de.itvsh.goofy.common.user.CurrentUserService;
import de.itvsh.goofy.common.user.UserRole;
import de.itvsh.goofy.kommentar.KommentarController.KommentarByVorgangController;
import de.itvsh.goofy.wiedervorlage.WiedervorlageCommandController.WiedervorlageCommandByVorgangController;
import de.itvsh.goofy.wiedervorlage.WiedervorlageController;

@Component
class VorgangModelAssembler implements RepresentationModelAssembler<Vorgang, EntityModel<Vorgang>> {

	@Autowired
	private CommandController commandController;

	public static final String REL_NEXT = "next";
	public static final String REL_PREVIOUS = "previous";
	public static final String REL_VORGANG_MIT_EINGANG = "vorgang-mit-eingang";
	public static final String REL_WIEDERVORLAGEN = "wiedervorlagen";
	public static final String REL_CREATE_WIEDERVORLAGE = "create-wiedervorlage";
	public static final String REL_KOMMENTARE = "kommentare";
	public static final String REL_PENDING_COMMANDS = "pending-commands";

	public static final String REL_VORGANG_ASSIGN = "assign";

	@Autowired
	private CurrentUserService userService;

	@Override
	public EntityModel<Vorgang> toModel(Vorgang vorgang) {
		var selfLink = linkTo(VorgangController.class).slash(vorgang.getId());

		var builder = ModelBuilder.fromEntity(vorgang).addLink(selfLink.withSelfRel())
				.addLink(linkTo(VorgangController.class).slash(vorgang.getId()).withRel(REL_VORGANG_MIT_EINGANG))
				.ifMatch(this::existsPendingCommands)
				.addLink(() -> linkTo(methodOn(CommandController.class).getPendingCommands(true, vorgang.getId())).withRel(REL_PENDING_COMMANDS))
				.addLink(linkTo(methodOn(KommentarByVorgangController.class).getAll(vorgang.getId())).withRel(REL_KOMMENTARE));

		if (userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)) {
			return builder.buildModel();
		} else {
			return builder
					.addLink(linkTo(methodOn(WiedervorlageController.class).findByVorgang(vorgang.getId())).withRel(REL_WIEDERVORLAGEN))//
					.addLink(linkTo(methodOn(WiedervorlageCommandByVorgangController.class).createCommand(null, vorgang.getId()))
							.withRel(REL_CREATE_WIEDERVORLAGE))
					.addLink(linkTo(methodOn(CommandByRelationController.class).createCommandByRelation(vorgang.getId(), vorgang.getId(),
							vorgang.getVersion(), null)).withRel(REL_VORGANG_ASSIGN))
					.buildModel();
		}
	}

	public CollectionModel<EntityModel<Vorgang>> toCollectionModel(Stream<? extends Vorgang> entities, Optional<Integer> prevPage,
			Optional<Integer> nextPage) {
		var model = CollectionModel.of(entities.map(this::toModel).collect(Collectors.toList()),
				linkTo(VorgangController.class).withSelfRel());

		nextPage.ifPresent(page -> model.add(linkTo(methodOn(VorgangController.class).getVorgangListByPage(page)).withRel(REL_NEXT)));
		prevPage.ifPresent(page -> model.add(linkTo(methodOn(VorgangController.class).getVorgangListByPage(page)).withRel(REL_PREVIOUS)));

		return model;
	}

	private boolean existsPendingCommands(Vorgang vorgang) {
		return commandController.existsPendingCommands(vorgang.getId());
	}
}