package de.itvsh.goofy.vorgang.command;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Objects;
import java.util.function.Function;
import java.util.function.Predicate;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.LinkRelation;
import org.springframework.hateoas.server.LinkBuilder;
import org.springframework.hateoas.server.RepresentationModelProcessor;
import org.springframework.stereotype.Component;

import de.itvsh.goofy.common.command.CommandController.CommandByRelationController;
import de.itvsh.goofy.common.user.CurrentUserService;
import de.itvsh.goofy.common.user.UserRole;
import de.itvsh.goofy.vorgang.Vorgang;
import de.itvsh.goofy.vorgang.Vorgang.VorgangStatus;

@Component
class VorgangCommandProzessor implements RepresentationModelProcessor<EntityModel<? extends Vorgang>> {

	static final LinkRelation REL_VORGANG_ANNEHMEN = LinkRelation.of("annehmen");
	static final LinkRelation REL_VORGANG_VERWERFEN = LinkRelation.of("verwerfen");
	static final LinkRelation REL_VORGANG_ZURUECKHOLEN = LinkRelation.of("zurueckholen");
	static final LinkRelation REL_VORGANG_BEARBEITEN = LinkRelation.of("bearbeiten");
	static final LinkRelation REL_VORGANG_BESCHEIDEN = LinkRelation.of("bescheiden");
	static final LinkRelation REL_VORGANG_ZURUECKSTELLEN = LinkRelation.of("zurueckstellen");
	static final LinkRelation REL_VORGANG_ABSCHLIESSEN = LinkRelation.of("abschliessen");
	static final LinkRelation REL_VORGANG_WIEDEREROEFFNEN = LinkRelation.of("wiedereroeffnen");

	@Autowired
	private CurrentUserService userService;

	@Override
	public EntityModel<? extends Vorgang> process(EntityModel<? extends Vorgang> model) {
		Vorgang vorgang = model.getContent();
		if (Objects.isNull(vorgang))
			return model;

		var result = addLinkIfTransactionAllowed(VorgangStatus.ANGENOMMEN, this::getRelationForAngenommenByStatus, vorgang, model);
		result = addLinkIfTransactionAllowed(VorgangStatus.VERWORFEN, REL_VORGANG_VERWERFEN, vorgang, result);
		result = addLinkIfTransactionAllowed(VorgangStatus.NEU, REL_VORGANG_ZURUECKHOLEN, vorgang, result);
		result = addLinkIfTransactionAllowed(VorgangStatus.IN_BEARBEITUNG, this::getRelationForInBearbeitungByStatus, vorgang, result);
		result = addLinkIfTransactionAllowed(VorgangStatus.BESCHIEDEN, REL_VORGANG_BESCHEIDEN, vorgang, result);
		result = addLinkIfTransactionAllowed(VorgangStatus.ABGESCHLOSSEN, REL_VORGANG_ABSCHLIESSEN, vorgang, result);

		return result;
	}

	private LinkRelation getRelationForAngenommenByStatus(Vorgang vorgang) {
		return vorgang.getStatus() == VorgangStatus.NEU ? REL_VORGANG_ANNEHMEN : REL_VORGANG_ZURUECKSTELLEN;
	}

	private LinkRelation getRelationForInBearbeitungByStatus(Vorgang vorgang) {
		if (vorgang.getStatus() == VorgangStatus.ANGENOMMEN) {
			return REL_VORGANG_BEARBEITEN;
		} else {
			return REL_VORGANG_WIEDEREROEFFNEN;
		}
	}

	private EntityModel<? extends Vorgang> addLinkIfTransactionAllowed(VorgangStatus newStatus, Function<Vorgang, LinkRelation> getRelation,
			Vorgang vorgang, EntityModel<? extends Vorgang> model) {
		return addLinkIfTransactionAllowed(newStatus, getRelation.apply(vorgang), vorgang, model);
	}

	private EntityModel<? extends Vorgang> addLinkIfTransactionAllowed(VorgangStatus newStatus, LinkRelation relation,
			Vorgang vorgang, EntityModel<? extends Vorgang> model) {
		return addLinkIfCondition(v -> vorgang.getStatus().isTransactionAllowed(getUserRole(), newStatus), relation, vorgang, model);
	}

	String getUserRole() {
		return isEinheitlicherAnsprechpartner() ? UserRole.EINHEITLICHER_ANSPRECHPARTNER : UserRole.VERWALTUNG_USER;
	}

	private boolean isEinheitlicherAnsprechpartner() {
		return userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER);
	}

	private EntityModel<? extends Vorgang> addLinkIfCondition(Predicate<Void> condition, LinkRelation relation, Vorgang vorgang,
			EntityModel<? extends Vorgang> model) {
		if (condition.test(null)) {
			return addLink(model, relation, vorgang);
		}
		return model;
	}

	private EntityModel<? extends Vorgang> addLink(EntityModel<? extends Vorgang> model, LinkRelation relation, Vorgang vorgang) {
		return model.add(buildCreateCommandLink(vorgang).withRel(relation));
	}

	private LinkBuilder buildCreateCommandLink(Vorgang vorgang) {
		return linkTo(
				methodOn(CommandByRelationController.class).createCommandByRelation(vorgang.getId(), vorgang.getId(), vorgang.getVersion(), null));
	}

}