package de.itvsh.goofy;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.time.LocalDateTime;
import java.time.ZoneOffset;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.boot.info.BuildProperties;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.itvsh.goofy.common.user.CurrentUserService;
import de.itvsh.goofy.common.user.UserRole;

class RootControllerTest {

	private final String PATH = "/api";

	@InjectMocks // NOSONAR
	private RootController controller;
	@Mock
	private BuildProperties properties;
	@Mock
	private CurrentUserService userService;

	private MockMvc mockMvc;

	@BeforeEach
	void initTest() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Test
	void shouldHaveSelfLink() {
		var model = controller.getRootResource();

		assertThat(model.getLink(IanaLinkRelations.SELF)).isPresent();
	}

	@Test
	void shouldHaveLinkToVorgangList() {
		var model = controller.getRootResource();

		assertThat(model.getLink(RootController.REL_VORGAENGE)).isPresent().get()
				.extracting(Link::getHref).isEqualTo("/api/vorgangs?page=0");
	}

	@Nested
	class TestLinkForSearch {

		@DisplayName("should exists on user with role EINHEITLICHER_ANSPRECHPARTNER")
		@Test
		void shouldExist() {
			when(userService.hasRole(anyString())).thenReturn(true);

			var model = controller.getRootResource();

			assertThat(model.getLink(RootController.REL_SEARCH)).isPresent().get()
					.extracting(Link::getHref).isEqualTo("/api/vorgangs?page=0{&searchBy}");
		}

		@DisplayName("should NOT exists on user without role EINHEITLICHER_ANSPRECHPARTNER")
		@Test
		void shouldNotExist() {
			when(userService.hasRole(anyString())).thenReturn(false);

			var model = controller.getRootResource();

			assertThat(model.getLink(RootController.REL_SEARCH)).isNotPresent();
		}
	}

	@Nested
	class TestLinkForProfileSearch {

		@BeforeEach
		void init() {
			when(userService.hasRole(anyString())).thenReturn(false);
		}

		@DisplayName("should exists on user with role VERWALTUNG_USER")
		@Test
		void shouldExist() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);

			var model = controller.getRootResource();

			assertThat(model.getLink(RootController.REL_SEARCH_USER)).isPresent().get().extracting(Link::getHref)
					.isEqualTo("/api/userProfiles?searchBy={searchBy}");
		}
	}

	@Test
	void shouldHaveJavaVersion() throws Exception {
		callEndpoint().andExpect(jsonPath("$.javaVersion").exists());
	}

	@Test
	void versionExists() throws Exception {
		when(properties.getVersion()).thenReturn("42");

		callEndpoint().andExpect(jsonPath("$.version").value("42"));
	}

	@Test
	void buildTimeExists() throws Exception {
		when(properties.getTime()).thenReturn(LocalDateTime.parse("2021-04-01T10:30").toInstant(ZoneOffset.UTC));

		callEndpoint().andExpect(jsonPath("$.buildTime").exists());
	}

	private ResultActions callEndpoint() throws Exception {
		return mockMvc.perform(get(PATH)).andExpect(status().isOk());
	}
}