package de.itvsh.goofy.binary;

import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.itvsh.goofy.common.file.OzgFile;
import de.itvsh.goofy.common.file.OzgFileData;
import de.itvsh.goofy.common.file.OzgFileDataTestFactory;
import de.itvsh.goofy.common.file.OzgFileTestFactory;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;

class BinaryFileControllerTest {

	private final String PATH = BinaryFileController.PATH;

	@InjectMocks
	private BinaryFileController controller;
	@Mock
	private BinaryFileService service;
	@Mock
	private BinaryFileModelAssembler assembler;
	@Mock
	private FileIdMapper fileIdMapper;

	private MockMvc mockMvc;

	@BeforeEach
	void initTest() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Nested
	class TestUploadFile {

		@BeforeEach
		void mockServiec() {
			when(service.uploadFile(any(), any(), any())).thenReturn(BinaryFileTestFactory.FILE_ID);
		}

		@Test
		void shouldCallService() throws Exception {
			callEndpoint();

			verify(service).uploadFile(anyString(), anyString(), any(OzgFileData.class));

		}

		private ResultActions callEndpoint() throws Exception {
			return mockMvc.perform(multipart(PATH + "/" + VorgangHeaderTestFactory.ID + "/" + BinaryFileTestFactory.FIELD_NAME + "/file")
					.file(BinaryFileTestFactory.TEST_FILE))
					.andExpect(status().isCreated())
					.andExpect(header().string("Location", "http://localhost/api/binaryFiles/" + BinaryFileTestFactory.FILE_ID));
		}
	}

	@Nested
	class TestGetFile {

		private OzgFile ozgFile = OzgFileTestFactory.create();

		@BeforeEach
		void mockServiec() {
			when(service.getFile(any())).thenReturn(ozgFile);
		}

		@Test
		void shouldCallService() throws Exception {
			callEndpoint();

			verify(service).getFile(BinaryFileTestFactory.FILE_ID);
		}

		@Test
		void shouldCallAssembler() throws Exception {
			callEndpoint();

			verify(assembler).toModel(ozgFile);
		}

		private ResultActions callEndpoint() throws Exception {
			return mockMvc.perform(get(BinaryFileController.PATH + '/' + BinaryFileTestFactory.FILE_ID)
					.contentType(MediaType.APPLICATION_JSON_VALUE))
					.andExpect(status().isOk());
		}
	}

	@Nested
	class TestGetFiles {

		private OzgFile file = OzgFileTestFactory.create();

		@BeforeEach
		void mockServiec() {
			when(service.getFiles(any())).thenReturn(Stream.of(file));
			when(fileIdMapper.toString(any())).then(fileId -> fileId.toString());
		}

		@Test
		void shouldCallService() throws Exception {
			callEndpoint();

			verify(service).getFiles(any());
		}

		@Test
		void shouldCallAssembler() throws Exception {
			callEndpoint();

			verify(assembler).toCollectionModel(Stream.of(file).collect(Collectors.toList()));
		}

		private CollectionModel<EntityModel<OzgFile>> callEndpoint() throws Exception {
			return controller.getFiles(List.of(BinaryFileTestFactory.FILE_ID));
		}
	}

	@Nested
	class TestGetFileData {

		@BeforeEach
		void mockServiceReturnValue() {
			when(service.getFileData(any())).thenReturn(OzgFileDataTestFactory.create());
		}

		@Test
		void shouldCallEndpoint() throws Exception {
			callEndpoint();
		}

		@Test
		void shouldCallService() throws Exception {
			callEndpoint();

			verify(service).getFileData(OzgFileTestFactory.ID);
		}

		@Nested
		class TestHeader {

			ResultActions result = null;

			@BeforeEach
			void doCall() throws Exception {
				result = callEndpoint();
			}

			@Test
			void shouldHaveContentDisposition() throws Exception {
				result.andExpect(header().string(HttpHeaders.CONTENT_DISPOSITION, "attachment; filename=" + OzgFileDataTestFactory.NAME));
			}

			@Test
			void shouldHaveContentType() throws Exception {
				result.andExpect(header().string(HttpHeaders.CONTENT_TYPE, OzgFileDataTestFactory.CONTENT_TYPE_STR));
			}

			@Test
			void shouldHaveContentLength() throws Exception {
				result.andExpect(header().longValue(HttpHeaders.CONTENT_LENGTH, OzgFileDataTestFactory.SIZE));
			}
		}

		private ResultActions callEndpoint() throws Exception {
			return mockMvc.perform(get(PATH + "/" + OzgFileTestFactory.ID).contentType(MediaType.APPLICATION_OCTET_STREAM_VALUE))
					.andExpect(status().isOk());
		}
	}
}