package de.itvsh.goofy.common.command;

import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;

import de.itvsh.goofy.common.ValidationMessageCodes;
import de.itvsh.goofy.vorgang.RedirectRequest;
import de.itvsh.goofy.vorgang.RedirectRequestTestFactory;

@AutoConfigureMockMvc
@SpringBootTest
@WithMockUser
class CommandControllerITCase {

	@Autowired
	private MockMvc mockMvc;

	@WithMockUser
	@Nested
	class TestValidation {

		@WithMockUser
		@Nested
		class TestEmailValidation {

			private final String FIELD = "email";

			@Test
			void shouldReturnErrorOnNullEMail() throws Exception {
				String content = buildRedirectRequestWithEmail(null);

				doRequest(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.redirectRequest." + FIELD))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_IS_EMPTY));
			}

			@Test
			void shouldReturnErrorOnInvalidEMail() throws Exception {
				String content = buildRedirectRequestWithEmail("local@@domain.com");

				doRequest(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.redirectRequest." + FIELD))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_INVALID));
			}

			private String buildRedirectRequestWithEmail(String eMail) {
				return buildRedirectRequest(RedirectRequestTestFactory.createBuilder().email(eMail).build());
			}
		}

		@WithMockUser
		@Nested
		class TestPasswordValidation {

			private final String FIELD = "password";

			@Test
			void shouldReturnErrorOnNullPassword() throws Exception {
				String content = buildRedirectRequestWithPassword(null);

				doRequest(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.redirectRequest." + FIELD))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_IS_EMPTY));
			}

			@Test
			void shouldReturnErrorOnToShortPassword() throws Exception {
				String content = buildRedirectRequestWithPassword("1234567");

				doRequest(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.redirectRequest." + FIELD))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_SIZE));
			}

			private String buildRedirectRequestWithPassword(String password) {
				return buildRedirectRequest(RedirectRequestTestFactory.createBuilder().password(password).build());
			}
		}

		private String buildRedirectRequest(RedirectRequest request) {
			return CommandTestFactory.createRedirectRequestContent(CommandTestFactory.createCreateCommandBuilder().redirectRequest(request).build());
		}

		private ResultActions doRequest(String content) throws Exception {
			return mockMvc.perform(post("/api/vorgangs/" + CommandTestFactory.VORGANG_ID + "/relations/" + CommandTestFactory.RELATION_ID + "/"
					+ CommandTestFactory.VERSION + "/commands")
							.contentType(MediaType.APPLICATION_JSON)
							.content(content));
		}
	}

}