package de.itvsh.goofy.common.command;

import static de.itvsh.goofy.common.command.CommandController.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.itvsh.goofy.TestUtils;
import de.itvsh.goofy.common.command.CommandController.CommandByRelationController;
import de.itvsh.goofy.common.errorhandling.ExceptionController;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;

class CommandControllerTest {

	@InjectMocks // NOSONAR
	private CommandController controller;
	@Mock
	private CommandService service;
	@Mock
	private CommandModelAssembler modelAssembler;

	private MockMvc mockMvc;

	@BeforeEach
	void initTest() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller)
				.setControllerAdvice(new ExceptionController())
				.build();
	}

	@Nested
	class TestGetCommand {

		private Command command = CommandTestFactory.create();

		@BeforeEach
		void initTest() {
			when(service.getById(anyString())).thenReturn(command);
		}

		@Test
		void shouldReturnOk() throws Exception {
			doRequest().andExpect(status().isOk());
		}

		@Test
		void shouldCallService() throws Exception {
			doRequest();

			verify(service).getById(CommandTestFactory.ID);
		}

		@Test
		void shouldCallModelAssembler() throws Exception {
			doRequest();

			verify(modelAssembler).toModel(command);
		}

		private ResultActions doRequest() throws Exception {
			return mockMvc.perform(get(CommandController.COMMANDS_PATH + "/" + CommandTestFactory.ID)).andExpect(status().is2xxSuccessful());
		}
	}

	@Nested
	@DisplayName("Revoke Command")
	class TestRevokeCommand {

		@Test
		void shouldReturnOk() throws Exception {
			when(service.revoke(anyString())).thenReturn(CommandTestFactory.create());

			doRequest(CommandStatus.REVOKED).andExpect(status().isOk());
		}

		@Test
		void shouldCallService() throws Exception {
			when(service.revoke(anyString())).thenReturn(CommandTestFactory.create());

			doRequest(CommandStatus.REVOKED);

			verify(service).revoke(CommandTestFactory.ID);
		}

		@Test
		void shouldReturnForbiddenForOtherStatus() throws Exception {
			doRequest(CommandStatus.FINISHED).andExpect(status().isForbidden());
		}

		private ResultActions doRequest(CommandStatus statusToPatch) throws Exception {
			return mockMvc.perform(patch(COMMANDS_PATH + "/" + CommandTestFactory.ID).contentType(MediaType.APPLICATION_JSON)
					.content(TestUtils.loadTextFile("jsonTemplates/command/patchStatus.json", statusToPatch.name())));

		}
	}

	@Nested
	class TestCommandByRelationController {

		@InjectMocks // NOSONAR
		private CommandByRelationController controller;
		@Mock
		private CommandService service;
		@Mock
		private CommandModelAssembler modelAssembler;

		private MockMvc mockMvc;

		@BeforeEach
		void initTest() {
			mockMvc = MockMvcBuilders.standaloneSetup(controller).build();

			lenient().when(service.createCommand(anyString(), anyString(), anyLong(), any())).thenReturn(CommandTestFactory.create());
		}

		@Nested
		@DisplayName("Create Command by Vorgang")
		class TestCreateCommand {

			@Captor // NOSONAR
			private ArgumentCaptor<CreateCommand> createCommandCaptor;

			@DisplayName("should return CREATED")
			@Test
			void returnCreated() throws Exception {
				doRequest().andExpect(status().isCreated());
			}

			@DisplayName("should call service")
			@Test
			void callService() throws Exception {
				doRequest();

				verify(service).createCommand(eq(VorgangHeaderTestFactory.ID), eq(CommandTestFactory.RELATION_ID),
						eq(VorgangHeaderTestFactory.VERSION),
						createCommandCaptor.capture());
				assertThat(createCommandCaptor.getValue().getOrder()).isEqualTo(CommandTestFactory.ORDER);
			}

			@DisplayName("should have location header")
			@Test
			void locationHeader() throws Exception {
				doRequest()
						.andExpect(header().stringValues("location", "http://localhost" + COMMANDS_PATH + "/" + CommandTestFactory.ID));
			}

			private ResultActions doRequest() throws Exception {
				return mockMvc
						.perform(post(CommandByRelationController.COMMAND_BY_RELATION_AND_VERSION_PATH, VorgangHeaderTestFactory.ID,
								CommandTestFactory.RELATION_ID, VorgangHeaderTestFactory.VERSION)
										.contentType(MediaType.APPLICATION_JSON)
										.content(TestUtils.loadTextFile("jsonTemplates/command/createVorgangCommand.json.tmpl",
												CommandOrder.VORGANG_ANNEHMEN.name())))
						.andExpect(status().is2xxSuccessful());
			}
		}
	}

	@Nested
	class TestExistsPendingCommands {

		@Test
		void shouldCallService() {
			controller.existsPendingCommands(VorgangHeaderTestFactory.ID);

			verify(service).existsPendingCommands(VorgangHeaderTestFactory.ID);
		}
	}

	@Nested
	class TestGetPendingCommands {

		@Nested
		class TestGetCommand {

			private final Command command = CommandTestFactory.create();
			private final Stream<Command> pendingCommands = Stream.of(command);

			@BeforeEach
			void initTest() {
				when(service.getPendingCommands(anyString())).thenReturn(pendingCommands);
			}

			@Test
			void shouldReturnOk() throws Exception {
				doRequest().andExpect(status().isOk());
			}

			@Test
			void shouldCallService() throws Exception {
				doRequest();

				verify(service).getPendingCommands(VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldCallModelAssembler() throws Exception {
				doRequest();

				verify(modelAssembler).toCollectionModel(pendingCommands);
			}

			private ResultActions doRequest() throws Exception {
				return mockMvc.perform(get(CommandController.COMMANDS_PATH)
						.param(CommandController.PARAM_PENDING, Boolean.toString(true))
						.param(CommandController.PARAM_VORGANG_ID, VorgangHeaderTestFactory.ID))
						.andExpect(status().is2xxSuccessful());
			}
		}
	}

}