package de.itvsh.goofy.common.command;

import static de.itvsh.goofy.common.command.CommandModelAssembler.*;
import static de.itvsh.goofy.common.command.CommandTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.InjectMocks;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;

import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;

class CommandModelAssemblerTest {

	@InjectMocks // NOSONAR
	private CommandModelAssembler modelAssembler;

	private final String COMMAND_SINGLE_PATH = "/api/commands/" + CommandTestFactory.ID;

	@Test
	void shouldAddSelfLink() {
		var model = modelAssembler.toModel(CommandTestFactory.create());

		assertThat(model.getLink(IanaLinkRelations.SELF)).isPresent().get()
				.extracting(Link::getHref).isEqualTo(COMMAND_SINGLE_PATH);
	}

	@Test
	void shouldAddRevokeLink() {
		var model = modelAssembler.toModel(CommandTestFactory.create());

		assertThat(model.getLink(CommandModelAssembler.REL_REVOKE)).isPresent().get()
				.extracting(Link::getHref).isEqualTo(COMMAND_SINGLE_PATH);
	}

	@Nested
	@DisplayName("Link to effected Resource")
	class TestEffectedResourceLink {

		@Nested
		@DisplayName("vorgang")
		class TestEffectedResourceIsVorgang {

			private final String VORGANG_URL = "/api/vorgangs/" + RELATION_ID;

			@Test
			void shouldBePresentWhenDone() {
				var model = assembleModelWithCommandInStatus(CommandStatus.FINISHED);

				assertThat(model.getLink(REL_EFFECTED_RESOURCE)).isPresent().get().extracting(Link::getHref).isEqualTo(VORGANG_URL);
			}

			@Test
			void shouldBePresentWhenRevoked() {
				var model = assembleModelWithCommandInStatus(CommandStatus.REVOKED);

				assertThat(model.getLink(REL_EFFECTED_RESOURCE)).isPresent().get().extracting(Link::getHref).isEqualTo(VORGANG_URL);
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.EXCLUDE, names = { "FINISHED", "REVOKED", "ERROR" })
			void shouldNOTbePresentOtherWise(CommandStatus status) {
				var model = assembleModelWithCommandInStatus(status);

				assertThat(model.getLink(CommandModelAssembler.REL_EFFECTED_RESOURCE)).isEmpty();
			}

			@Test
			void shouldPointToForwarding() {
				var model = modelAssembler
						.toModel(CommandTestFactory.createBuilder().status(CommandStatus.FINISHED).order(CommandOrder.FORWARD_SUCCESSFULL).build());

				assertThat(model.getLink(CommandModelAssembler.REL_EFFECTED_RESOURCE)).isPresent().get()
						.extracting(Link::getHref).isEqualTo("/api/forwardings?vorgangId=" + VorgangHeaderTestFactory.ID);
			}
		}

		@Nested
		@DisplayName("wiedervorlage")
		class TestEffectedResourceIsWiedervorlage {

			private final String WIEDERVORLAGE_URL = "/api/wiedervorlages/" + RELATION_ID;

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "CREATE_WIEDERVORLAGE", "EDIT_WIEDERVORLAGE",
					"WIEDERVORLAGE_ERLEDIGEN", "WIEDERVORLAGE_WIEDEREROEFFNEN" })
			void shouldLinkToWiedervorlage(CommandOrder order) {
				var model = modelAssembler.toModel(
						CommandTestFactory.createBuilder().status(CommandStatus.FINISHED).order(CommandOrder.WIEDERVORLAGE_ERLEDIGEN).build());

				assertThat(model.getLink(CommandModelAssembler.REL_EFFECTED_RESOURCE)).isPresent().get().extracting(Link::getHref)
						.isEqualTo(WIEDERVORLAGE_URL);
			}
		}

		private EntityModel<Command> assembleModelWithCommandInStatus(CommandStatus status) {
			return modelAssembler.toModel(CommandTestFactory.createBuilder().status(status).build());
		}
	}

	@Nested
	class TestRevokeLink {

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "CREATE_WIEDERVORLAGE", "WIEDERVORLAGE_ERLEDIGEN", "EDIT_WIEDERVORLAGE",
				"WIEDERVORLAGE_WIEDEREROEFFNEN", "CREATE_KOMMENTAR", "EDIT_KOMMENTAR", "REDIRECT_VORGANG", "FORWARD_SUCCESSFULL",
				"FORWARD_FAILED", "ASSIGN_USER" })
		void shouldHaveLink(CommandOrder order) {
			var model = modelAssembler.toModel(CommandTestFactory.createBuilder().order(order).build());

			assertThat(model.getLink(REL_REVOKE)).isPresent().get().extracting(Link::getHref).isEqualTo(COMMAND_SINGLE_PATH);
		}

		@ParameterizedTest
		@EnumSource(names = { "CREATE_WIEDERVORLAGE", "WIEDERVORLAGE_ERLEDIGEN", "EDIT_WIEDERVORLAGE",
				"WIEDERVORLAGE_WIEDEREROEFFNEN", "CREATE_KOMMENTAR", "EDIT_KOMMENTAR", "REDIRECT_VORGANG", "FORWARD_SUCCESSFULL" })
		void shouldNotHaveLink(CommandOrder order) {
			var model = modelAssembler.toModel(CommandTestFactory.createBuilder().order(order).build());

			assertThat(model.getLink(REL_REVOKE)).isNotPresent();
		}
	}

	@Nested
	class TestUpdateLink {

		@ParameterizedTest
		@EnumSource(names = { "PENDING", "REVOKE_PENDING" })
		void shouldHave(CommandStatus status) {
			var model = modelAssembler.toModel(CommandTestFactory.createBuilder().status(status).build());

			assertThat(model.getLink(REL_UPDATE)).isPresent().get().extracting(Link::getHref).isEqualTo(COMMAND_SINGLE_PATH);
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "PENDING", "REVOKE_PENDING" })
		void shouldHaveLink(CommandStatus status) {
			var model = modelAssembler.toModel(CommandTestFactory.createBuilder().status(status).build());

			assertThat(model.getLink(REL_UPDATE)).isNotPresent();
		}
	}
}