package de.itvsh.goofy.common.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.itvsh.goofy.common.ContextService;
import de.itvsh.goofy.vorgang.RedirectRequest;
import de.itvsh.goofy.vorgang.RedirectRequestTestFactory;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;
import de.itvsh.ozg.pluto.grpc.command.CommandServiceGrpc.CommandServiceBlockingStub;
import de.itvsh.ozg.pluto.grpc.command.GrpcCallContext;
import de.itvsh.ozg.pluto.grpc.command.GrpcCommand;
import de.itvsh.ozg.pluto.grpc.command.GrpcCommandResponse;
import de.itvsh.ozg.pluto.grpc.command.GrpcCommandsResponse;
import de.itvsh.ozg.pluto.grpc.command.GrpcCreateCommandRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcExistsPendingCommandsRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcExistsPendingCommandsResponse;
import de.itvsh.ozg.pluto.grpc.command.GrpcFindCommandsRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcGetCommandRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcGetPendingCommandsRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcGetPendingCommandsResponse;
import de.itvsh.ozg.pluto.grpc.command.GrpcRedirectRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcRevokeCommandRequest;

class CommandRemoteServiceTest {

	@Spy
	@InjectMocks
	private CommandRemoteService service;
	@Mock
	private CommandServiceBlockingStub commandServiceStub;
	@Mock
	private CommandMapper mapper;
	@Mock
	private CommandBodyMapper bodyMapper;
	@Mock
	private ContextService contextService;

	private GrpcCallContext grpcCallContext = GrpcCallContext.newBuilder().build();

	@Nested
	@DisplayName("Create CreateCommand")
	class TestCreateCommand {

		private GrpcCreateCommandRequest request = GrpcCreateCommandRequest.newBuilder().build();
		private CreateCommand createCommand = CommandTestFactory.createCreateCommand();

		@Nested
		@DisplayName("call grpc endpoint")
		class TestCallGrpcEndpoint {
			private Command resultCommand = CommandTestFactory.create();

			@BeforeEach
			void initTest() {
				doReturn(request).when(service).createCreateCommandRequest(any(), anyLong());
				when(commandServiceStub.createCommand(any())).thenReturn(GrpcCommandResponse.newBuilder().build());
				when(mapper.toCommand(any())).thenReturn(resultCommand);
			}

			@Test
			void shouldCreateRequest() {

				service.createCommand(createCommand, CommandTestFactory.VERSION);

				verify(service).createCreateCommandRequest(createCommand, CommandTestFactory.VERSION);
			}

			@Test
			void shouldCallServiceStub() {

				service.createCommand(createCommand, CommandTestFactory.VERSION);

				verify(commandServiceStub).createCommand(request);
			}

			@Test
			void shouldReturnResultCommand() {

				var result = service.createCommand(createCommand, CommandTestFactory.VERSION);

				assertThat(result).isSameAs(resultCommand);
			}
		}

		@Nested
		@DisplayName("Create Request")
		class TestCreateRequest {

			private GrpcRedirectRequest grpcRedirectRequest = GrpcRedirectRequest.newBuilder().build();

			@BeforeEach
			void initTest() {
				when(contextService.createCallContext()).thenReturn(grpcCallContext);
			}

			@Test
			void shouldCreateRequest() {
				var request = service.createCreateCommandRequest(CommandTestFactory.createCreateCommand(), CommandTestFactory.VERSION);

				assertThat(request).isNotNull();
			}

			@Test
			void shouldHaveOrder() {
				var request = service.createCreateCommandRequest(CommandTestFactory.createCreateCommand(), CommandTestFactory.VERSION);

				assertThat(request.getOrder().name()).isEqualTo(CommandTestFactory.ORDER.name());
			}

			@Test
			void shouldHaveCallContext() {
				var request = service.createCreateCommandRequest(CommandTestFactory.createCreateCommand(), CommandTestFactory.VERSION);

				assertThat(request.getCallContext()).isSameAs(grpcCallContext);
			}

			@Test
			void shouldHaveRelationId() {
				var request = service.createCreateCommandRequest(CommandTestFactory.createCreateCommand(), CommandTestFactory.VERSION);

				assertThat(request.getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
			}

			@Test
			void shouldHaveVorgangId() {
				var request = service.createCreateCommandRequest(CommandTestFactory.createCreateCommand(), CommandTestFactory.VERSION);

				assertThat(request.getVorgangId()).isEqualTo(CommandTestFactory.VORGANG_ID);
			}

			@Test
			void shouldHaveVersion() {
				var request = service.createCreateCommandRequest(CommandTestFactory.createCreateCommand(), CommandTestFactory.VERSION);

				assertThat(request.getRelationVersion()).isEqualTo(VorgangHeaderTestFactory.VERSION);
			}

			@Test
			void shouldHaveRedirectRequest() {
				when(mapper.toGrpcRedirectRequest(any())).thenReturn(grpcRedirectRequest);
				RedirectRequest redirectRequest = RedirectRequestTestFactory.create();
				var command = CommandTestFactory.createCreateCommandBuilder().redirectRequest(redirectRequest).build();

				var request = service.createCreateCommandRequest(command, CommandTestFactory.VERSION);

				assertThat(request.getRedirectRequest()).isNotNull().isSameAs(grpcRedirectRequest);
			}

			@ParameterizedTest
			@EnumSource()
			void shouldHaveOrder(CommandOrder order) {
				var request = service.createCreateCommandRequest(CommandTestFactory.createCreateCommandBuilder().order(order).build(),
						CommandTestFactory.VERSION);

				assertThat(request.getOrder()).hasToString(order.name());
			}
		}
	}

	@Nested
	@DisplayName("Revoke Command")
	class TestRevokeCommand {
		@DisplayName("create request")
		@Nested
		class TestCreateRequest {

			@BeforeEach
			void initTest() {
				when(contextService.createCallContext()).thenReturn(grpcCallContext);
			}

			@Test
			void shouldReturnRequest() {
				var request = service.createRevokeRequest(CommandTestFactory.ID);

				assertThat(request).isNotNull().isInstanceOf(GrpcRevokeCommandRequest.class);
			}

			@Test
			void shouldHaveCallContext() {
				var request = service.createRevokeRequest(CommandTestFactory.ID);

				assertThat(request.getContext()).isNotNull().isSameAs(grpcCallContext);
				verify(contextService).createCallContext();
			}

			@Test
			void shouldHaveCommandId() {
				var request = service.createRevokeRequest(CommandTestFactory.ID);

				assertThat(request.getId()).isEqualTo(CommandTestFactory.ID);
			}
		}

		@DisplayName("send request")
		@Nested
		class TestSendRequest {

			private GrpcRevokeCommandRequest request = GrpcRevokeCommandRequest.newBuilder().build();

			@BeforeEach
			void initTest() {
				doReturn(request).when(service).createRevokeRequest(anyString());
				when(commandServiceStub.revokeCommand(any())).thenReturn(GrpcCommandResponse.newBuilder().build());
			}

			@Test
			void shouldCreateRequest() {
				service.revokeCommand(CommandTestFactory.ID);

				verify(service).createRevokeRequest(CommandTestFactory.ID);
			}

			@Test
			void shouldCallGrpc() {
				service.revokeCommand(CommandTestFactory.ID);

				verify(commandServiceStub).revokeCommand(request);
			}

			@Test
			void shouldMapCommand() {
				service.revokeCommand(CommandTestFactory.ID);

				verify(mapper).toCommand(any());
			}
		}
	}

	@Nested
	@DisplayName("get command")
	class TestGetCommand {
		@DisplayName("create Request")
		@Nested
		class TestCreateRequest {

			@BeforeEach
			void initTest() {
				when(contextService.createCallContext()).thenReturn(grpcCallContext);
			}

			@Test
			void shouldReturnRequest() {
				var request = service.createGetCommandRequest(CommandTestFactory.ID);

				assertThat(request).isNotNull().isInstanceOf(GrpcGetCommandRequest.class);
			}

			@Test
			void shouldHaveCallContext() {
				var request = service.createGetCommandRequest(CommandTestFactory.ID);

				assertThat(request.getContext()).isNotNull().isSameAs(grpcCallContext);
				verify(contextService).createCallContext();
			}

			@Test
			void shouldHaveCommandId() {
				var request = service.createGetCommandRequest(CommandTestFactory.ID);

				assertThat(request.getId()).isEqualTo(CommandTestFactory.ID);
			}
		}

		@DisplayName("send request")
		@Nested
		class TestSendRequest {

			private GrpcGetCommandRequest request = GrpcGetCommandRequest.newBuilder().build();

			@BeforeEach
			void initTest() {
				doReturn(request).when(service).createGetCommandRequest(anyString());
				when(commandServiceStub.getCommand(any())).thenReturn(GrpcCommand.newBuilder().build());
			}

			@Test
			void shouldCreateRequest() {
				service.getCommand(CommandTestFactory.ID);

				verify(service).createGetCommandRequest(CommandTestFactory.ID);
			}

			@Test
			void shouldCallGrpc() {
				service.getCommand(CommandTestFactory.ID);

				verify(commandServiceStub).getCommand(request);
			}

			@Test
			void shouldMapCommand() {
				service.getCommand(CommandTestFactory.ID);

				verify(mapper).toCommand(any());
			}
		}
	}

	@Nested
	class TestExistsPendingCommands {

		private final GrpcExistsPendingCommandsRequest request = GrpcExistsPendingCommandsRequest.newBuilder()
				.setContext(grpcCallContext)
				.setVorgangId(VorgangHeaderTestFactory.ID).build();

		@BeforeEach
		void mockServiceStub() {
			when(commandServiceStub.existsPendingCommands(any()))
					.thenReturn(GrpcExistsPendingCommandsResponse.newBuilder().setExistsPendingCommands(true).build());

			when(contextService.createCallContext()).thenReturn(grpcCallContext);
		}

		@Test
		void shouldCallServiceStub() {
			service.existsPendingCommands(VorgangHeaderTestFactory.ID);

			verify(commandServiceStub).existsPendingCommands(request);
		}

		@Test
		void shouldCallContextService() {
			service.existsPendingCommands(VorgangHeaderTestFactory.ID);

			verify(contextService).createCallContext();
		}

		@Test
		void shouldReturnValue() {
			var result = service.existsPendingCommands(VorgangHeaderTestFactory.ID);

			assertThat(result).isTrue();
		}
	}

	@Nested
	class TestFindCommands {

		private final GrpcCommand grpcCommand = GrpcCommand.newBuilder().build();
		private final GrpcCommandsResponse response = GrpcCommandsResponse.newBuilder().addAllCommand(Collections.singleton(grpcCommand)).build();

		@Captor // NOSONAR
		private ArgumentCaptor<GrpcFindCommandsRequest> requestCaptor;

		@BeforeEach
		void init() {
			when(commandServiceStub.findCommands(any())).thenReturn(response);
			when(contextService.createCallContext()).thenReturn(grpcCallContext);
		}

		@Test
		void shouldCallServiceStub() {
			service.findCommands(VorgangHeaderTestFactory.ID, Optional.empty(), Optional.empty());

			verify(commandServiceStub).findCommands(notNull());
		}

		@Test
		void shouldRequestWithVorgangId() {
			service.findCommands(VorgangHeaderTestFactory.ID, Optional.empty(), Optional.empty());

			verify(commandServiceStub).findCommands(requestCaptor.capture());

			assertThat(requestCaptor.getValue().getVorgangId()).isEqualTo(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldRequestWithStatus() {
			service.findCommands(VorgangHeaderTestFactory.ID, Optional.of(CommandStatus.FINISHED), Optional.empty());

			verify(commandServiceStub).findCommands(requestCaptor.capture());

			assertThat(requestCaptor.getValue().getStatusList()).contains(CommandStatus.FINISHED.name());
		}

		@Test
		void shouldRequestWithOrder() {
			service.findCommands(VorgangHeaderTestFactory.ID, Optional.empty(), Optional.of(CommandOrder.REDIRECT_VORGANG));

			verify(commandServiceStub).findCommands(requestCaptor.capture());

			assertThat(requestCaptor.getValue().getOrder().name()).isEqualTo(CommandOrder.REDIRECT_VORGANG.name());
		}

		@Test
		void shouldCallMapper() {
			service.findCommands(VorgangHeaderTestFactory.ID, Optional.empty(), Optional.empty()).forEach(command -> {
			});

			verify(mapper).toCommand(grpcCommand);
		}
	}

	@Nested
	class TestGetPendingCommands {

		private final GrpcGetPendingCommandsRequest request = GrpcGetPendingCommandsRequest.newBuilder()
				.setContext(grpcCallContext)
				.setVorgangId(VorgangHeaderTestFactory.ID).build();
		private final GrpcCommand grpcCommand = GrpcCommand.newBuilder().build();
		private final Set<GrpcCommand> commandList = Collections.singleton(grpcCommand);
		private final GrpcGetPendingCommandsResponse response = GrpcGetPendingCommandsResponse.newBuilder().addAllCommand(commandList).build();
		private final Command command = CommandTestFactory.create();

		@BeforeEach
		void mockServiceStub() {
			when(commandServiceStub.getPendingCommands(any())).thenReturn(response);
			when(contextService.createCallContext()).thenReturn(grpcCallContext);
		}

		@Test
		void shouldCallServiceStub() {
			service.getPendingCommands(VorgangHeaderTestFactory.ID);

			verify(commandServiceStub).getPendingCommands(request);
		}

		@Test
		void shouldCallContextService() {
			service.getPendingCommands(VorgangHeaderTestFactory.ID);

			verify(contextService).createCallContext();
		}

		@Test
		void shouldCallMapper() {
			var result = service.getPendingCommands(VorgangHeaderTestFactory.ID);
			forceStreamOperationsDone(result);

			verify(mapper).toCommand(grpcCommand);
		}

		@Test
		void shouldReturnValue() {
			when(mapper.toCommand(any())).thenReturn(command);

			var result = service.getPendingCommands(VorgangHeaderTestFactory.ID);

			assertThat(result.collect(Collectors.toList()).get(0)).isEqualTo(command);
		}

		private void forceStreamOperationsDone(Stream<Command> stream) {
			stream.collect(Collectors.toList()).stream();
		}
	}
}