package de.itvsh.goofy.kommentar;

import static de.itvsh.goofy.kommentar.KommentarCommandTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.itvsh.goofy.common.command.CommandOrder;
import de.itvsh.goofy.kommentar.KommentarCommandController.KommentarCommandByVorgangController;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;

class KommentarCommandControllerTest {

	@Captor
	private ArgumentCaptor<KommentarCommand> commandCaptor;

	@Nested
	class TestCreateKommentarCommandByVorgang {

		@InjectMocks
		private KommentarCommandByVorgangController controller;
		@Mock
		private KommentarService service;

		private MockMvc mockMvc;

		@BeforeEach
		void init() {
			mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
			when(service.createCommand(anyString(), any())).thenReturn(KommentarCommandTestFactory.create());
		}

		@Test
		void shouldReturnCreated() throws Exception {
			doRequest().andExpect(status().isCreated());
		}

		@Test
		void shouldCallService() throws Exception {
			doRequest();

			verify(service).createCommand(eq(VorgangHeaderTestFactory.ID), notNull());
		}

		@Test
		void shouldGiveCommandToService() throws Exception {
			doRequest();

			verify(service).createCommand(eq(VorgangHeaderTestFactory.ID), commandCaptor.capture());
			var command = commandCaptor.getValue();

			assertThat(command).usingRecursiveComparison()
					.ignoringFields("kommentar.id", "kommentar.createdBy", "kommentar.createdByName", "kommentar.createdAt")
					.isEqualTo(KommentarCommandTestFactory.createBuilder().id(null).build());
		}

		private ResultActions doRequest() throws Exception {
			return mockMvc.perform(post(KommentarCommandByVorgangController.KOMMENTAR_COMMANDS_BY_VORGANG, VorgangHeaderTestFactory.ID)
					.content(createValidRequestContent()).contentType(MediaType.APPLICATION_JSON))
					.andExpect(status().is2xxSuccessful());
		}
	}

	@Nested
	class TestCreateEditKommentarCommand {

		final String REPONSE_HEADER = "http://localhost/api/commands/" + KommentarCommandTestFactory.ID;

		@InjectMocks
		private KommentarCommandController controller;
		@Mock
		private KommentarService service;

		private MockMvc mockMvc;

		@BeforeEach
		void init() {
			mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
			when(service.createCommand(anyString(), any())).thenReturn(KommentarCommandTestFactory.createBuilder()
					.order(CommandOrder.EDIT_KOMMENTAR).build());
		}

		@Test
		void shouldReturnCreated() throws Exception {
			doRequest().andExpect(status().isCreated());
		}

		@Test
		void shouldCallService() throws Exception {
			doRequest();

			verify(service).createCommand(anyString(), notNull());
		}

		@Test
		void shouldGiveCommandToService() throws Exception {
			doRequest();

			verify(service).createCommand(anyString(), commandCaptor.capture());
			var command = commandCaptor.getValue();

			assertThat(command).usingRecursiveComparison()
					.ignoringFields("kommentar.id", "kommentar.createdBy", "kommentar.createdByName", "kommentar.createdAt")
					.isEqualTo(KommentarCommandTestFactory.createBuilder().id(null).build());
		}

		@Test
		void shouldRespondWithLocationHeader() throws Exception {
			doRequest().andExpect(header().string("Location", REPONSE_HEADER));
		}

		ResultActions doRequest() throws Exception {
			String content = createValidRequestContent();
			return mockMvc.perform(post(KommentarCommandController.KOMMENTAR_COMMANDS, KommentarTestFactory.ID)
					.content(content).contentType(MediaType.APPLICATION_JSON))
					.andExpect(status().is2xxSuccessful());
		}
	}
}
