package de.itvsh.goofy.kommentar;

import static de.itvsh.goofy.kommentar.KommentarCommandTestFactory.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;

import de.itvsh.goofy.common.ValidationMessageCodes;
import de.itvsh.goofy.common.command.CommandOrder;
import de.itvsh.goofy.kommentar.KommentarCommandController.KommentarCommandByVorgangController;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;

@AutoConfigureMockMvc
@SpringBootTest
@WithMockUser
class KommentarCommandITCase {

	@Autowired
	private MockMvc mockMvc;
	@MockBean
	private KommentarRemoteService remoteService;

	@WithMockUser
	@Nested
	class TestCreateCommandByKommentarId {

		@BeforeEach
		void initTest() {
			when(remoteService.sendKommentarCommand(anyString(), any())).thenReturn(KommentarCommandTestFactory.create());
		}

		@Test
		void shouldCreateCommand() throws Exception {
			doRequestByKommentarId(createValidRequestContent()).andExpect(status().isCreated());

			verify(remoteService).sendKommentarCommand(eq(KommentarTestFactory.ID), any());
		}

		@WithMockUser
		@DisplayName("should return validation error")
		@Nested
		class TestValidation {

			@DisplayName("for null Text")
			@Test
			void createCommandWithInvalidText() throws Exception {
				String content = buildContentWithText(null);

				doRequestByKommentarId(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.kommentar.text"))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_IS_EMPTY));
			}

			@DisplayName("for empty String in Text")
			@Test
			void createcommandWithShortText() throws Exception {
				String content = buildContentWithText(StringUtils.EMPTY);

				doRequestByKommentarId(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.[0].field").value("command.kommentar.text"));

			}

			@DisplayName("for invalid text should have parameter")
			@Test
			void minMaxParameter() throws Exception {
				String content = buildContentWithText(StringUtils.EMPTY);

				doRequestByKommentarId(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues[0].parameters.length()").value(2));
			}

			private String buildContentWithText(String text) {
				return createRequestContent(KommentarCommandTestFactory.createBuilder()
						.order(CommandOrder.EDIT_KOMMENTAR)
						.kommentar(KommentarTestFactory.createBuilder().text(text).build())
						.build());
			}
		}

		private ResultActions doRequestByKommentarId(String content) throws Exception {
			return mockMvc.perform(post(KommentarCommandController.KOMMENTAR_COMMANDS, KommentarTestFactory.ID)
					.contentType(MediaType.APPLICATION_JSON)
					.content(content));
		}
	}

	@WithMockUser
	@Nested
	class TestCreateCommandByVorgangId {

		@BeforeEach
		void initTest() {
			when(remoteService.sendKommentarCommand(anyString(), any())).thenReturn(KommentarCommandTestFactory.create());
		}

		@Test
		void shouldCreateCommand() throws Exception {
			doRequestByVorgangId(createValidRequestContent()).andExpect(status().isCreated());

			verify(remoteService).sendKommentarCommand(eq(VorgangHeaderTestFactory.ID), any());
		}

		@WithMockUser
		@DisplayName("should return validation error")
		@Nested
		class TestValidation {

			@DisplayName("for null Text")
			@Test
			void createCommandWithInvalidText() throws Exception {
				String content = buildContentWithText(null);

				doRequestByVorgangId(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.kommentar.text"))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_IS_EMPTY));
			}

			@DisplayName("for empty String in Text")
			@Test
			void createcommandWithShortText() throws Exception {
				String content = buildContentWithText(StringUtils.EMPTY);

				doRequestByVorgangId(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.[0].field").value("command.kommentar.text"));

			}

			@DisplayName("for invalid text should have parameter")
			@Test
			void minMaxParameter() throws Exception {
				String content = buildContentWithText(StringUtils.EMPTY);

				doRequestByVorgangId(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues[0].parameters.length()").value(2));
			}

			private String buildContentWithText(String text) {
				return createRequestContent(KommentarCommandTestFactory.createBuilder()
						.order(CommandOrder.CREATE_KOMMENTAR)
						.kommentar(KommentarTestFactory.createBuilder().text(text).build())
						.build());
			}
		}

		private ResultActions doRequestByVorgangId(String content) throws Exception {
			return mockMvc.perform(post(KommentarCommandByVorgangController.KOMMENTAR_COMMANDS_BY_VORGANG, VorgangHeaderTestFactory.ID)
					.contentType(MediaType.APPLICATION_JSON)
					.content(content));
		}
	}
}