package de.itvsh.goofy.kommentar;

import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.itvsh.goofy.common.ContextService;
import de.itvsh.goofy.common.command.GrpcCommandResponseTestFactory;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;
import de.itvsh.ozg.pluto.grpc.command.GrpcCallContext;
import de.itvsh.ozg.pluto.grpc.command.GrpcCommandResponse;
import de.itvsh.ozg.pluto.kommentar.GrpcFindKommentarsByVorgangIdRequest;
import de.itvsh.ozg.pluto.kommentar.GrpcFindKommentarsByVorgangIdResponse;
import de.itvsh.ozg.pluto.kommentar.GrpcGetKommentarRequest;
import de.itvsh.ozg.pluto.kommentar.GrpcGetKommentarResponse;
import de.itvsh.ozg.pluto.kommentar.GrpcKommentar;
import de.itvsh.ozg.pluto.kommentar.KommentarServiceGrpc.KommentarServiceBlockingStub;

class KommentarRemoteServiceTest {

	@Spy
	@InjectMocks
	private KommentarRemoteService service;

	@Mock
	private KommentarMapper mapper;
	@Mock
	private KommentarServiceBlockingStub serviceStub;
	@Mock
	private ContextService contextService;

	private GrpcKommentar kommentar = GrpcKommentarTestFactory.create();

	@Nested
	class TestFindByVorgangId {

		private GrpcFindKommentarsByVorgangIdRequest request = GrpcFindKommentarsByVorgangIdRequest.newBuilder()
				.setVorgangId(VorgangHeaderTestFactory.ID).build();
		private GrpcFindKommentarsByVorgangIdResponse response = GrpcFindKommentarsByVorgangIdResponse.newBuilder()
				.addKommentars(kommentar)
				.addKommentars(kommentar).build();

		@BeforeEach
		void mockStub() {
			when(serviceStub.findKommentarsByVorgangId(any())).thenReturn(response);
		}

		@Test
		void shouldCallStub() {
			service.findByVorgangId(VorgangHeaderTestFactory.ID);

			verify(serviceStub).findKommentarsByVorgangId(request);
		}

		@Test
		void shouldCallMapper() {
			Stream<Kommentar> result = service.findByVorgangId(VorgangHeaderTestFactory.ID);
			collectStreamElementsToTriggerLazyStream(result);

			verify(mapper, times(2)).toKommentar(kommentar);
		}

		private void collectStreamElementsToTriggerLazyStream(Stream<Kommentar> stream) {
			stream.collect(Collectors.toList());
		}
	}

	@Nested
	class TestGetById {

		private GrpcGetKommentarRequest request = GrpcGetKommentarRequest.newBuilder().setId(KommentarTestFactory.ID).build();
		private GrpcGetKommentarResponse response = GrpcGetKommentarResponse.newBuilder().setKommentar(kommentar).build();

		@BeforeEach
		void mockStub() {
			when(serviceStub.getKommentar(any())).thenReturn(response);
		}

		@Test
		void shouldCallStub() {
			service.getById(KommentarTestFactory.ID);

			verify(serviceStub).getKommentar(request);
		}

		@Test
		void shouldCallMapper() {
			service.getById(KommentarTestFactory.ID);

			verify(mapper).toKommentar(kommentar);
		}
	}

	@Nested
	class TestSendKommentarCommand {

		private GrpcCallContext callContext = GrpcCallContext.newBuilder().build();
		private GrpcCommandResponse response = GrpcCommandResponseTestFactory.createBuilder().build();

		@BeforeEach
		void initTest() {
			when(serviceStub.createCommand(any())).thenReturn(response);
			when(contextService.createCallContext()).thenReturn(callContext);
			when(mapper.toGrpcKommentar(any())).thenReturn(GrpcKommentarTestFactory.create());
			when(mapper.toKommentarCommand(any())).thenReturn(KommentarCommandTestFactory.create());
		}

		@Test
		@DisplayName("Should call buildCreateCommandRequest with VorgangId")
		void shouldCallCreateRequestWithVorgangId() {
			doRequest();

			verify(service).buildCreateCommandRequest(eq(VorgangHeaderTestFactory.ID), any());
		}

		@Test
		void shouldCallStub() {
			doRequest();

			verify(serviceStub).createCommand(any());
		}

		@Test
		void shouldMapResult() {
			doRequest();

			verify(mapper).toKommentarCommand(any());
		}

		private void doRequest() {
			service.sendKommentarCommand(VorgangHeaderTestFactory.ID, KommentarCommandTestFactory.create());
		}
	}
}
