package de.itvsh.goofy.vorgang;

import static de.itvsh.goofy.vorgang.GrpcSubFormTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.itvsh.ozg.pluto.vorgang.GrpcEingang;
import de.itvsh.ozg.pluto.vorgang.GrpcSubForm;

class EingangMapperTest {

	@InjectMocks
	private EingangMapper mapper = Mappers.getMapper(EingangMapper.class);

	@Mock
	private AntragstellerMapper antragstellerMapper;
	@Mock
	private EingangHeaderMapper eingangHeaderMapper;

	@Nested
	class TestGrpcEingangToGoofyEingang {

		private final Antragsteller antragsteller = AntragstellerTestFactory.create();
		private final EingangHeader eingangHeader = EingangHeaderTestFactory.create();

		@BeforeEach
		void mockMapperReturnValues() {
			lenient().when(antragstellerMapper.toAntragsteller(any())).thenReturn(antragsteller);
			lenient().when(eingangHeaderMapper.toEingangHeader(any())).thenReturn(eingangHeader);
		}

		@Test
		void shouldMapIdToId() {
			GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getId()).isEqualTo(GrpcEingangTestFactory.ID);
		}

		@Test
		void shouldHaveAntragsteller() {
			GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getAntragsteller()).isEqualTo(antragsteller);
		}

		@Test
		void shouldHaveEingangHeader() {
			GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getHeader()).isEqualTo(eingangHeader);
		}

		@Test
		void shouldProceedWithEmptyValues() {
			GrpcEingang grpcEingang = GrpcEingang.newBuilder().build();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getAntragsteller()).isNull();
			assertThat(eingang.getHeader()).isNull();
		}

		@DisplayName("Mapped SubFormField")
		@Nested
		class TestSubFormFieldMapping {

			private final GrpcSubForm subForm = GrpcSubFormTestFactory.create();

			@Test
			void shouldContainField() {
				var mapped = mapper.mapSubFormFields(subForm.getFieldList());

				assertThat(mapped).containsKey(FIELD_NAME);
			}

			@Test
			void shouldContainValue() {
				var mapped = mapper.mapSubFormFields(subForm.getFieldList());

				assertThat(mapped).containsEntry(FIELD_NAME, FIELD_VALUE);
			}
		}

		@DisplayName("Mapped FormData")
		@Nested
		class TestMapSubFormToGoofySubForm {

			@Test
			void shouldNotBeEmpty() {
				GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

				var eingang = mapper.fromGrpc(grpcEingang);

				assertThat(eingang.getFormData()).isNotEmpty();
			}

			@Test
			void shouldHaveMappedSubForm() {
				GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

				var eingang = mapper.fromGrpc(grpcEingang);

				assertThat(eingang.getFormData()).containsKey(TITLE);
			}

			@DisplayName("mapped FormData should contain fields")
			@Test
			void fieldsInSubForm() {
				GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

				var eingang = mapper.fromGrpc(grpcEingang);
				@SuppressWarnings("unchecked")
				Map<String, Object> formData = (Map<String, Object>) eingang.getFormData().get(TITLE);

				assertThat(formData).containsEntry(FIELD_NAME, FIELD_VALUE);
			}

			@DisplayName("mapped FormData should contains subForm")
			@Test
			void shouldHaveMappedSubFormWithSubForm() {
				GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

				var eingang = mapper.fromGrpc(grpcEingang);
				@SuppressWarnings("unchecked")
				Map<String, Object> formData = (Map<String, Object>) eingang.getFormData().get(TITLE);

				assertThat(formData).containsKey(GrpcSubFormTestFactory.SUBFORM_NAME);
			}

			@DisplayName("mapped SubForm should contain field")
			@Test
			void subFormShouldContainField() {
				GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

				var eingang = mapper.fromGrpc(grpcEingang);
				@SuppressWarnings("unchecked")
				Map<String, Object> subForm = (Map<String, Object>) ((Map<String, Object>) eingang.getFormData().get(TITLE)).get(SUBFORM_NAME);

				assertThat(subForm).containsEntry(SUBFORM_FIELD_NAME, SUBFORM_FIELD_VALUE);
			}
		}
	}
}