package de.itvsh.goofy.vorgang;

import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;

import de.itvsh.goofy.common.command.CommandController;
import de.itvsh.goofy.common.errorhandling.ResourceNotFoundException;
import de.itvsh.goofy.common.user.UserProfileTestFactory;
import de.itvsh.goofy.vorgang.forwarding.ForwardingController;

@AutoConfigureMockMvc
@SpringBootTest
@WithMockUser
class VorgangControllerITCase {

	public static final String SINGLE_PATH = VorgangController.PATH + "/{id}";

	@MockBean
	private VorgangRemoteService service;

	@MockBean
	private CommandController commandController;

	@MockBean
	private ForwardingController forwardingController;

	@Autowired
	private MockMvc mockMvc;

	@WithMockUser
	@Nested
	class GetSingleVorgang {

		@BeforeEach()
		void mockCommandController() {
			when(commandController.existsPendingCommands(any())).thenReturn(true);
		}

		@Test
		void shouldFormatDateTime() throws Exception {
			when(service.findVorgangWithEingang(VorgangHeaderTestFactory.ID)).thenReturn(VorgangWithEingangTestFactory.create());

			doRequest().andExpect(status().isOk())
					.andExpect(jsonPath("$.createdAt").value(VorgangHeaderTestFactory.CREATED_AT_STR));
		}

		@Test
		void shouldReturnNotFound() throws Exception {
			when(service.findVorgangWithEingang(any())).thenThrow(new ResourceNotFoundException("ResourceNotFoundMessage"));

			mockMvc.perform(get(SINGLE_PATH, VorgangHeaderTestFactory.ID)).andExpect(status().isNotFound());
		}

		@Test
		void shouldHaveAnnehmenLink() throws Exception {
			when(service.findVorgangWithEingang(anyString())).thenReturn(VorgangWithEingangTestFactory.create());

			doRequest().andExpect(jsonPath("$._links.annehmen.href")
					.value("http://localhost/api/vorgangs/" + VorgangHeaderTestFactory.ID + "/relations/" + VorgangHeaderTestFactory.ID + "/"
							+ VorgangHeaderTestFactory.VERSION + "/commands"));
		}

		@WithMockUser
		@Nested
		class LinkToAssignedUser {
			@Test
			void shouldBePresentIfAssigned() throws Exception {
				when(service.findVorgangWithEingang(anyString())).thenReturn(VorgangWithEingangTestFactory.create());

				doRequest().andExpect(jsonPath("$._links.assignedTo.href").value("http://localhost/api/userProfiles/" + UserProfileTestFactory.ID));
			}

			@Test
			void shouldNOTBePresentIfNOSTAssigned() throws Exception {
				when(service.findVorgangWithEingang(anyString())).thenReturn(VorgangWithEingangTestFactory.createBuilder().assignedTo(null).build());

				doRequest().andExpect(jsonPath("$._links.assignedTo.href").doesNotExist());
			}
		}

		private ResultActions doRequest() throws Exception {
			return mockMvc.perform(get(SINGLE_PATH, VorgangHeaderTestFactory.ID)).andExpect(status().is2xxSuccessful());
		}
	}
}