package de.itvsh.goofy.vorgang;

import static de.itvsh.goofy.vorgang.EingangTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.hateoas.EntityModel;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.itvsh.goofy.wiedervorlage.WiedervorlageController;
import de.itvsh.goofy.wiedervorlage.WiedervorlageTestFactory;

class VorgangControllerTest {

	private final String PATH = VorgangController.PATH + "/";

	@InjectMocks // NOSONAR
	private VorgangController controller;

	@Mock
	private VorgangService vorgangService;
	@Mock
	private VorgangModelAssembler modelAssembler;
	@Mock
	private WiedervorlageController wiedervorlageController;

	private MockMvc mockMvc;

	@BeforeEach
	void initTest() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Nested
	class TestVorgangListByPage {

		final int page = 5;

		@BeforeEach
		void initTest() {
			when(vorgangService.findVorgangHeader(anyInt(), anyInt(), any())).thenReturn(VorgangListResponseTestFactory.create());
		}

		@Test
		void shouldCallService() throws Exception {
			callEndpoint();

			verify(vorgangService).findVorgangHeader(VorgangController.PAGE_SIZE, 500, Optional.empty());
		}

		@Test
		void shouldCallServiceWithDefaultPage() throws Exception {
			mockMvc.perform(get(PATH)).andExpect(status().isOk());

			verify(vorgangService).findVorgangHeader(VorgangController.PAGE_SIZE, 0, Optional.empty());
		}

		@Test
		void shouldCallModelAssembler() throws Exception {
			callEndpoint();

			verify(modelAssembler).toCollectionModel(any(), any(), any());
		}

		private void callEndpoint() throws Exception {
			mockMvc.perform(get(PATH).param(VorgangController.PARAM_PAGE, Integer.toString(page))).andExpect(status().isOk());
		}
	}

	@Nested
	class TestSearchVorgangList {

		static final String SEARCH_STRING = "Looking for freedom";

		@BeforeEach
		void initTest() {
			when(vorgangService.findVorgangHeader(anyInt(), anyInt(), any())).thenReturn(VorgangListResponseTestFactory.create());
		}

		@Test
		void shouldReturnOk() throws Exception {
			doRequest().andExpect(status().isOk());
		}

		@Test
		void shouldCallService() throws Exception {
			doRequest();

			verify(vorgangService).findVorgangHeader(anyInt(), anyInt(), eq(Optional.of(SEARCH_STRING)));
		}

		private ResultActions doRequest() throws Exception {
			return mockMvc.perform(get(PATH).param(VorgangController.PARAM_SEARCH, SEARCH_STRING))
					.andExpect(status().is2xxSuccessful());
		}
	}

	@Nested
	class AddNextFrist {

		@Test
		void shouldCallWiedervorlageController() {
			controller.addNextFrist(VorgangHeaderTestFactory.create());

			verify(wiedervorlageController).findWithNextFrist(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldSetNextFrist() {
			when(wiedervorlageController.findWithNextFrist(anyString())).thenReturn(Optional.of(WiedervorlageTestFactory.create()));

			var result = controller.addNextFrist(VorgangHeaderTestFactory.create());

			assertThat(result.getNextFrist()).isPresent().get().isEqualTo(WiedervorlageTestFactory.FRIST);
		}

		@Test
		void shouldByFindWithNoWiedervorlage() {
			when(wiedervorlageController.findWithNextFrist(anyString())).thenReturn(Optional.empty());

			var result = controller.addNextFrist(VorgangHeaderTestFactory.create());

			assertThat(result.getNextFrist()).isEmpty();
		}
	}

	@Nested
	class TestVorgangWithEingang {

		final int page = 0;

		@BeforeEach
		void initTest() {
			when(vorgangService.findVorgangWithEingang(any())).thenReturn(VorgangWithEingangTestFactory.create());
			when(modelAssembler.toModel(any())).then(i -> EntityModel.of(i.getArgument(0)));
		}

		@Test
		void shouldCallService() throws Exception {
			performRequest();

			verify(vorgangService).findVorgangWithEingang(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldCallModelAssembler() throws Exception {
			performRequest();

			verify(modelAssembler).toModel(any());
		}

		@Nested
		class TestDataMapping {

			@DisplayName("Single Field should be in json first level")
			@Test
			void singleField() throws Exception {
				performRequest().andExpect(jsonPath("$.eingang.formData." + SINGLE_FIELD_NAME).value(SINGLE_FIELD_VALUE));
			}

			@DisplayName("Numeric field should be numeric in json")
			@Test
			void numericField() throws Exception {
				performRequest().andExpect(jsonPath("$.eingang.formData." + NUMERIC_FIELD_NAME).isNumber());
				performRequest().andExpect(jsonPath("$.eingang.formData." + NUMERIC_FIELD_NAME).value(NUMERIC_FIELD_VALUE));
			}

			@DisplayName("SubForm should have name as key")
			@Test
			void keyOfsubForm() throws Exception {
				performRequest().andExpect(jsonPath("$.eingang.formData." + SUBFORM_NAME).isMap());
			}

			@DisplayName("SubForm-Field should be present")
			@Test
			void subFormField() throws Exception {
				performRequest().andExpect(jsonPath("$.eingang.formData." + SUBFORM_NAME + "." + SUBFORM_FIELD_NAME).value(SUBFORM_FIELD_VALUE));
			}

		}

		private ResultActions performRequest() throws Exception {
			return mockMvc.perform(get(PATH + "/" + VorgangHeaderTestFactory.ID)).andExpect(status().isOk());
		}
	}

	@Nested
	class TestHasNextPage {

		private final int pageSize = 10;
		private VorgangListResponse response;

		@BeforeEach
		void createVorgangList() {
			response = VorgangListResponse.builder().total(28).vorgaengeHeader(VorgangListResponseTestFactory.createList()).build();
		}

		@Test
		void firstPage() {
			int page = 0;

			Optional<Integer> nextPage = controller.hasNextPage(page, response, pageSize);

			assertThat(nextPage).isPresent().contains(1);
		}

		@Test
		void secondPage() {
			int page = 1;

			Optional<Integer> nextPage = controller.hasNextPage(page, response, pageSize);

			assertThat(nextPage).isPresent().contains(2);
		}

		@Test
		void lastPage() {
			int page = 2;

			Optional<Integer> nextPage = controller.hasNextPage(page, response, pageSize);

			assertThat(nextPage).isNotPresent();
		}
	}
}