package de.itvsh.goofy.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;
import org.springframework.security.test.context.support.WithMockUser;

import de.itvsh.goofy.common.command.CommandController;
import de.itvsh.goofy.common.user.CurrentUserService;
import de.itvsh.goofy.common.user.UserRole;

class VorgangModelAssemblerTest {

	private final static String BASE_PATH = VorgangController.PATH;

	@InjectMocks // NOSONAR
	private VorgangModelAssembler modelAssembler;

	@Mock
	private CommandController commandController;
	@Mock
	private CurrentUserService userService;

	@Nested
	class TestCollectionModel {

		@Test
		void shouldHaveSize() throws Exception {
			assertThat(toCollectionModel(null, null).getContent()).hasSize(1);
		}

		@Test
		void shouldHavePreviousLink() {
			var link = toCollectionModel(1, null).getLink(VorgangModelAssembler.REL_PREVIOUS);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo(BASE_PATH + "?page=1");
		}

		@Test
		void shouldHaveNextLink() {
			var link = toCollectionModel(0, 2).getLink(VorgangModelAssembler.REL_NEXT);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo(BASE_PATH + "?page=2");
		}

		private CollectionModel<EntityModel<Vorgang>> toCollectionModel(Integer prevPage, Integer nextPage) {
			return modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create()), //
					Optional.ofNullable(prevPage),
					Optional.ofNullable(nextPage));
		}
	}

	@Nested
	class TestLinksOnModel {

		@DisplayName("Test links at model with user role 'EINHEITLICHER_ANSPRECHPARTNER'")
		@Nested
		class TestWithRole {

			@BeforeEach
			void init() {
				when(userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(true);
			}

			@Test
			void shouldHaveSelfLink() {
				var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(IanaLinkRelations.SELF_VALUE);

				assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(BASE_PATH + "/" + VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldHaveVorgangMitEingangLink() {
				var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_VORGANG_MIT_EINGANG);

				assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(BASE_PATH + "/" + VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldNotHaveWiedervorlagenLink() {
				Optional<Link> model = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_WIEDERVORLAGEN);

				assertThat(model).isNotPresent();
			}

			@Test
			void shouldNotHaveCreateWiedervorlageLink() {
				Optional<Link> model = modelAssembler.toModel(VorgangHeaderTestFactory.create())
						.getLink(VorgangModelAssembler.REL_CREATE_WIEDERVORLAGE);

				assertThat(model).isNotPresent();
			}

			@Test
			void shouldNotHaveBearbeiterLink() {
				Optional<Link> model = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

				assertThat(model).isNotPresent();
			}
		}

		@DisplayName("Test links at model WITHOUT any role")
		@Nested
		class TestWithoutRole {

			@Test
			void shouldHaveSelfLink() {
				var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(IanaLinkRelations.SELF_VALUE);

				assertThat(link).isPresent();
				assertThat(link.get().getHref()).isEqualTo(BASE_PATH + "/" + VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldHaveVorgangMitEingangLink() {
				var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_VORGANG_MIT_EINGANG);

				assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(BASE_PATH + "/" + VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldHaveWiedervorlagenLink() {
				var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_WIEDERVORLAGEN);

				assertThat(link).isPresent();
				assertThat(link.get().getHref()).isEqualTo("/api/wiedervorlages?vorgangId=" + VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldHaveCreateWiedervorlageLink() {
				var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_CREATE_WIEDERVORLAGE);

				assertThat(link).isPresent().get().extracting(Link::getHref)
						.isEqualTo("/api/vorgangs/" + VorgangHeaderTestFactory.ID + "/wiedervorlageCommands");
			}

			@Test
			void shouldHaveAssignLink() {
				var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

				assertThat(link).isPresent().get().extracting(Link::getHref)
						.isEqualTo(String.format("/api/vorgangs/%s/relations/%s/%s/commands", VorgangHeaderTestFactory.ID,
								VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.VERSION));
			}

			@Test
			void shouldHaveKommentareLink() {
				var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_KOMMENTARE);

				assertThat(link).isPresent();
				assertThat(link.get().getHref()).isEqualTo(BASE_PATH + "/" + VorgangHeaderTestFactory.ID + "/kommentars");
			}
		}
	}

	@Nested
	class TestPendingCommandsLink {

		private final String EXPECTED_HREF = "/api/commands?pending=true&vorgangId=" + VorgangHeaderTestFactory.ID;

		@Test
		void shouldExists() {
			when(commandController.existsPendingCommands(any())).thenReturn(true);

			var link = modelAssembler.toModel(VorgangWithEingangTestFactory.create()).getLink(VorgangModelAssembler.REL_PENDING_COMMANDS);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo(EXPECTED_HREF);
		}

		@Test
		void shouldNotExists() {
			when(commandController.existsPendingCommands(any())).thenReturn(false);

			assertThat(modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_PENDING_COMMANDS)).isNotPresent();
		}

		@DisplayName("should have link with user role " + UserRole.EINHEITLICHER_ANSPRECHPARTNER)
		@WithMockUser(roles = UserRole.EINHEITLICHER_ANSPRECHPARTNER)
		void shouldExistsByRoleEA() {
			when(commandController.existsPendingCommands(any())).thenReturn(false);

			var link = modelAssembler.toModel(VorgangHeaderTestFactory.create()).getLink(VorgangModelAssembler.REL_PENDING_COMMANDS);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo(EXPECTED_HREF);
		}
	}
}