package de.itvsh.goofy.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.LinkRelation;

import de.itvsh.goofy.common.command.CommandController;
import de.itvsh.goofy.common.user.CurrentUserService;
import de.itvsh.goofy.common.user.UserRole;
import de.itvsh.goofy.vorgang.Vorgang;
import de.itvsh.goofy.vorgang.Vorgang.VorgangStatus;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;
import de.itvsh.goofy.vorgang.VorgangWithEingang;
import de.itvsh.goofy.vorgang.VorgangWithEingangTestFactory;
import de.itvsh.goofy.vorgang.forwarding.ForwardingController;
import de.itvsh.goofy.vorgang.forwarding.ForwardingTestFactory;

class VorgangWithEingangCommandProzessorTest {

	@Spy // NOSONAR
	@InjectMocks
	private VorgangWithEingangCommandProzessor processor;
	@Mock
	private CurrentUserService userService;
	@Mock
	private CommandController commandController;
	@Mock
	private ForwardingController forwardingController;

	private VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
	private EntityModel<VorgangWithEingang> vorgangEntityModel = EntityModel.of(vorgang);

	@Test
	void shouldReturnEntityModel() {
		var result = processor.process(vorgangEntityModel);

		assertThat(result).isNotNull();
	}

	@Nested
	class TestAddForwardLink {

		private VorgangWithEingang vorgang = VorgangWithEingangTestFactory.createBuilder().status(VorgangStatus.IN_BEARBEITUNG).build();

		@BeforeEach
		void init() {
			when(userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(true);
		}

		@DisplayName("should be true for Role EINHEITLICHER_ANSPRECHPARTNER, Status IN_BEARBEITUNG and Forwarding with Error")
		@Test
		void shouldBeTrue() {
			when(forwardingController.findFailedForwardings(anyString())).thenReturn(Collections.singletonList(ForwardingTestFactory.create()));

			var addLink = processor.addForwardLink(vorgang);

			assertThat(addLink).isTrue();
		}

		@Test
		void shouldBeTrueInStatusNeu() {
			var vorgang = VorgangWithEingangTestFactory.createBuilder().status(VorgangStatus.NEU).build();

			var addLink = processor.addForwardLink(vorgang);

			assertThat(addLink).isTrue();
		}

		@Test
		void shouldBeFalseForRoleNotEA() {
			when(userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(false);

			var addLink = processor.addForwardLink(vorgang);

			assertThat(addLink).isFalse();
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "NEU", "IN_BEARBEITUNG" })
		void shouldBeFalseForVorgangInOtherStatus(VorgangStatus status) {
			var vorgang = VorgangWithEingangTestFactory.createBuilder().status(status).build();

			var addLink = processor.addForwardLink(vorgang);

			assertThat(addLink).isFalse();
		}

		@Test
		void shouldBeFalseIfNoFailedForwarding() {
			when(forwardingController.findFailedForwardings(anyString())).thenReturn(Collections.emptyList());

			var addLink = processor.addForwardLink(vorgang);

			assertThat(addLink).isFalse();
		}

		@Test
		void shouldBeFalseOnStatusWeitergeleitet() {
			VorgangWithEingang vorgang = VorgangWithEingangTestFactory.createBuilder().status(VorgangStatus.WEITERGELEITET).build();

			var addLink = processor.addForwardLink(vorgang);

			assertThat(addLink).isFalse();
		}

	}

	@Nested
	@DisplayName("add 'forward' link")
	class TestForwardLink {

		private final LinkRelation linkRel = VorgangWithEingangCommandProzessor.REL_VORGANG_FORWARD;

		@Test
		void shouldNotBePresent() {
			doReturn(false).when(processor).addForwardLink(any());

			var processed = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()));

			assertThat(processed.getLink(linkRel)).isEmpty();
		}

		@Test
		void shouldBePresent() {
			doReturn(true).when(processor).addForwardLink(any());

			var processed = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()));

			assertThat(processed.getLink(linkRel)).isPresent().get()
					.extracting(Link::getHref)
					.isEqualTo("/api/vorgangs/" + VorgangHeaderTestFactory.ID + "/relations/" + VorgangHeaderTestFactory.ID + "/"
							+ VorgangHeaderTestFactory.VERSION + "/commands");
		}

	}

	@Nested
	class TestForwardingLink {
		@DisplayName("should NOT be present in other Role then EA")
		@ParameterizedTest
		@EnumSource
		void shouldNotBePresentWithoutRole(Vorgang.VorgangStatus status) {
			when(userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(false);

			var processed = processor.process(buildVorgangInStatus(status));

			assertThat(processed.getLink(VorgangWithEingangCommandProzessor.REL_VORGANG_FORWARDING)).isEmpty();
		}

		@DisplayName("with role EinheitlicherAnsprechpartner")
		@Nested
		class TestWithRoleEinheitlicherAnsprechpartner {
			@BeforeEach
			void init() {
				when(userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(true);
			}

			@DisplayName("should be present in any Status")
			@ParameterizedTest
			@EnumSource()
			void shouldBePresent(VorgangStatus status) {
				var processed = processor.process(buildVorgangInStatus(status));

				assertThat(processed.getLink(VorgangWithEingangCommandProzessor.REL_VORGANG_FORWARDING)).isPresent()
						.get().extracting(Link::getHref)
						.isEqualTo("/api/forwardings?vorgangId=" + VorgangHeaderTestFactory.ID);
			}
		}
	}

	private EntityModel<VorgangWithEingang> buildVorgangInStatus(VorgangStatus status) {
		return EntityModel.of(VorgangWithEingangTestFactory.createBuilder().status(status).build());
	}
}
