package de.itvsh.goofy.wiedervorlage;

import static de.itvsh.goofy.wiedervorlage.WiedervorlageCommandTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.itvsh.goofy.common.command.CommandOrder;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;
import de.itvsh.goofy.wiedervorlage.WiedervorlageCommandController.WiedervorlageCommandByVorgangController;

class WiedervorlageCommandControllerTest {

	@InjectMocks
	private WiedervorlageCommandController controller;
	@Mock
	private WiedervorlageService service;

	@Captor
	private ArgumentCaptor<WiedervorlageCommand> commandCaptor;

	private MockMvc mockMvc;

	@BeforeEach
	void init() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Nested
	class TestCreateWiedervorlageCommandByVorgang {

		@InjectMocks
		private WiedervorlageCommandByVorgangController controller;
		@Mock
		private WiedervorlageService service;

		private MockMvc mockMvc;

		@BeforeEach
		void init() {
			mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
			when(service.createCreateCommand(anyString(), any())).thenReturn(WiedervorlageCommandTestFactory.create());
		}

		@Test
		void shouldReturnCreated() throws Exception {
			doRequest().andExpect(status().isCreated());
		}

		@Test
		void shouldCallService() throws Exception {
			doRequest();

			verify(service).createCreateCommand(eq(VorgangHeaderTestFactory.ID), notNull());
		}

		@Test
		void shouldGiveCommandToService() throws Exception {
			doRequest();

			verify(service).createCreateCommand(eq(VorgangHeaderTestFactory.ID), commandCaptor.capture());
			var command = commandCaptor.getValue();

			assertThat(command).usingRecursiveComparison()
					.ignoringFields("wiedervorlage.id", "wiedervorlage.createdBy", "wiedervorlage.createdByName", "wiedervorlage.createdAt",
							"wiedervorlage.attachments")
					.isEqualTo(WiedervorlageCommandTestFactory.createBuilder().id(null).build());
		}

		private ResultActions doRequest() throws Exception {
			return mockMvc.perform(post(WiedervorlageCommandByVorgangController.WIEDERVORLAGE_COMMANDS_BY_VORGANG, VorgangHeaderTestFactory.ID)
					.content(createValidRequestContent()).contentType(MediaType.APPLICATION_JSON))
					.andExpect(status().is2xxSuccessful());
		}
	}

	@Nested
	class TestCreateWiedervorlageCommand {

		final String REPONSE_HEADER = "http://localhost/api/commands/" + WiedervorlageCommandTestFactory.ID;

		@Nested
		class TestCreateEditWiedervorlageCommand {

			@BeforeEach
			void init() {
				when(service.createCommand(anyString(), any())).thenReturn(WiedervorlageCommandTestFactory.create());
			}

			@Test
			void shouldReturnCreated() throws Exception {
				doRequest().andExpect(status().isCreated());
			}

			@Test
			void shouldCallService() throws Exception {
				doRequest();

				verify(service).createCommand(anyString(), notNull());
			}

			@Test
			void shouldGiveCommandToService() throws Exception {
				doRequest();

				verify(service).createCommand(anyString(), commandCaptor.capture());
				var command = commandCaptor.getValue();

				assertThat(command).usingRecursiveComparison()
						.ignoringFields("wiedervorlage.id", "wiedervorlage.createdBy", "wiedervorlage.createdByName", "wiedervorlage.createdAt",
								"wiedervorlage.attachments")
						.isEqualTo(WiedervorlageCommandTestFactory.createBuilder().id(null).build());
			}

			@Test
			void shouldResponseWithLocationHeader() throws Exception {
				doRequest().andExpect(header().string("Location", REPONSE_HEADER));
			}

			private ResultActions doRequest() throws Exception {
				return TestCreateWiedervorlageCommand.this.doRequest(createValidRequestContent());
			}
		}

		@Nested
		class TestCreateWiedervorlageCommandByOrder {

			@BeforeEach
			void init() {
				when(service.createCommand(anyString(), any())).thenReturn(WiedervorlageCommandTestFactory.create());
			}

			@ParameterizedTest
			@EnumSource(names = { "WIEDERVORLAGE_ERLEDIGEN", "WIEDERVORLAGE_WIEDEREROEFFNEN" })
			void shouldReturnCreated(CommandOrder order) throws Exception {
				doRequest(order).andExpect(status().isCreated());
			}

			@ParameterizedTest
			@EnumSource(names = { "WIEDERVORLAGE_ERLEDIGEN", "WIEDERVORLAGE_WIEDEREROEFFNEN" })
			void shouldCallService(CommandOrder order) throws Exception {
				var command = WiedervorlageCommand.builder().order(order).build();

				doRequest(order);

				verify(service).createCommand(WiedervorlageTestFactory.ID, command);
			}

			@ParameterizedTest
			@EnumSource(names = { "WIEDERVORLAGE_ERLEDIGEN", "WIEDERVORLAGE_WIEDEREROEFFNEN" })
			void shouldResponseWithLocationHeader(CommandOrder order) throws Exception {
				doRequest(order).andExpect(header().string("Location", REPONSE_HEADER));

			}

			private ResultActions doRequest(CommandOrder order) throws Exception {
				return TestCreateWiedervorlageCommand.this.doRequest(createRequestContent(order));
			}
		}

		ResultActions doRequest(String content) throws Exception {
			return mockMvc.perform(post(WiedervorlageCommandController.WIEDERVORLAGE_COMMANDS, WiedervorlageTestFactory.ID)
					.content(content).contentType(MediaType.APPLICATION_JSON))
					.andExpect(status().is2xxSuccessful());
		}
	}
}