package de.itvsh.goofy.wiedervorlage;

import static de.itvsh.goofy.wiedervorlage.WiedervorlageCommandTestFactory.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.time.LocalDate;

import org.assertj.core.internal.bytebuddy.utility.RandomString;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;

import de.itvsh.goofy.common.ValidationMessageCodes;

@AutoConfigureMockMvc
@SpringBootTest
@WithMockUser
class WiedervorlageCommandITCase {

	@Autowired
	private MockMvc mockMvc;
	@MockBean
	private WiedervorlageRemoteService remoteService;

	@WithMockUser
	@Nested
	class TestCreateEditCommand {

		@BeforeEach
		void initTest() {
			when(remoteService.sendCreateWiedervorlageCommand(anyString(), any())).thenReturn(WiedervorlageCommandTestFactory.create());
			when(remoteService.sendWiedervorlageCommand(anyString(), any())).thenReturn(WiedervorlageCommandTestFactory.create());
		}

		@Test
		void shouldCreateCommand() throws Exception {
			doRequest(createValidRequestContent()).andExpect(status().isCreated());

			verify(remoteService).sendWiedervorlageCommand(eq(WiedervorlageTestFactory.ID), any());
		}

		@WithMockUser
		@DisplayName("should return validation error")
		@Nested
		class TestValidation {

			@DisplayName("for null Betreff")
			@Test
			void createCommandWithInvalidBetreff() throws Exception {
				String content = buildContentWithBetreff(null);

				doRequest(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.wiedervorlage.betreff"))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_IS_EMPTY));
			}

			@DisplayName("for only 1 character in Betreff")
			@Test
			void createcommandWithShortBetreff() throws Exception {
				String content = buildContentWithBetreff("a");

				doRequest(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.[0].field").value("command.wiedervorlage.betreff"));

			}

			@DisplayName("for invalid betreff should have paramter")
			@Test
			void minMaxParameter() throws Exception {
				String content = buildContentWithBetreff("a");

				doRequest(content).andExpect(status().isUnprocessableEntity()).andExpect(jsonPath("$.issues[0].parameters.length()").value(2));
			}

			@DisplayName("for 41 character in Betreff")
			@Test
			void createCommandWithLongBetreff() throws Exception {
				var content = buildContentWithBetreff(RandomString.make(41));

				doRequest(content).andExpect(status().isUnprocessableEntity());
			}

			private String buildContentWithBetreff(String betreff) {
				return createRequestContent(createWithWiedervorlage(WiedervorlageTestFactory.createBuilder().betreff(betreff).build()));
			}

			@DisplayName("for frist in past")
			@Test
			void createCommandWithInvalidDate() throws Exception {
				String content = createRequestContent(
						createWithWiedervorlage(WiedervorlageTestFactory.createBuilder().frist(LocalDate.parse("2020-01-01")).build()));

				doRequest(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.wiedervorlage.frist"))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_DATE_PAST));
			}

			@DisplayName("for empty frist")
			@Test
			void createCommandWithoutDate() throws Exception {
				String content = createRequestContent(
						createWithWiedervorlage(WiedervorlageTestFactory.createBuilder().frist(null).build()));

				doRequest(content).andExpect(status().isUnprocessableEntity())
						.andExpect(jsonPath("$.issues.length()").value(1))
						.andExpect(jsonPath("$.issues.[0].field").value("command.wiedervorlage.frist"))
						.andExpect(jsonPath("$.issues.[0].messageCode").value(ValidationMessageCodes.FIELD_IS_EMPTY));
			}

		}

		private ResultActions doRequest(String content) throws Exception {
			return mockMvc.perform(post(WiedervorlageCommandController.WIEDERVORLAGE_COMMANDS, WiedervorlageTestFactory.ID)
					.contentType(MediaType.APPLICATION_JSON)
					.content(content));
		}
	}
}