package de.itvsh.goofy.wiedervorlage;

import static org.assertj.core.api.Assertions.*;

import java.util.Collections;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.server.EntityLinks;

import de.itvsh.goofy.common.file.OzgFileTestFactory;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;

class WiedervorlageModelAssemblerTest {

	private final String PATH = WiedervorlageController.WIEDERVORLAGE_PATH + "/";

	@InjectMocks
	private WiedervorlageModelAssembler modelAssembler;

	@Mock
	private EntityLinks entityLinks;

	@Nested
	class TestLinksOnModel {

		final String COMMAND_BY_WIEDERVORLAGE_PATH = //
				WiedervorlageCommandController.WIEDERVORLAGE_COMMANDS.replace("{wiedervorlageId}", WiedervorlageTestFactory.ID);

		@Test
		void shouldHaveSelfLink() {
			var link = toModel().getLink(IanaLinkRelations.SELF);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo(PATH + WiedervorlageTestFactory.ID);
		}

		@Test
		void shouldHaveEditLink() {
			var link = toModel().getLink(WiedervorlageModelAssembler.REL_EDIT);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo(COMMAND_BY_WIEDERVORLAGE_PATH);
		}

		@Nested
		class TestErledigenLink {

			@Test
			void shouldHaveLink() {
				var link = toModel().getLink(WiedervorlageModelAssembler.REL_ERLEDIGEN);

				assertThat(link).isPresent();
				assertThat(link.get().getHref()).isEqualTo(COMMAND_BY_WIEDERVORLAGE_PATH);
			}

			@Test
			void shouldNotHaveLink() {
				EntityModel<Wiedervorlage> entityModel = modelAssembler.toModel(WiedervorlageTestFactory.createBuilder().done(true).build());

				var link = entityModel.getLink(WiedervorlageModelAssembler.REL_ERLEDIGEN);

				assertThat(link).isNotPresent();
			}
		}

		@Nested
		class TestWiedereroeffnenLink {

			@Test
			void shouldHaveLink() {
				EntityModel<Wiedervorlage> entityModel = modelAssembler.toModel(WiedervorlageTestFactory.createBuilder().done(true).build());

				var link = entityModel.getLink(WiedervorlageModelAssembler.REL_WIEDEREROEFFNEN);

				assertThat(link).isPresent();
				assertThat(link.get().getHref()).isEqualTo(COMMAND_BY_WIEDERVORLAGE_PATH);
			}

			@Test
			void shouldNotHaveLink() {
				EntityModel<Wiedervorlage> entityModel = modelAssembler.toModel(WiedervorlageTestFactory.create());

				var link = entityModel.getLink(WiedervorlageModelAssembler.REL_WIEDEREROEFFNEN);

				assertThat(link).isNotPresent();
			}
		}

		@Nested
		class TestAttachmentsLink {

			@Test
			void shouldHaveLink() {
				EntityModel<Wiedervorlage> entityModel = modelAssembler
						.toModel(WiedervorlageTestFactory.createBuilder().attachment(OzgFileTestFactory.ID).build());

				var link = entityModel.getLink(WiedervorlageModelAssembler.REL_ATTACHMENTS);

				assertThat(link).isPresent();
				assertThat(link.get().getHref()).isEqualTo("/api/wiedervorlages/" + WiedervorlageTestFactory.ID + "/attachments");
			}

			@Test
			void shouldNotHaveLink() {
				EntityModel<Wiedervorlage> entityModel = modelAssembler
						.toModel(WiedervorlageTestFactory.createBuilder().attachments(Collections.emptyList()).build());

				var link = entityModel.getLink(WiedervorlageModelAssembler.REL_ATTACHMENTS);

				assertThat(link).isNotPresent();
			}
		}

		EntityModel<Wiedervorlage> toModel() {
			return modelAssembler.toModel(WiedervorlageTestFactory.create());
		}
	}

	@Nested
	class TestLinksOnCollectionModel {

		@Test
		void shouldHaveCreateWiedervorlageLink() {
			var collectionModel = modelAssembler.toCollectionModel(Collections.singleton(WiedervorlageTestFactory.create()).stream(),
					VorgangHeaderTestFactory.ID);

			var link = collectionModel.getLink(WiedervorlageModelAssembler.REL_CREATE);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo("/api/vorgangs/" + VorgangHeaderTestFactory.ID + "/wiedervorlageCommands");
		}

		@Test
		void shouldHaveUploadFileLink() {
			var collectionModel = modelAssembler.toCollectionModel(Collections.singleton(WiedervorlageTestFactory.create()).stream(),
					VorgangHeaderTestFactory.ID);

			var link = collectionModel.getLink(WiedervorlageModelAssembler.REL_UPLOAD_FILE);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo("/api/binaryFiles/" + VorgangHeaderTestFactory.ID + "/wiedervorlageAttachment/file");
		}
	}
}