package de.itvsh.goofy.wiedervorlage;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.itvsh.goofy.common.ContextService;
import de.itvsh.goofy.vorgang.VorgangHeaderTestFactory;
import de.itvsh.ozg.pluto.grpc.command.GrpcCallContext;
import de.itvsh.ozg.pluto.grpc.command.GrpcCommandResponse;
import de.itvsh.ozg.pluto.grpc.command.GrpcCreateCommandRequest;
import de.itvsh.ozg.pluto.grpc.command.GrpcOrder;
import de.itvsh.ozg.pluto.wiedervorlage.GrpcFindWiedervorlagenByVorgangRequest;
import de.itvsh.ozg.pluto.wiedervorlage.GrpcFindWiedervorlagenByVorgangResponse;
import de.itvsh.ozg.pluto.wiedervorlage.GrpcGetWiedervorlageRequest;
import de.itvsh.ozg.pluto.wiedervorlage.GrpcGetWiedervorlageResponse;
import de.itvsh.ozg.pluto.wiedervorlage.GrpcWiedervorlage;
import de.itvsh.ozg.pluto.wiedervorlage.WiedervorlageServiceGrpc.WiedervorlageServiceBlockingStub;

class WiedervorlageRemoteServiceTest {

	@Spy // NOSONAR
	@InjectMocks
	private WiedervorlageRemoteService service;

	@Mock
	private WiedervorlageServiceBlockingStub serviceStub;
	@Mock
	private WiedervorlageMapper mapper;
	@Mock
	private ContextService contextService;

	private GrpcWiedervorlage wiedervorlage = GrpcWiedervorlage.newBuilder().setId("42").build();

	@Nested
	class TestFindByVorgangId {

		private GrpcFindWiedervorlagenByVorgangRequest request = GrpcFindWiedervorlagenByVorgangRequest.newBuilder()
				.setVorgangId(VorgangHeaderTestFactory.ID)
				.build();
		private GrpcFindWiedervorlagenByVorgangResponse response = GrpcFindWiedervorlagenByVorgangResponse.newBuilder()
				.addWiedervorlagen(wiedervorlage)
				.addWiedervorlagen(wiedervorlage).build();

		@BeforeEach
		void mockStub() {
			when(serviceStub.findWiedervorlagenByVorgangId(any())).thenReturn(response);
		}

		@Test
		void shouldCallStub() {
			service.findByVorgangId(VorgangHeaderTestFactory.ID);

			verify(serviceStub).findWiedervorlagenByVorgangId(request);
		}

		@Test
		void shouldCallMapper() {
			Stream<Wiedervorlage> result = service.findByVorgangId(VorgangHeaderTestFactory.ID);
			collectStreamElementsToTriggerLazyStream(result);

			verify(mapper, times(2)).toWiedervorlage(wiedervorlage);
		}

		private void collectStreamElementsToTriggerLazyStream(Stream<Wiedervorlage> stream) {
			stream.collect(Collectors.toList());
		}
	}

	@Nested
	class TestGetById {

		private GrpcGetWiedervorlageRequest request = GrpcGetWiedervorlageRequest.newBuilder().setId(WiedervorlageTestFactory.ID).build();
		private GrpcGetWiedervorlageResponse response = GrpcGetWiedervorlageResponse.newBuilder().setWiedervorlage(wiedervorlage).build();

		@BeforeEach
		void mockStub() {
			when(serviceStub.getWiedervorlage(any())).thenReturn(response);
		}

		@Test
		void shouldCallStub() {
			service.getById(WiedervorlageTestFactory.ID);

			verify(serviceStub).getWiedervorlage(request);
		}

		@Test
		void shouldCallMapper() {
			service.getById(WiedervorlageTestFactory.ID);

			verify(mapper, times(1)).toWiedervorlage(wiedervorlage);
		}
	}

	@Nested
	class TestBuildCommandRequest {

		private GrpcCallContext callContext = GrpcCallContext.newBuilder().build();
		private WiedervorlageCommand command = WiedervorlageCommandTestFactory.create();
		private GrpcWiedervorlage grpcWiedervorlage = GrpcWiedervorlageTestFactory.create();

		@BeforeEach
		void initTest() {
			when(contextService.createCallContext()).thenReturn(callContext);
			when(mapper.toGrpcWiedervorlage(any())).thenReturn(grpcWiedervorlage);
		}

		@Test
		@DisplayName("should have CallContext")
		void shouldHaveCallContext() {
			var request = doBuildRequest();

			assertThat(request.getCallContext()).isNotNull().isSameAs(callContext);
		}

		@Test
		@DisplayName("should have vorgangId as relation Id")
		void shouldHaveRelationId() {
			var request = doBuildRequest();

			assertThat(request.getRelationId()).isEqualTo(VorgangHeaderTestFactory.ID);
		}

		@Test
		@DisplayName("should set create Wiedervorlage order")
		void shouldSetOrder() {
			var request = doBuildRequest();

			assertThat(request.getOrder()).isEqualTo(GrpcOrder.CREATE_WIEDERVORLAGE);
		}

		@Test
		@DisplayName("should map wiedervorlage")
		void shouldMapWiedervorlage() {
			var request = doBuildRequest();

			verify(mapper).toGrpcWiedervorlage(command.getWiedervorlage());
			assertThat(request.getWiedervorlage()).isSameAs(grpcWiedervorlage);
		}

		private GrpcCreateCommandRequest doBuildRequest() {
			return service.buildCreateCommandRequest(VorgangHeaderTestFactory.ID, command);
		}
	}

	@Nested
	class TestCreateCreateCommand {
		private GrpcCreateCommandRequest request = GrpcCreateCommandRequest.newBuilder().build();

		@BeforeEach
		void initTest() {
			doReturn(request).when(service).buildCreateCommandRequest(anyString(), any());
			when(serviceStub.createCommand(any())).thenReturn(GrpcCommandResponse.newBuilder().build());
		}

		@Test
		@DisplayName("Should call buildCreateCommandRequest with VorgangId")
		void shouldCallCreateRequestWithVorgangId() {
			doRequest();

			verify(service).buildCreateCommandRequest(eq(VorgangHeaderTestFactory.ID), any());
		}

		@Test
		void shouldCallStub() {
			doRequest();

			verify(serviceStub).createCommand(request);
		}

		@Test
		void shouldMapResult() {
			doRequest();

			verify(mapper).toWiedervorlageCommand(any());
		}

		private void doRequest() {
			service.sendCreateWiedervorlageCommand(VorgangHeaderTestFactory.ID, WiedervorlageCommandTestFactory.create());
		}
	}

	@Nested
	class TestCreateEditCommand {

		private GrpcCreateCommandRequest request = GrpcCreateCommandRequest.newBuilder().build();

		@BeforeEach
		void initTest() {
			doReturn(request).when(service).buildCreateCommandRequest(anyString(), any());
			when(serviceStub.createCommand(any())).thenReturn(GrpcCommandResponse.newBuilder().build());
		}

		@Test
		void shouldCreateRequestWithWiedervorlageId() {
			doRequest();

			verify(service).buildCreateCommandRequest(eq(WiedervorlageTestFactory.ID), any());
		}

		@Test
		void shouldCallStub() {
			doRequest();

			verify(serviceStub).createCommand(request);
		}

		@Test
		void shouldMapResult() {
			doRequest();

			verify(mapper).toWiedervorlageCommand(any());
		}

		private void doRequest() {
			service.sendWiedervorlageCommand(WiedervorlageTestFactory.ID, WiedervorlageCommandTestFactory.create());
		}
	}

	@Nested
	class TestAddWiedervorlageToRequest {

		private GrpcCreateCommandRequest.Builder requestBuilder = GrpcCreateCommandRequest.newBuilder();

		@Nested
		class TestWithWiedervorlage {

			WiedervorlageCommand command = WiedervorlageCommandTestFactory.create();

			@BeforeEach
			void initTest() {
				when(mapper.toGrpcWiedervorlage(any())).thenReturn(GrpcWiedervorlageTestFactory.create());
			}

			@Test
			void shouldMapWiedervorlage() {
				service.addWiedervorlage(requestBuilder, command);

				verify(mapper).toGrpcWiedervorlage(any());
			}

			@Test
			void shouldAddWiedervorlageToRequest() {
				var result = service.addWiedervorlage(requestBuilder, command);

				assertThat(result.getWiedervorlage().getId()).isEqualTo(WiedervorlageTestFactory.ID);
			}
		}

		@Nested
		class TestWithoutWiedervorlage {

			WiedervorlageCommand command = WiedervorlageCommandTestFactory.createBuilder().wiedervorlage(null).build();

			@Test
			void shouldNotMapWiedervorlage() {
				service.addWiedervorlage(requestBuilder, command);

				verifyNoInteractions(mapper);
			}
		}
	}
}